﻿using System.Collections.Generic;

namespace E39
{
    /// <summary>
    /// Configuration for the E39 wave data pipeline.
    /// </summary>
    /// <remarks>
    /// The configuration consists of general settings which usually don't change.
    /// This includes for instance the area of the Excel templates in which to fill data,
    /// the order of the fields in the templates, and parameters for the imputation algorithms.
    /// </remarks>
    public class Config
    {
        /// <summary>
        /// The base URL of the thredds.met.no file server from which to download the raw data files.
        /// </summary>
        public string DownloadUrl { get; set; }

        /// <summary>
        /// The types of raw data files that will be downloaded and loaded. E.g. "wind" and "wave".
        /// </summary>
        public List<string> FileDataTypes { get; set; }

        /// <summary>
        /// List of all available station names.
        /// </summary>
        public List<string> AllStationNames { get; set; }

        /// <summary>
        /// Path to the Python executable.
        /// </summary>
        public string PythonExecutablePath { get; set; }

        /// <summary>
        /// The path to the directory in which raw data files reside, or into which they'll be downloaded and converted.
        /// </summary>
        public string RawDataDirectory { get; set; }

        /// <summary>
        /// The path to the directory in which generated Excel reports will be placed.
        /// </summary>
        public string ExcelOutputDirectory { get; set; }

        /// <summary>
        /// The path to the directory containing the Excel templates.
        /// </summary>
        public string ExcelTemplatesDirectory { get; set; }

        /// <summary>
        /// The row number of the first row to fill in the Excel templates. Index starts at 1.
        /// </summary>
        public int ExcelStartRow { get; set; }

        /// <summary>
        /// The column number of the first column to fill in the Excel templates. Index starts at 1.
        /// </summary>
        public int ExcelStartColumn { get; set; }

        /// <summary>
        /// An ordered list of feature names, in the order of the columns in the Excel template,
        /// with names matching the variable names in the NetCDF files.
        /// </summary>
        public List<string> ExcelFeaturesOrder { get; set; }

        /// <summary>
        /// An ordered list of feature names, in the order of the columns in the Excel SCAT template,
        /// with names matching the variable names in the NetCDF files.
        /// </summary>
        public List<string> ExcelScatFeaturesOrder { get; set; }

        /// <summary>
        /// A list of feature names to impute.
        /// </summary>
        public List<string> FeaturesToImpute { get; set; }

        /// <summary>
        /// The K parameter for the kNN algorithm, which is the number of nearest neighbors
        /// to calculate weighted average of when imputing an empty data point.
        /// </summary>
        public int KNNParameter_K { get; set; }

        /// <summary>
        /// The N parameter for the kNN algorithm, which is the number of data points back and forth in time
        /// to consider when looking for the nearest neighbors of an empty data point.
        /// </summary>
        public int KNNParameter_N { get; set; }

        /// <summary>
        /// Whether to debug the pipeline. Indicates to the components in the pipeline that they can be
        /// extra verbose and possibly write stat files for debugging.
        /// </summary>
        public bool Debug { get; set; }

        /// <summary>
        /// Returns a string representation of the <see cref="Config"/> object with all its properties and their values.
        /// </summary>
        /// <returns>The string representation of the <see cref="Config"/>.</returns>
        public override string ToString()
        {
            var str = "Config {\n";

            foreach (var prop in GetType().GetProperties())
            {
                var propValue = prop.GetValue(this);
                if (propValue is System.Collections.IList list)
                {
                    var strList = new List<string>();
                    foreach (var item in list)
                    {
                        strList.Add(item.ToString());
                    }
                    propValue = string.Join(", ", strList);
                }
                str += $"    {prop.Name}: {propValue}\n";
            }

            return str + "}";
        }
    }
}