﻿using CsvHelper;
using System;
using System.Collections.Generic;
using System.Globalization;
using System.IO;

namespace E39.DataLoading
{
    public class CsvDataLoader : DataLoader
    {
        public CsvDataLoader(Config config, string stationName) : base(config, stationName)
        {
        }

        /// <summary>
        ///     Loads data from a CSV file into a DataTable. The file must be semicolon separated.
        /// </summary>
        /// <param name="filePath">Path of the CSV file.</param>
        public override void LoadData(string filePath)
        {
            var file = new FileInfo(filePath);
            if (!file.Exists)
            {
                throw new FileNotFoundException($"The csvFilePath: {filePath} is not correct.", filePath);
            }

            using (var streamReader = new StreamReader(filePath))
            using (var csvReader = new CsvReader(streamReader))
            {
                csvReader.Configuration.Delimiter = ";";
                csvReader.Configuration.MissingFieldFound = null;
                csvReader.Configuration.CultureInfo = CultureInfo.InvariantCulture;

                csvReader.Read();
                csvReader.ReadHeader();

                // Find all fields which we can read from the csv into the datatable.
                // Start with all column names in the csv, and exclude those which are not in the datatable schema
                List<string> fieldsToRead = new List<string>(csvReader.Context.HeaderRecord);
                fieldsToRead.RemoveAll(field => !Data.Columns.Contains(field));
                // Skip reading time field because we'll do that manually
                fieldsToRead.Remove("time");

                // Read all rows in the csv file
                while (csvReader.Read())
                {
                    double csvUnixTimeStamp = csvReader.GetField<double>("time");
                    var csvDateTime = new DateTime(1970, 1, 1, 0, 0, 0, 0, DateTimeKind.Utc).AddSeconds(csvUnixTimeStamp);

                    // Find the row in the datatable where timeInterval matches the data point timestamp
                    var rows = Data.Select($"timeInterval = #{csvDateTime.ToString(CultureInfo.InvariantCulture.DateTimeFormat)}#");

                    // Make sure there exists a datatable row for this timestamp
                    if (rows.Length > 0)
                    {
                        var row = rows[0];
                        row["time"] = csvDateTime;

                        foreach (var field in fieldsToRead)
                        {
                            string dataInField = csvReader.GetField<string>(field);

                            if (double.TryParse(dataInField, NumberStyles.Number, CultureInfo.InvariantCulture, out double parsedData))
                            {
                                row[field] = parsedData;
                            }
                        }
                    }
                }
            }
        }
    }
}