﻿using E39.Imputation;
using System;
using System.Data;
using System.Linq;

namespace E39.DataLoading
{
    public abstract class DataLoader
    {
        public Config Config { get; }
        public string StationName { get; }
        public DataTable Data { get; set; }
        public int MissingRows { get; private set; }

        protected DataLoader(Config config, string stationName)
        {
            Config = config;
            StationName = stationName;
            Data = CreateDataTable();
        }

        /// <summary>
        /// Creates the DataTable with column headers and types.
        /// </summary>
        /// <returns>The newly created DataTable</returns>
        public DataTable CreateDataTable()
        {
            DataTable dt = new DataTable("StationData_" + StationName);
            dt.ExtendedProperties.Add("StationName", StationName);

            dt.Columns.Add("timeInterval", typeof(DateTime));
            dt.Columns.Add("time", typeof(DateTime));

            dt.PrimaryKey = new DataColumn[] { dt.Columns["timeInterval"] };

            foreach (string columnHeader in Config.ExcelFeaturesOrder.Union(Config.ExcelScatFeaturesOrder))
            {
                if (columnHeader != "timeInterval" && columnHeader != "time")
                {
                    dt.Columns.Add(columnHeader, typeof(double));
                }
            }

            // Make a column indicating whether a datapoint in the station data came from imputation
            dt.Columns.Add(new DataColumn("Imputed", typeof(bool)) { DefaultValue = false });

            return dt;
        }

        /// <summary>
        /// Fills inn the timeInterval column in the DataTable. There are 10 minutes between each row added.
        /// </summary>
        /// <param name="startDate">Start date of interval.</param>
        /// <param name="endDate">End date of interval.</param>
        public void FillInTimeColumn(DateTimeOffset startDate, DateTimeOffset endDate)
        {
            if (startDate.Year < 1970 || startDate > endDate)
            {
                throw new ArgumentException("Parameters startDate or endDate have invalid values. " +
                                            "Make sure that startDate.Year is bigger than 1970," +
                                            "Make sure that startDate is equal to or before endDate", nameof(startDate));
            }

            for (var timeInterval = startDate; timeInterval <= endDate; timeInterval = timeInterval.AddMinutes(10))
            {
                DataRow newRow = Data.NewRow();
                newRow["timeInterval"] = timeInterval.UtcDateTime;
                Data.Rows.Add(newRow);
            }
        }

        public int CountEmptyRows()
        {
            int missingRows = 0;

            foreach (DataRow row in Data.Rows)
            {
                if (DataPointValidator.IsDataPointEmpty(row))
                {
                    missingRows++;
                }
            }

            this.MissingRows = missingRows;
            return missingRows;
        }

        /// <summary>
        /// Loads data from a raw data file into the <see cref="Data"/> DataTable.
        /// </summary>
        /// <param name="filePath">Path to the raw data file.</param>
        public abstract void LoadData(string filePath);
    }
}