﻿using Microsoft.Research.Science.Data;
using System;
using System.Collections.Generic;
using System.Globalization;
using System.Linq;

namespace E39.DataLoading
{
    public class NetCDFDataLoader : DataLoader
    {
        public NetCDFDataLoader(Config config, string stationName) : base(config, stationName)
        {
        }

        public override void LoadData(string filePath)
        {
            using (var netcdf = DataSet.Open(filePath))
            {
                Data.ExtendedProperties["BuoySerialNo"] = netcdf.Metadata["buoy_serialno"];

                // Pre-load data from all variables
                var variableData = new Dictionary<string, Array>();
                foreach (var variable in netcdf.Variables)
                {
                    variableData[variable.Name] = variable.GetData();
                }

                // Find all fields which we can read from the netcdf into the datatable.
                // Start with all column names in the netcdf, and exclude those which are not in the datatable schema
                // Skip reading time field because we'll do that manually
                var featuresToRead = netcdf
                    .Select(variable => variable.Name)
                    .Where(feature => Data.Columns.Contains(feature) && feature != "time");

                // Loop through each index of the dimension of the data
                for (int i = 0; i < netcdf.Dimensions[0].Length; ++i)
                {
                    var unixTime = (double)variableData["time"].GetValue(i);
                    var time = new DateTime(1970, 1, 1, 0, 0, 0, 0, DateTimeKind.Utc).AddSeconds(unixTime);

                    var rows = Data.Select($"timeInterval = #{time.ToString(CultureInfo.InvariantCulture.DateTimeFormat)}#");

                    if (rows.Length > 0)
                    {
                        var row = rows[0];
                        row["time"] = time;

                        foreach (var feature in featuresToRead)
                        {
                            try
                            {
                                var value = Convert.ToDouble(variableData[feature].GetValue(i));
                                if (!double.IsNaN(value))
                                {
                                    row[feature] = value;
                                }
                            }
                            catch (IndexOutOfRangeException)
                            {
                                // If, somehow, this variable has fewer items than the time-dimension
                            }
                        }
                    }
                }
            }
        }
    }
}