﻿using E39;
using E39.Imputation;
using OfficeOpenXml;
using OfficeOpenXml.Drawing;
using OfficeOpenXml.Drawing.Chart;
using OfficeOpenXml.Style;
using OfficeOpenXml.Table;
using System;
using System.Collections.Generic;
using System.Data;
using System.Diagnostics;
using System.Drawing;
using System.IO;
using System.Threading.Tasks;

namespace E39.Excel
{
    /// <summary>
    /// Writes Excel report files for station data.
    /// </summary>
    public class ExcelWriter : IDisposable
    {
        private Config Config { get; }
        private int StartRow { get; }
        private int StartColumn { get; }
        private List<string> ColumnHeaders { get; }
        private readonly int NumberOfRowsPerDataTable;

        public DateTimeOffset StartDate { get; }
        public DateTimeOffset EndDate { get; }

        private DataTable _data;

        /// <summary>
        /// The <see cref="DataTable"/> containing the station data to write to the Excel reports.
        /// </summary>
        public DataTable Data
        {
            get
            {
                return _data;
            }
            set
            {
                ClearExcelDocuments();
                _data = value;
            }
        }

        private readonly ExcelPackage ExcelTemplate;
        private readonly ExcelPackage ScatTotalExcelTemplate;
        private readonly ExcelPackage ScatSwellExcelTemplate;
        private readonly ExcelPackage ScatVindExcelTemplate;

        private ExcelWorksheet WorksheetRawdata;
        private ExcelWorksheet WorksheetScatterPlots;

        private ExcelWorksheet ScatTotalWorksheetRawData;
        private ExcelWorksheet ScatTotalWorksheetData;
        private ExcelWorksheet ScatTotalWorksheetHs_dir;
        private ExcelWorksheet ScatTotalWorksheetHs_Tp;
        private ExcelWorksheet ScatTotalWorksheet3D;
        private ExcelWorksheet ScatTotalWorksheet3D_max;

        private ExcelWorksheet ScatSwellWorksheetRawData;
        private ExcelWorksheet ScatSwellWorksheetData;
        private ExcelWorksheet ScatSwellWorksheetHswell_dir;
        private ExcelWorksheet ScatSwellWorksheetHswell_Hwind;

        private ExcelWorksheet ScatVindWorksheetRawData;
        private ExcelWorksheet ScatVindWorksheetData;
        private ExcelWorksheet ScatVindWorksheetU_dir;

        private List<ExcelDrawing> AddedCharts;
        private List<ExcelWorksheet> Worksheets;

        /// <summary>
        /// Constructs a new ExcelWriter. Loads Excel templates and worksheets.
        /// </summary>
        /// <param name="config">The pipeline configuration.</param>
        /// <param name="startDate">The starting date of the station data to write reports for. Used for determining graph range.</param>
        /// <param name="endDate">The end date of the station data to write reports for. Used for determining graph range.</param>
        public ExcelWriter(Config config, DateTimeOffset startDate, DateTimeOffset endDate, string templateDirectory, int numberOfRowsPerDataTable)
        {
            Config = config;
            StartDate = startDate;
            EndDate = endDate;
            Data = null;
            ColumnHeaders = Config.ExcelFeaturesOrder;
            StartRow = config.ExcelStartRow;
            StartColumn = config.ExcelStartColumn;
            NumberOfRowsPerDataTable = numberOfRowsPerDataTable;

            if (startDate >= endDate)
            {
                throw new ArgumentException("Parameter endDate is set to before the excels start date. " +
                    "endDate need to be set to after the excels start date.", nameof(endDate));
            }

            // Locate the template file
            var excelTemplateFilePath = Path.Combine(templateDirectory, "excel_template.xlsx");
            var excelTemplateFile = new FileInfo(excelTemplateFilePath);
            if (!excelTemplateFile.Exists)
            {
                throw new FileNotFoundException("The excel template file does not exist.", excelTemplateFilePath);
            }

            // Locate the SCAT_total_excel_template file
            var ScatTotalExcelTemplateFilePath = Path.Combine(templateDirectory, "SCAT_total_excel_template.xlsx");
            var ScatTotalExcelTemplateFile = new FileInfo(ScatTotalExcelTemplateFilePath);
            if (!ScatTotalExcelTemplateFile.Exists)
            {
                throw new FileNotFoundException("The SCAT_total_excel_template.xlsx file does not exist.", ScatTotalExcelTemplateFilePath);
            }

            // Locate the SCAT_swell_excel_template file
            var ScatSwellExcelTemplateFilePath = Path.Combine(templateDirectory, "SCAT_swell_excel_template.xlsx");
            var ScatSwellExcelTemplateFile = new FileInfo(ScatSwellExcelTemplateFilePath);
            if (!ScatSwellExcelTemplateFile.Exists)
            {
                throw new FileNotFoundException("The SCAT_swell_excel_template.xlsx file does not exist.", ScatSwellExcelTemplateFilePath);
            }

            // Locate the SCAT_vind_excel_template file
            var ScatVindExcelTemplateFilePath = Path.Combine(templateDirectory, "SCAT_vind_excel_template.xlsx");
            var ScatVindExcelTemplateFile = new FileInfo(ScatVindExcelTemplateFilePath);
            if (!ScatVindExcelTemplateFile.Exists)
            {
                throw new FileNotFoundException("The SCAT_vind_excel_template.xlsx file does not exist.", ScatVindExcelTemplateFilePath);
            }


            // Load the templates and worksheets
            ExcelTemplate = new ExcelPackage(excelTemplateFile);
            ScatTotalExcelTemplate = new ExcelPackage(ScatTotalExcelTemplateFile);
            ScatSwellExcelTemplate = new ExcelPackage(ScatSwellExcelTemplateFile);
            ScatVindExcelTemplate = new ExcelPackage(ScatVindExcelTemplateFile);

            WorksheetRawdata = ExcelTemplate.Workbook.Worksheets[0];
            WorksheetScatterPlots = ExcelTemplate.Workbook.Worksheets[1];

            ScatTotalWorksheetRawData = ScatTotalExcelTemplate.Workbook.Worksheets[0];
            ScatTotalWorksheetData = ScatTotalExcelTemplate.Workbook.Worksheets[1];
            ScatTotalWorksheetHs_dir = ScatTotalExcelTemplate.Workbook.Worksheets[2];
            ScatTotalWorksheetHs_Tp = ScatTotalExcelTemplate.Workbook.Worksheets[3];
            ScatTotalWorksheet3D = ScatTotalExcelTemplate.Workbook.Worksheets[4];
            ScatTotalWorksheet3D_max = ScatTotalExcelTemplate.Workbook.Worksheets[5];

            ScatSwellWorksheetRawData = ScatSwellExcelTemplate.Workbook.Worksheets[0];
            ScatSwellWorksheetData = ScatSwellExcelTemplate.Workbook.Worksheets[1];
            ScatSwellWorksheetHswell_dir = ScatSwellExcelTemplate.Workbook.Worksheets[2];
            ScatSwellWorksheetHswell_Hwind = ScatSwellExcelTemplate.Workbook.Worksheets[3];

            ScatVindWorksheetRawData = ScatVindExcelTemplate.Workbook.Worksheets[0];
            ScatVindWorksheetData = ScatVindExcelTemplate.Workbook.Worksheets[1];
            ScatVindWorksheetU_dir = ScatVindExcelTemplate.Workbook.Worksheets[2];

            Worksheets = new List<ExcelWorksheet>()
            {
                WorksheetScatterPlots,

                ScatTotalWorksheetRawData,
                ScatTotalWorksheetData,
                ScatTotalWorksheetHs_dir,
                ScatTotalWorksheetHs_Tp, 
                ScatTotalWorksheet3D,
                ScatTotalWorksheet3D_max,

                ScatSwellWorksheetRawData,
                ScatSwellWorksheetData,
                ScatSwellWorksheetHswell_dir,
                ScatSwellWorksheetHswell_Hwind,

                ScatVindWorksheetRawData,
                ScatVindWorksheetData,
                ScatVindWorksheetU_dir
            };

            // Extend formulas in templates to correct size
            ExtendFormulas(NumberOfRowsPerDataTable);

            AddedCharts = new List<ExcelDrawing>();
        }

        /// <summary>
        /// Fills data for a station into both the main and SCAT Excel templates and makes the neccessary graphs.
        /// </summary>
        /// <param name="stationData">The <see cref="DataTable"/> with data for the station to make reports for.</param>
        public void MakeReports(DataTable stationData)
        {
            Data = stationData;

            AddMetaData();

            WriteToExcel();
            WriteScatToExcel();

            WriteTimeSeriesGraphToExcel("chart1", $"Waves - {Data.ExtendedProperties["StationName"]}", 500, 2000, 26, 8, 2, "Date", new string[] { "Hm0", "hm0a", "hm0b", "tp" }, "Significant wave height, Hmo", new string[] { "thtp" }, "Wave direction");
            WriteTimeSeriesGraphToExcel("chart2", $"Wind - {Data.ExtendedProperties["StationName"]}", 500, 2000, 26, 38, 2, "Date", new string[] { "WindSpeed" }, "Wind speed, 10 min m/s", new string[] { "WindDirection" }, "Wind direction");
            WriteTimeSeriesGraphToExcel("chart3", $"Significant Wave height, Hs - {Data.ExtendedProperties["StationName"]}", 500, 2000, 26, 68, 2, "Date", new string[] { "Hm0" }, "Significant wave height, Hmo");
            WriteTimeSeriesGraphToExcel("chart4", $"Spectral top period, Tp - {Data.ExtendedProperties["StationName"]}", 500, 2000, 26, 98, 2, "Date", new string[] { "tp" }, "Spectral Top period, Tp");
            WriteTimeSeriesGraphToExcel("chart5", $"Maximum Wave height, Hmax - {Data.ExtendedProperties["StationName"]}", 500, 2000, 26, 128, 2, "Date", new string[] { "hmax" }, "Maximum wave height, Hmo");
            WriteTimeSeriesGraphToExcel("chart6", $"Wave direction, thtp - {Data.ExtendedProperties["StationName"]}", 500, 2000, 26, 158, 2, "Date", new string[] { "thtp" }, "Wave direction, Hmo");

            CreateScatterCharts($"Wind speed vs wave height - {Data.ExtendedProperties["StationName"]}", "Wind speed m/s", "Wave height, hm0", $"Vindhastighet vs bølgehøyde - {Data.ExtendedProperties["StationName"]}", "Vindhastighet m/s", "hm0", 10, 1000, 700);
        }

        /// <summary>
        /// Writes the station data into the main Excel report, according to the feature names,
        /// column order and starting position in the <see cref="E39.Config"/>.
        /// </summary>
        /// <exception cref="InvalidOperationException">If the <see cref="Data"/> property not has been set to an instance of a <see cref="DataTable"/>.</exception>
        /// <seealso cref="Config.ExcelFeaturesOrder"/>
        /// <seealso cref="Config.ExcelStartRow"/>
        /// <seealso cref="Config.ExcelStartColumn"/>
        public void WriteToExcel()
        {
            if (Data == null)
            {
                throw new InvalidOperationException("You must set the Data property to an instance of a DataTable first.");
            }

            Console.WriteLine($"Writing {Data.Rows.Count} rows to the excel spreadsheet");

            // Variable row is the current excel row index
            for (int row = StartRow; row < Data.Rows.Count + StartRow; ++row)
            {
                // DataTable row index is row - StartRow
                DataRow data = Data.Rows[row - StartRow];

                // Variable row is the current excel row index
                int col = StartColumn;
                foreach (var feature in Config.ExcelFeaturesOrder)
                {
                    WorksheetRawdata.SetValue(row, col++, data[feature]);
                }

                if (data.Field<bool>("Imputed"))
                {
                    var excelRow = WorksheetRawdata.Cells[row, StartColumn, row, StartColumn + Config.ExcelFeaturesOrder.Count - 1];  // -1 since range is inclusive
                    excelRow.Style.Fill.PatternType = ExcelFillStyle.Solid;
                    excelRow.Style.Fill.BackgroundColor.SetColor(Color.LightYellow);
                }
            }
        }

        /// <summary>
        /// Writes the station data into the SCAT Excel report, according to the feature names,
        /// column order and starting position in the <see cref="E39.Config"/>.
        /// </summary>
        /// <exception cref="InvalidOperationException">If the <see cref="Data"/> property not has been set to an instance of a <see cref="DataTable"/>.</exception>
        /// <seealso cref="Config.ExcelScatFeaturesOrder"/>
        /// <seealso cref="Config.ExcelStartRow"/>
        /// <seealso cref="Config.ExcelStartColumn"/>
        public void WriteScatToExcel()
        {
            if (Data == null)
            {
                throw new InvalidOperationException("You must set the Data property to an instance of a DataTable first.");
            }

            Console.WriteLine($"Writing {Data.Rows.Count} rows to the excel spreadsheet (SCAT)");

            // Variable row is the current excel row index
            for (int row = StartRow; row < Data.Rows.Count + StartRow; ++row)
            {
                // DataTable row index is row - StartRow
                DataRow data = Data.Rows[row - StartRow];

                // Variable col is the current excel column index
                int col = StartColumn;
                foreach (var feature in Config.ExcelScatFeaturesOrder)
                {
                    ScatTotalWorksheetRawData.SetValue(row, col, data[feature]);
                    ScatSwellWorksheetRawData.SetValue(row, col, data[feature]);
                    ScatVindWorksheetRawData.SetValue(row, col, data[feature]);

                    col++;
                }

                if (data.Field<bool>("Imputed"))
                {
                    var excelRowScatTotal = ScatTotalWorksheetRawData.Cells[row, StartColumn, row, StartColumn + Config.ExcelScatFeaturesOrder.Count - 1];  // -1 since range is inclusive
                    excelRowScatTotal.Style.Fill.PatternType = ExcelFillStyle.Solid;
                    excelRowScatTotal.Style.Fill.BackgroundColor.SetColor(Color.LightYellow);

                    var excelRowScatSwell = ScatSwellWorksheetRawData.Cells[row, StartColumn, row, StartColumn + Config.ExcelScatFeaturesOrder.Count - 1];  // -1 since range is inclusive
                    excelRowScatSwell.Style.Fill.PatternType = ExcelFillStyle.Solid;
                    excelRowScatSwell.Style.Fill.BackgroundColor.SetColor(Color.LightYellow);

                    var excelRowScatVind = ScatVindWorksheetRawData.Cells[row, StartColumn, row, StartColumn + Config.ExcelScatFeaturesOrder.Count - 1];  // -1 since range is inclusive
                    excelRowScatVind.Style.Fill.PatternType = ExcelFillStyle.Solid;
                    excelRowScatVind.Style.Fill.BackgroundColor.SetColor(Color.LightYellow);
                }
            }
        }

        /// <summary>
        /// Writes a graph to the main Excel report. It is important that chartName is unique. Parameters height and width are in pixels.
        /// Parameters xPos and yPos are column and row respectivly. Parameter endDate is the date where the graph stops getting data. Start date for the
        /// time series is the top row from the excel spread sheet. Parameters secondaryYAxis and secondaryYAxisTitle are optional, and null as default.
        /// </summary>
        /// <param name="chartName"> Name of chart, must be unique. </param>
        /// <param name="graphTitle"> Top title on graph. </param>
        /// <param name="height"> Height of graph box in pixels. </param>
        /// <param name="width"> Width of graph box in pixels. </param>
        /// <param name="xPos"> Column start for top left corner of graph box. </param>
        /// <param name="yPos"> Row start for top left corner of graph box. </param>
        /// <param name="xAxis"> Column position of xAxis data. </param>
        /// <param name="xAxisTitle"> Title for the xAxis. </param>
        /// <param name="yAxis"> yAxis column names as an array. </param>
        /// <param name="yAxisTitle"> Title for the yAxis. </param>
        /// <param name="secondaryYAxis"> Secondary yAxis column names as an array. </param>
        /// <param name="secondaryYAxisTitle"> Title for the secondary yAxis. </param>
        /// <exception cref="InvalidOperationException">If the <see cref="Data"/> property not has been set to an instance of a <see cref="DataTable"/>.</exception>
        public void WriteTimeSeriesGraphToExcel(string chartName, string graphTitle, int height, int width, int xPos, int yPos, int xAxis, string xAxisTitle, string[] yAxis, string yAxisTitle, string[] secondaryYAxis = null, string secondaryYAxisTitle = null)
        {
            if (Data == null)
            {
                throw new InvalidOperationException("You must set the Data property to an instance of a DataTable first.");
            }

            Console.WriteLine("Making graph: {0}", graphTitle);

            if (height <= 0) throw new ArgumentException(string.Format("Parameter {0} must be larger than 0.", height), nameof(height));
            if (width <= 0) throw new ArgumentException(string.Format("Parameter {0} must be larger than 0.", width), nameof(width));
            if (xPos <= 0) throw new ArgumentException(string.Format("Parameter {0} must be larger than 0.", xPos), nameof(xPos));
            if (yPos <= 0) throw new ArgumentException(string.Format("Parameter {0} must be larger than 0.", yPos), nameof(yPos));
            if (xAxis <= 0) throw new ArgumentException(string.Format("Parameter {0} must be larger than 0.", xAxis), nameof(xAxis));

            var scatterPlot = (ExcelScatterChart)WorksheetRawdata.Drawings.AddChart(chartName, eChartType.XYScatterLines);
            AddedCharts.Add(scatterPlot);

            scatterPlot.SetPosition(yPos, 0, xPos, 0);
            scatterPlot.PlotArea.Fill.Color = Color.FloralWhite;
            scatterPlot.SetSize(width, height);
            scatterPlot.RoundedCorners = false;
            scatterPlot.XAxis.Title.Text = xAxisTitle;
            scatterPlot.XAxis.Title.Font.Size = 14;
            scatterPlot.YAxis.Title.Text = yAxisTitle;
            scatterPlot.YAxis.Title.Font.Size = 14;

            scatterPlot.XAxis.MinValue = (StartDate - new DateTimeOffset(1900, 1, 1, 0, 0, 0, TimeSpan.Zero)).TotalDays;
            scatterPlot.XAxis.MaxValue = (EndDate - new DateTimeOffset(1900, 1, 1, 0, 0, 0, TimeSpan.Zero)).TotalDays + 5;

            int endRowOffset = Data.Rows.Count - 1;  // -1 because excel ranges are inclusive

            if (secondaryYAxis != null)
            {
                var scatterPlot2 = scatterPlot.PlotArea.ChartTypes.Add(eChartType.XYScatterLines);

                foreach (string columnNameToPlot in secondaryYAxis)
                {
                    // The number of columns this column to plot is from StartColumn
                    var columnOffset = ColumnHeaders.IndexOf(columnNameToPlot);
                    // Check if the column name is in ColumnHeaders
                    if (columnOffset >= 0)
                    {
                        var serie2 = (ExcelScatterChartSerie)scatterPlot2.Series.Add(
                            // The Y-axis of the series
                            ExcelCellBase.GetAddress(
                                StartRow, StartColumn + columnOffset,                 // Start cell of column to plot
                                StartRow + endRowOffset, StartColumn + columnOffset), // End cell of column to plot

                            // The X-axis of the series
                            ExcelCellBase.GetAddress(
                                StartRow, xAxis,                // Start cell of column to plot
                                StartRow + endRowOffset, xAxis) // End cell of column to plot
                        );
                        serie2.Marker = eMarkerStyle.None;
                        serie2.Header = columnNameToPlot;

                        serie2.LineWidth = 0.5;
                    }
                }

                scatterPlot2.UseSecondaryAxis = true;
                scatterPlot2.YAxis.LabelPosition = eTickLabelPosition.High;
                scatterPlot2.YAxis.Title.Text = secondaryYAxisTitle;
                scatterPlot2.YAxis.Title.Font.Size = 14;
            }

            foreach (string columnNameToPlot in yAxis)
            {
                // The number of columns this column to plot is from StartColumn
                var columnOffset = ColumnHeaders.IndexOf(columnNameToPlot);
                // Check if the column name is in ColumnHeaders
                if (columnOffset >= 0)
                {
                    var serie1 = (ExcelScatterChartSerie)scatterPlot.Series.Add(
                        // The Y-axis of the series
                        ExcelCellBase.GetAddress(
                            StartRow, StartColumn + columnOffset,
                            StartRow + endRowOffset, StartColumn + columnOffset),

                        // The X-axis of the series
                        ExcelCellBase.GetAddress(
                            StartRow, xAxis,
                            StartRow + endRowOffset, xAxis)
                    );
                    serie1.Marker = eMarkerStyle.None;
                    serie1.Header = columnNameToPlot;
                    serie1.Border.LineStyle = eLineStyle.Solid;
                    serie1.LineWidth = 0.5;
                }
            }

            scatterPlot.Title.Text = graphTitle;
            scatterPlot.XAxis.Format = "dd-MM-yyyy";
        }

        /// <summary>
        /// Add meta data to the excel and excel_scat files.
        /// </summary>
        private void AddMetaData()
        {
            if (Data == null)
            {
                throw new InvalidOperationException("You must set the Data property to an instance of a DataTable first.");
            }

            int rowsMissing = 0;
            bool valuesImputed = false;

            foreach (DataRow row in Data.Rows)
            {
                if ((bool)row["Imputed"] || DataPointValidator.IsDataPointEmpty(row))
                {
                    if ((bool)row["Imputed"]) valuesImputed = true;

                    rowsMissing++;
                }
            }

            // Add meta data to Excel report
            WorksheetRawdata.Cells[1, 2].Value = Data.ExtendedProperties["StationName"];
            WorksheetRawdata.Cells[2, 2].Value = Data.Rows[0]["timeInterval"];
            WorksheetRawdata.Cells[3, 2].Value = Data.Rows[Data.Rows.Count - 1]["timeInterval"];
            WorksheetRawdata.Cells[1, 6].Value = Data.Rows.Count;
            WorksheetRawdata.Cells[2, 6].Value = rowsMissing;
            WorksheetRawdata.Cells[1, 12].Value = valuesImputed;
            WorksheetRawdata.Cells[7, 3].Value = Data.ExtendedProperties["StationName"];

            // Fomart meta data for Excel report
            WorksheetRawdata.Cells[1, 1, 3, 11].Style.Font.Bold = true;
            WorksheetRawdata.Cells[1, 1, 3, 11].Style.Font.Size = 11;

            // Add meta data to SCAT_total Excel report
            ScatTotalWorksheetRawData.Cells[1, 2].Value = Data.ExtendedProperties["StationName"];
            ScatTotalWorksheetRawData.Cells[2, 2].Value = Data.Rows[0]["timeInterval"];
            ScatTotalWorksheetRawData.Cells[3, 2].Value = Data.Rows[Data.Rows.Count - 1]["timeInterval"];
            ScatTotalWorksheetRawData.Cells[1, 6].Value = Data.Rows.Count;
            ScatTotalWorksheetRawData.Cells[2, 6].Value = rowsMissing;
            ScatTotalWorksheetRawData.Cells[1, 11].Value = valuesImputed;
            // Format meta data for SCAT_total Excel report
            ScatTotalWorksheetRawData.Cells[1, 1, 3, 11].Style.Font.Bold = true;
            ScatTotalWorksheetRawData.Cells[1, 1, 3, 11].Style.Font.Size = 11;

            // Add meta data to SCAT_swell Excel report
            ScatSwellWorksheetRawData.Cells[1, 2].Value = Data.ExtendedProperties["StationName"];
            ScatSwellWorksheetRawData.Cells[2, 2].Value = Data.Rows[0]["timeInterval"];
            ScatSwellWorksheetRawData.Cells[3, 2].Value = Data.Rows[Data.Rows.Count - 1]["timeInterval"];
            ScatSwellWorksheetRawData.Cells[1, 6].Value = Data.Rows.Count;
            ScatSwellWorksheetRawData.Cells[2, 6].Value = rowsMissing;
            ScatSwellWorksheetRawData.Cells[1, 11].Value = valuesImputed;
            // Format meta data for SCAT_swell Excel report
            ScatSwellWorksheetRawData.Cells[1, 1, 3, 11].Style.Font.Bold = true;
            ScatSwellWorksheetRawData.Cells[1, 1, 3, 11].Style.Font.Size = 11;

            // Add meta data to SCAT_vind Excel report
            ScatVindWorksheetRawData.Cells[1, 2].Value = Data.ExtendedProperties["StationName"];
            ScatVindWorksheetRawData.Cells[2, 2].Value = Data.Rows[0]["timeInterval"];
            ScatVindWorksheetRawData.Cells[3, 2].Value = Data.Rows[Data.Rows.Count - 1]["timeInterval"];
            ScatVindWorksheetRawData.Cells[1, 6].Value = Data.Rows.Count;
            ScatVindWorksheetRawData.Cells[2, 6].Value = rowsMissing;
            ScatVindWorksheetRawData.Cells[1, 11].Value = valuesImputed;
            // Format meta data for SCAT_vind Excel report
            ScatVindWorksheetRawData.Cells[1, 1, 3, 11].Style.Font.Bold = true;
            ScatVindWorksheetRawData.Cells[1, 1, 3, 11].Style.Font.Size = 11;

            int startRow = Config.ExcelStartRow;

            // Add meta data to Data worksheet in ScatWorksheet
            ScatTotalWorksheetData.Cells[2, 23].Formula = $"COUNT($V${startRow}:$V${Data.Rows.Count + startRow})";
            ScatTotalWorksheetData.Cells[3, 23].Formula = $"SUM(V{startRow}:V{Data.Rows.Count + startRow})";

            // Add meta data to Hs_dir worksheet in ScatWorksheet
            ScatTotalWorksheetHs_dir.Cells[3, 24].Formula = $"COUNT(B{startRow}:B{Data.Rows.Count + startRow})";
            ScatTotalWorksheetHs_dir.Cells[3, 26].Formula = $"COUNT(D{startRow}:D{Data.Rows.Count + startRow})";
            ScatTotalWorksheetHs_dir.Cells[4, 24].Formula = $"MIN(B{startRow}:B{Data.Rows.Count + startRow})";
            ScatTotalWorksheetHs_dir.Cells[4, 26].Formula = $"MIN(D{startRow}:D{Data.Rows.Count + startRow})";
            ScatTotalWorksheetHs_dir.Cells[5, 24].Formula = $"Average(B{startRow}:B{Data.Rows.Count + startRow})";
            ScatTotalWorksheetHs_dir.Cells[5, 26].Formula = $"Average(D{startRow}:D{Data.Rows.Count + startRow})";
            ScatTotalWorksheetHs_dir.Cells[6, 24].Formula = $"MAX(B{startRow}:B{Data.Rows.Count + startRow})";
            ScatTotalWorksheetHs_dir.Cells[6, 26].Formula = $"MAX(D{startRow}:D{Data.Rows.Count + startRow})";
            ScatTotalWorksheetHs_dir.Cells[6, 6].Value = $"{((DateTime)Data.Rows[0]["timeInterval"]).ToString("dd-MMM-yyyy")}";
            ScatTotalWorksheetHs_dir.Cells[6, 7].Value = $"{((DateTime)Data.Rows[Data.Rows.Count - 1]["timeInterval"]).ToString("dd-MMM-yyyy")}";

            // Add meta data to Hs_Tp worksheet in ScatWorksheet
            ScatTotalWorksheetHs_Tp.Cells[2, 27].Formula = $"+MIN(B{startRow}:B{Data.Rows.Count + startRow})";
            ScatTotalWorksheetHs_Tp.Cells[2, 28].Formula = $"+MIN(C{startRow}:C{Data.Rows.Count + startRow})";
            ScatTotalWorksheetHs_Tp.Cells[3, 27].Formula = $"+Average(B{startRow}:B{Data.Rows.Count + startRow})";
            ScatTotalWorksheetHs_Tp.Cells[3, 28].Formula = $"+Average(C{startRow}:C{Data.Rows.Count + startRow})";
            ScatTotalWorksheetHs_Tp.Cells[4, 27].Formula = $"+MAX(B{startRow}:B{Data.Rows.Count + startRow})";
            ScatTotalWorksheetHs_Tp.Cells[4, 28].Formula = $"+MAX(C{startRow}:C{Data.Rows.Count + startRow})";
        }


        /// <summary>
        /// Saves the Excel report files.
        /// </summary>
        /// <param name="excelFinishedFilePath">The pathname of the main Excel report.</param>
        /// <param name="scatTotalExcelFinishedFilePath">The pathname of the SCAT Excel report.</param>
        public void Save(string excelFinishedFilePath, string scatTotalExcelFinishedFilePath, string scatSwellExcelFinishedFilePath, string scatVindExcelFinishedFilePath)
        {
            ExcelTemplate.SaveAs(new FileInfo(excelFinishedFilePath));
            ScatTotalExcelTemplate.SaveAs(new FileInfo(scatTotalExcelFinishedFilePath));
            ScatSwellExcelTemplate.SaveAs(new FileInfo(scatSwellExcelFinishedFilePath));
            ScatVindExcelTemplate.SaveAs(new FileInfo(scatVindExcelFinishedFilePath));
        }

        /// <summary>
        /// Extends the formulas for each worksheet in the excel_template.xlsx and SCAT_excel_template.xlsx.
        /// </summary>
        public void ExtendFormulas(int numberOfRows)
        {

            int startRow = 10;
            int startRowScat = 190;

            int numberOfRowsScat = numberOfRows - (startRowScat - startRow);

            // Add additional rows to sheet 2 and copy formulas
            int startColumn = 1;
            int endColumnHs_Tp = 24;
            int endColumnHswell_Hwind = 24;
            int endColumnData = 22;
            int endColumnHs_dir = 17;
            int endColumnHswell_dir = 17;
            int endColumnU_dir = 17;
            int endColumnScatterPlots = 16;
            int endColumn3D = 8;
            int endColumn3D_max = 5;
            int endColRawData = 1;

            Dictionary<string, int> wsNameEndCol = new Dictionary<string, int>()
            {
                ["Hs_Tp"] = endColumnHs_Tp,
                ["Hswell_Hwind"] = endColumnHswell_Hwind,
                ["Data"] = endColumnData,
                ["Hs_dir"] = endColumnHs_dir,
                ["Hswell_dir"] = endColumnHswell_dir,
                ["U_dir"] = endColumnU_dir,
                ["ScatterPlots"] = endColumnScatterPlots,
                ["3D"] = endColumn3D,
                ["3D_max"] = endColumn3D_max,
                ["RawData"] = endColRawData

            };

            ScatTotalWorksheetRawData.InsertRow(startRowScat + 1, numberOfRowsScat, startRowScat);
            ScatTotalWorksheetData.InsertRow(startRowScat + 1, numberOfRowsScat, startRowScat);
            ScatTotalWorksheetHs_dir.InsertRow(startRowScat + 1, numberOfRowsScat, startRowScat);
            ScatTotalWorksheetHs_Tp.InsertRow(startRowScat + 1, numberOfRowsScat, startRowScat);
            ScatTotalWorksheet3D.InsertRow(startRowScat + 1, numberOfRowsScat, startRowScat);
            ScatTotalWorksheet3D_max.InsertRow(startRowScat + 1, numberOfRowsScat, startRowScat);

            ScatSwellWorksheetRawData.InsertRow(startRowScat + 1, numberOfRowsScat, startRowScat);
            ScatSwellWorksheetData.InsertRow(startRowScat + 1, numberOfRowsScat, startRowScat);
            ScatSwellWorksheetHswell_dir.InsertRow(startRowScat + 1, numberOfRowsScat, startRowScat);
            ScatSwellWorksheetHswell_Hwind.InsertRow(startRowScat + 1, numberOfRowsScat, startRowScat);

            ScatVindWorksheetRawData.InsertRow(startRowScat + 1, numberOfRowsScat, startRowScat);
            ScatVindWorksheetData.InsertRow(startRowScat + 1, numberOfRowsScat, startRowScat);
            ScatVindWorksheetU_dir.InsertRow(startRowScat + 1, numberOfRowsScat, startRowScat);

            WorksheetRawdata.InsertRow(startRow + 1, numberOfRows, startRow);
            WorksheetScatterPlots.InsertRow(startRow + 1, numberOfRows, startRow);

            // Extend formulas in WorksheetScatterPlots to row startRowScat      
            for (int row = startRow; row < startRowScat; ++row)
            {
                for (int col = startColumn; col <= endColumnScatterPlots; col++)
                {
                    WorksheetScatterPlots.Cells[row, col].FormulaR1C1 = WorksheetScatterPlots.Cells[row - 1, col].FormulaR1C1;
                }

                WorksheetRawdata.Cells[row, 1].FormulaR1C1 = WorksheetScatterPlots.Cells[row - 1, 1].FormulaR1C1;
            }

            Parallel.ForEach<ExcelWorksheet>(Worksheets, worksheet =>
            {
                for (int row = startRowScat; row < numberOfRowsScat + startRowScat; row++)
                {
                    for (int col = startColumn; col <= wsNameEndCol[worksheet.Name]; col++)
                    {
                        worksheet.Cells[row, col].FormulaR1C1 = worksheet.Cells[row - 1, col].FormulaR1C1;
                    }
                }
            });
        }

        /// <summary>
        /// Creates two scatter charts.
        /// </summary>
        /// <param name="graphTitle1">Title of chart 1</param>
        /// <param name="xAxisTitle1">xAxis title of chart 1</param>
        /// <param name="yAxisTitle1">yAxis title of chart 1</param>
        /// <param name="graphTitle2">Title of chart 2</param>
        /// <param name="xAxisTitle2">xAxis title of chart 2</param>
        /// <param name="yAxisTitle2">yAxis title of chart 2</param>
        /// <param name="startRow">The start row for data used in the charts</param>
        /// <param name="graphWidth">Width of the graph</param>
        /// <param name="graphHeight">Height of the graph</param>
        public void CreateScatterCharts(string graphTitle1, string xAxisTitle1, string yAxisTitle1, string graphTitle2, string xAxisTitle2, string yAxisTitle2, int startRow, int graphWidth, int graphHeight)
        {
            if (Data == null)
            {
                throw new InvalidOperationException("You must set the Data property to an instance of a DataTable first.");
            }

            // Create chart 1
            var scatterChart = ExcelTemplate.Workbook.Worksheets[1].Drawings.AddChart("scatterplot1", eChartType.XYScatter);
            var series1 = (ExcelScatterChartSerie)scatterChart.Series.Add($"ScatterPlots!$C${startRow}:$C${Data.Rows.Count + startRow - 1}", $"ScatterPlots!$B${startRow}:$B${Data.Rows.Count + startRow - 1}");
            series1.HeaderAddress = new ExcelAddressBase("ScatterPlots!$C$9");
            series1.MarkerSize = 4;
            scatterChart.SetSize(graphWidth, graphHeight);
            scatterChart.SetPosition(5, 0, 22, 0);
            scatterChart.PlotArea.Fill.Color = Color.FloralWhite;
            scatterChart.RoundedCorners = false;
            scatterChart.Title.Text = graphTitle1;

            scatterChart.XAxis.Title.Text = xAxisTitle1;
            scatterChart.XAxis.Title.Font.Size = 14;
            scatterChart.YAxis.Title.Text = yAxisTitle1;
            scatterChart.YAxis.Title.Font.Size = 14;

            // Create chart 2
            var scatterChart2 = ExcelTemplate.Workbook.Worksheets[1].Drawings.AddChart("scatterplot2", eChartType.XYScatter);
            for (char c = 'E'; c <= 'P'; c++)
            {
                var series2 = (ExcelScatterChartSerie)scatterChart2.Series.Add($"ScatterPlots!${c}${startRow}:${c}${Data.Rows.Count + startRow - 1}", $"ScatterPlots!$B${startRow}:$B${Data.Rows.Count + startRow - 1}");
                series2.HeaderAddress = new ExcelAddressBase($"ScatterPlots!${c}$9");
                series2.MarkerSize = 4;
            }
            scatterChart2.SetSize(graphWidth, graphHeight);
            scatterChart2.SetPosition(43, 0, 22, 0);
            scatterChart2.PlotArea.Fill.Color = Color.FloralWhite;
            scatterChart2.RoundedCorners = false;
            scatterChart2.Title.Text = graphTitle2;

            scatterChart2.XAxis.Title.Text = xAxisTitle2;
            scatterChart2.XAxis.Title.Font.Size = 14;
            scatterChart2.YAxis.Title.Text = yAxisTitle2;
            scatterChart2.YAxis.Title.Font.Size = 14;
        }

        /// <summary>
        /// Clears the loaded Excel documents and worksheets, so that it can be used for another data table.
        /// </summary>
        public void ClearExcelDocuments()
        {
            if (Data != null)
            {
                Console.WriteLine("Clearing Excel document objects");

                WorksheetRawdata.Cells[StartRow, StartColumn, Data.Rows.Count + StartRow - 1, StartColumn + Config.ExcelFeaturesOrder.Count - 1].Value = null;
                foreach (var chart in AddedCharts)
                {
                    WorksheetRawdata.Drawings.Remove(chart);
                }
                AddedCharts.Clear();

                // Remove scatterplots on worksheet 2
                ExcelTemplate.Workbook.Worksheets[1].Drawings.Remove("scatterplot1");
                ExcelTemplate.Workbook.Worksheets[1].Drawings.Remove("scatterplot2");

                ScatTotalWorksheetRawData.Cells[StartRow, StartColumn, Data.Rows.Count + StartRow - 1, StartColumn + Config.ExcelScatFeaturesOrder.Count - 1].Value = null;
            }
        }

        /// <summary>
        /// Disposes the underlying EPPlus ExcelPackages.
        /// </summary>
        public void Dispose()
        {
            if (ExcelTemplate != null)
            {
                ExcelTemplate.Dispose();
            }
            if (ScatTotalExcelTemplate != null)
            {
                ScatTotalExcelTemplate.Dispose();
            }
        }
    }
}