﻿using System;
using System.Collections.Generic;
using System.IO;
using System.Net;

namespace E39.FileDownloading
{
    public class FileGetter
    {
        // File to get: http://thredds.met.no/thredds/fileServer/obs/buoy-svv-e39/2017/01/201701_E39_G_Halsafjorden_wind.nc

        public List<string> SuccesfullyDownloadedURLs { get; }
        public List<string> FailedDownloadURLs { get; }
        private readonly Config config;

        public FileGetter(Config config)
        {
            this.config = config;
            SuccesfullyDownloadedURLs = new List<string>();
            FailedDownloadURLs = new List<string>();
        }

        /// <summary>
        /// Downloads a file from a given URL.
        /// </summary>
        /// <param name="url">URL where the file is stored.</param>
        /// <param name="destinationDirectory">URI of the local directory where the file is going to be saved.</param>
        /// <returns>true if successfull and false if not.</returns>
        public bool DownloadFile(string url, string destinationDirectory)
        {
            // Get the filname from the url
            string filename = GetFilename(url);

            if (filename == null)
            {
                throw new ArgumentException("The URL:" + url + " has the wrong format.", nameof(url));
            }

            using (var client = new WebClient())
            {
                // Check if there is internet
                if (System.Net.NetworkInformation.NetworkInterface.GetIsNetworkAvailable())
                {
                    try
                    {
                        client.DownloadFile(url, Path.Combine(destinationDirectory, filename));
                        SuccesfullyDownloadedURLs.Add(url);  // TODO: Test this part of the try catch.
                        return true;
                    }
                    catch (WebException e)
                    {
                        Console.WriteLine($"(FileGetter.cs) Could not download file: {url}. Exception: {e}");
                        FailedDownloadURLs.Add(url);
                    }
                }
                else
                {
                    Console.WriteLine("No internet connection found");
                }
            }

            return false;
        }

        /// <summary>
        /// Gets the name of a file given a URL.
        /// </summary>
        /// <param name="URL">The URL that you want to exctract the filename from.</param>
        /// <returns>The extraced filename if possible, null if not.</returns>
        public static string GetFilename(string URL)
        {
            try
            {
                Uri uri = new Uri(URL);
                return Path.GetFileName(uri.LocalPath);
            }
            catch (UriFormatException)
            {
                Console.WriteLine("The URL is empty or has an invalid format");
            }

            return null;
        }

        /// <summary>
        /// Download all files, given year, start month and end month
        /// </summary>
        /// <param name="startYear">What year the files belong to, format YYYY</param>
        /// <param name="startMonth">Starting month of time period for the data, format M if one digit and MM if two</param>
        /// <param name="endMonth">Ending month of time period for the data, format M if one digit and MM if two</param>
        public void DownloadFiles(int startYear, int startMonth, int endYear, int endMonth)
        {
            var startDateTime = new DateTime(startYear, startMonth, 1, 0, 0, 0, DateTimeKind.Utc);
            var endDateTime = new DateTime(endYear, endMonth, 1, 0, 0, 0, DateTimeKind.Utc);

            for (var i = startDateTime; startDateTime <= endDateTime; startDateTime = startDateTime.AddMonths(1))
            {
                string year = startDateTime.Year.ToString();
                string month = startDateTime.Month.ToString();

                if (month.Length < 2) month = "0" + month;

                foreach (string station in config.AllStationNames)
                {
                    foreach (string filetype in config.FileDataTypes) // A loop for each file ending
                    {
                        //DownloadFile(mainURL + year + "/" + month + "/" + year + month + station + fileEnding, CH.RawDataDirectory);
                        string rawDataFilename = $"{year}{month}_E39_{station}_{filetype}.nc";
                        DownloadFile($"{config.DownloadUrl}/{year}/{month}/{rawDataFilename}", config.RawDataDirectory);
                    }
                }
            }
        }
    }
}