﻿using System;
using System.Data;

namespace E39.Imputation
{
    /// <summary>
    /// Class with predicate functions for deciding whether a data point is considered empty.
    /// </summary>
    public static class DataPointValidator
    {
        /// <summary>
        /// A predicate method which checks if a data point has missing coordinates and should be considered empty.
        /// </summary>
        /// <remarks>The coordinates of a data point are its features "latitude" and "longitude", and is
        /// considered empty if both those features equals zero.</remarks>
        /// <param name="dataPoint">The data point to check for missing coordinates.</param>
        /// <returns>Whether the data point is missing coordinates.</returns>
        public static bool IsMissingCoordinates(DataRow dataPoint)
        {
            var lat = dataPoint["latitude"] as double? ?? 0;
            var lon = dataPoint["longitude"] as double? ?? 0;

            return (lat == 0 || double.IsNaN(lat))
                && (lon == 0 || double.IsNaN(lon));
        }

        /// <summary>
        /// Predicate yielding true for data points either where Hm0, thtp and tp features
        /// are all zero or DBNull, or time feature is zero or DBNull.
        /// </summary>
        /// <param name="dataPoint">The data point as a <see cref="DataRow"/>.</param>
        /// <returns>Whether the data point is considered empty or not.</returns>
        public static bool IsDataPointEmpty(DataRow dataPoint)
        {
            // DataRow is a class, therefore a nullable type, must therefore check
            if (dataPoint == null)
            {
                return true;
            }

            var dataPointTimeRaw = dataPoint["time"] as DateTime?;
            if (dataPointTimeRaw == null)
            {
                return true;
            }

            var dataPointTimeOffset = new DateTimeOffset(dataPointTimeRaw.Value, TimeSpan.Zero);
            if (dataPointTimeOffset.ToUnixTimeSeconds() == 0)
            {
                return true;
            }

            // Checks Hm0, thtp and tp features.
            // If these are all 0 or DBNull (which means zero anyway) or NaN, the data point is considered empty.
            var Hm0 = dataPoint["Hm0"] as double? ?? 0;  // The weirdest syntax
            var thtp = dataPoint["thtp"] as double? ?? 0;
            var tp = dataPoint["tp"] as double? ?? 0;

            return (Hm0 == 0 || double.IsNaN(Hm0))
                && (thtp == 0 || double.IsNaN(thtp))
                && (tp == 0 || double.IsNaN(tp));
        }
    }
}