﻿using System;
using System.Collections.Generic;
using System.Data;
using System.Linq;

namespace E39.Imputation
{
    /// <summary>
    /// kNN regression to use for imputing missing data points in stations' data sets.
    /// </summary>
    public class KNNRegression
    {
        public struct KNNParameters
        {
            public int K { get; }
            public int N { get; }

            public KNNParameters(int k, int n)
            {
                K = k;
                N = n;
            }
        }

        public DataSet StationsData { get; }
        public List<string> FeaturesToImpute { get; }
        public KNNParameters Params { get; }
        public Predicate<DataRow> IsDataPointEmpty { get; }

        public KNNRegression(DataSet stationsData, List<string> featuresToImpute, KNNParameters parameters)
            : this(stationsData, featuresToImpute, parameters, DataPointValidator.IsDataPointEmpty) { }

        public KNNRegression(DataSet stationsData, List<string> featuresToImpute, KNNParameters parameters, Predicate<DataRow> isDataPointEmpty)
        {
            StationsData = stationsData;
            FeaturesToImpute = featuresToImpute;
            Params = parameters;
            IsDataPointEmpty = isDataPointEmpty;
        }

        /// <summary>
        /// Prepares the stations data for kNN regression, by generating possibly missing coordinate values.
        /// </summary>
        /// <remarks>
        /// This method should always be called before calling other methods in this class.
        /// </remarks>
        public void PrepareStationsData()
        {
            foreach (DataTable stationData in StationsData.Tables)
            {
                Console.WriteLine("  Linearly interpolating missing coordinates for station " + stationData.ExtendedProperties["StationName"]);
                var coordinateInterpolator = new LinearInterpolation(
                    stationData,
                    new List<string> { "latitude", "longitude" },
                    DataPointValidator.IsMissingCoordinates
                );
                coordinateInterpolator.LinearInterpolateEmptyDataPoints();
            }
        }

        /// <summary>
        /// Finds distances from all empty data points to each its closest non-empty non-imputated data points.
        /// </summary>
        /// <remarks>
        /// For each empty data point, the method only checks a window of neighboring data points N time steps
        /// back and forth from the empty point. Distance and weight is then calculated for those neighbor data points
        /// which are non-empty and non-imputated within this window.
        /// </remarks>
        /// <param name="filterKNearest">Whether to filter and return only the top K nearest neighboring data points,
        /// or return all distances and weights to all neighbors. Default is to return only K nearest.</param>
        /// <returns>A <see cref="List{T}"/> with a <see cref="KNNDataPointNeighbors"/> object for each empty data point,
        /// containing distances and weights to its nearest neighbors.</returns>
        public List<KNNDataPointNeighbors> FindEmptyDataPointsAndNeighborWeights(bool filterKNearest = true)
        {
            // Make the kNN distance table
            var dataPointWeights = new List<KNNDataPointNeighbors>();

            // The highest number of data points.
            // TODO: Avoid index out of bounds for stations with less data points
            var maxStationRowCount = StationsData.Tables.Cast<DataTable>().Max(station => station.Rows.Count);

            // For all rows in all station datasets, vertical
            for (var row = 0; row < maxStationRowCount; ++row)
            {
                // For each station at a given timestamp, horizontal
                foreach (DataTable stationFrom in StationsData.Tables)
                {
                    var dataPointFrom = stationFrom.Rows[row];
                    if (IsDataPointEmpty(dataPointFrom) && !DataPointValidator.IsMissingCoordinates(dataPointFrom))
                    {
                        var neighbors = new KNNDataPointNeighbors(dataPointFrom);
                        dataPointWeights.Add(neighbors);

                        // Sliding window for kNN neighbors: n back and forth
                        var indexBehind = Math.Max(0, row - Params.N);
                        var indexForward = Math.Min(maxStationRowCount, row + Params.N);

                        // Find kNN distances from this empty data point to all other non-empty data points n back and forth in time
                        // For all stations, horizontal
                        foreach (DataTable stationTo in StationsData.Tables)
                        {
                            // For all datapoints n back and forth from the empty one, vertical
                            for (var i = indexBehind; i < indexForward && i < stationTo.Rows.Count; ++i)
                            {
                                var dataPointTo = stationTo.Rows[i];

                                // Calculate kNN distance only if datapointTo is non-empty, not itself imputed, and don't calculate distance to itself
                                // Theoretically, only IsDataPointEmpty-check is necessary, because all data points marked "Imputed" are already empty,
                                // and the current datapointFrom is also always empty.
                                if (!IsDataPointEmpty(dataPointTo) && !dataPointTo.Field<bool>("Imputed") && dataPointTo != dataPointFrom)
                                {
                                    var neighbor = new KNNDataPointNeighbors.KNNDistance(dataPointTo, 0, 0);
                                    double dist = Distance(dataPointFrom, dataPointTo, neighbor);

                                    // We want weight to be high when distance is low, and vice versa
                                    double weight = 1 / dist;

                                    neighbor.Distance = dist;
                                    neighbor.Weight = weight;
                                    neighbors.Neighbors.Add(neighbor);
                                }
                            }
                        }

                        // Should we return only the k nearest neighbors?
                        if (filterKNearest)
                        {
                            neighbors.Neighbors = neighbors.Neighbors
                                .OrderBy(neighbor => neighbor.Distance)
                                .Take(Params.K)
                                .ToList();
                        }
                    }
                }
            }

            return dataPointWeights;
        }

        /// <summary>
        /// Filters the list of neighbors in a <see cref="KNNDataPointNeighbors"/> object, and returns a new
        /// <see cref="KNNDataPointNeighbors"/> with only the top K nearest neighboring data points.
        /// </summary>
        /// <remarks>
        /// This method has no effect if the <see cref="KNNDataPointNeighbors"/> already contains K or less neighbors.
        /// For instance, this method is not necessary if <see cref="FindEmptyDataPointsAndNeighborWeights"/>
        /// was called with filterKNearest = true.
        /// </remarks>
        /// <param name="neighborDistances">The <see cref="KNNDataPointNeighbors"/> to filter K nearest neighbors for.</param>
        /// <returns>A new <see cref="KNNDataPointNeighbors"/> object containing only the K nearest neighbors.</returns>
        public KNNDataPointNeighbors FilterKNearestNeighbors(KNNDataPointNeighbors neighborDistances)
        {
            return new KNNDataPointNeighbors(neighborDistances.DataPoint)
            {
                Neighbors = neighborDistances.Neighbors
                    .OrderBy(neighbor => neighbor.Distance)
                    .Take(Params.K)
                    .ToList()
            };
        }

        /// <summary>
        /// Calculates the weighted average of neighbor data points and fills each data point with these values in-place.
        /// </summary>
        /// <remarks>This is the same as calling <see cref="CalculateAndSetWeightedAverage(KNNDataPointNeighbors)"/>
        /// on each element in the list.</remarks>
        /// <param name="emtpyDataPointNeighbors">A <see cref="List{T}"/> of <see cref="KNNDataPointNeighbors"/> for each
        /// empty data point, with its neighbor data points and their distances and weights.</param>
        public void ImputeEmptyDataPoints(List<KNNDataPointNeighbors> emtpyDataPointNeighbors)
        {
            foreach (var dataPointNeighbors in emtpyDataPointNeighbors)
            {
                // Calculate weighted average of neighbors and fill data point in-place
                CalculateAndSetWeightedAverage(dataPointNeighbors);
            }
        }

        /// <summary>
        /// Calculates the weighted average of neighbor data points and fills a data point with these values in-place.
        /// </summary>
        /// <remarks>Field "Imputed" is set to true for the imputed data point.</remarks>
        /// <param name="dataPointAndNeighbors">The empty data point, its neighbors and each its weight and distance.</param>
        public void CalculateAndSetWeightedAverage(KNNDataPointNeighbors dataPointAndNeighbors)
        {
            // Find weight sum
            double weightSum = dataPointAndNeighbors.Neighbors.Sum(neighbor => neighbor.Weight);
            dataPointAndNeighbors.WeightSum = weightSum;

            // For each neighbor (Z-axis), find the coefficient weight divided by weight sum
            foreach (var neighbor in dataPointAndNeighbors.Neighbors)
            {
                neighbor.Coefficient = neighbor.Weight / weightSum;
            }

            // For each feature to impute, horizontal
            foreach (var feature in FeaturesToImpute)
            {
                double featureValue = 0;

                // For each neighbor (Z-axis), take current feature multiplied by coefficient, add to features sum
                foreach (var neighbor in dataPointAndNeighbors.Neighbors)
                {
                    featureValue += neighbor.DataPointTo.Field<double>(feature) * (neighbor.Coefficient ?? 0);
                }

                // Set value of current feature on empty data point to the feature sum
                dataPointAndNeighbors.DataPoint[feature] = featureValue;
            }
            dataPointAndNeighbors.DataPoint["Imputed"] = true;
        }

        /// <summary>
        /// Finds the Euclidean distance between two data points' spatial location and time difference.
        /// </summary>
        /// <remarks>
        /// Basically, this method find the length of a 2d vector where one axis is time difference between
        /// data points, and the other axis is physical distance between data points.
        /// It uses the Haversine formula to find the spatial distance between data points, by considering their
        /// latitude and longitude. The difference in time in seconds is scaled down, such that the time difference
        /// of one time step (10 minutes) has the same magnitude as 200 meters in spatial distance.
        /// </remarks>
        /// <param name="dp1">The first data point</param>
        /// <param name="dp2">The second data point</param>
        /// <returns>The space- and time-distance between the data points.</returns>
        public static double Distance(DataRow dp1, DataRow dp2, KNNDataPointNeighbors.KNNDistance distances = null)
        {
            // Physical distance between stations when these data points were measured, in meters
            var spatialDist = HaversineDistance(dp1, dp2);

            double dp1Time = new DateTimeOffset((DateTime)dp1["timeInterval"], TimeSpan.Zero).ToUnixTimeSeconds();
            double dp2Time = new DateTimeOffset((DateTime)dp2["timeInterval"], TimeSpan.Zero).ToUnixTimeSeconds();
            // "Distance" in time between the data points, which is the difference in measurement timestamps in seconds
            double timeDist = Math.Abs(dp1Time - dp2Time);

            // Scale time distance such that a data point one step back or forth in time at the same spatial location
            // approximately weighs the same as a data point 200 meters away at the same timestamp.
            timeDist /= 3;

            if (distances != null)
            {
                distances.SpatialDist = spatialDist;
                distances.TimeDist = timeDist;
            }

            return Math.Sqrt(spatialDist * spatialDist + timeDist * timeDist); ;
        }

        /// <summary>
        /// Finds the great-circle distance between two data points in meters, using the Haversine formula.
        /// </summary>
        /// <remarks>This method only finds spatial distance between two points.
        /// It only considers latitude and longitude of two data points, and not any possible difference in time.</remarks>
        /// <param name="dp1">A data point.</param>
        /// <param name="dp2">Another data point.</param>
        /// <returns>The great-circle distance in meters between the data points.</returns>
        public static double HaversineDistance(DataRow dp1, DataRow dp2)
        {
            var piOver180 = Math.PI / 180.0D;
            var lat1rad = (dp1["latitude"] as double? ?? 0) * piOver180;  // Latitude (breddegrad) of point 1 in radians
            var lat2rad = (dp2["latitude"] as double? ?? 0) * piOver180;  // Latitude (breddegrad) of point 2 in radians
            var lon1rad = (dp1["longitude"] as double? ?? 0) * piOver180; // Longitude (lengdegrad) of point 1 in radians
            var lon2rad = (dp2["longitude"] as double? ?? 0) * piOver180; // Longitude (lengdegrad) of point 2 in radians

            // Distance between stations/data points in meters
            return 12742000 * Math.Asin(Math.Sqrt(
                Math.Pow(Math.Sin((lat1rad - lat2rad) / 2), 2)
                + Math.Cos(lat1rad) * Math.Cos(lat2rad) * Math.Pow(Math.Sin((lon1rad - lon2rad) / 2), 2)
            ));
        }
    }
}
