﻿using System;
using System.Collections.Generic;
using System.Data;

namespace E39.Imputation
{
    /// <summary>
    /// A class for generating and filling missing data points in a station's data set, by using linear interpolation of adjacent data points.
    /// </summary>
    public class LinearInterpolation
    {
        public DataTable StationData { get; }
        public List<string> Features { get; }
        public Predicate<DataRow> IsDataPointEmpty { get; }

        /// <summary>
        /// Creates a new linear interpolator for the given station data.
        /// </summary>
        /// <param name="stationData">The DataTable containing the station's measurement data, whose gaps of missing data points will be filled.</param>
        /// <param name="features">A List with names of the features to fill missing data for.</param>
        /// <param name="isDataPointEmpty">A predicate yielding true for data points considered empty.</param>
        public LinearInterpolation(DataTable stationData, List<string> features, Predicate<DataRow> isDataPointEmpty)
        {
            StationData = stationData;
            Features = features;
            IsDataPointEmpty = isDataPointEmpty;
        }

        /// <summary>
        /// Finds and returns the index of the first non-empty data point after the given index in a station's data.
        /// </summary>
        /// <remarks>
        /// If there are no non-empty data points after the given index, or if the index would yield a next
        /// index out of range, the method will return -1.
        /// </remarks>
        /// <param name="stationData">The station measurement data.</param>
        /// <param name="afterIndex">The index after which to find the next non-empty data point.</param>
        /// <param name="isEmptyPredicate">A predicate yielding true for data points considered empty.</param>
        /// <returns>The index of the first non-empty data point, or -1 if none exists.</returns>
        public int FindNextNonemptyDataPointIndex(int afterIndex)
        {
            for (int i = afterIndex + 1; i < StationData.Rows.Count; ++i)
            {
                if (!IsDataPointEmpty(StationData.Rows[i]))
                {
                    return i;
                }
            }

            return -1;
        }

        /// <summary>
        /// Fills gaps of missing data points in a station's data in-place, by linearly interpolating existing data points in the station data.
        /// </summary>
        /// <remarks>
        /// This method finds gaps in the station's data points and tries to fill them, by linearly interpolating from the
        /// previously known data point to the next known data point. The missing data points are filled in-place, meaning
        /// the supplied DataTable is modified. Field "Imputed" is set to true for each data point which was filled with
        /// interpolated data.
        /// Data points are checked against the supplied predicate, and are considered empty if the predicate yields true
        /// for those data points. If a station's first data point is missing, it tries to copy the first known data point
        /// backwards to the start. If a station's last data point is missing, it tries to copy the last known data point
        /// forwards to the end. If all data points are missing in the station data, the method can't do anything and returns
        /// the data unaltered.
        /// Only the features listed in the Features propery will be interpolated or copied for empty data points. It is assumed
        /// that the features to be interpolated are numeric values, as they are cast to double during calculation.
        /// </remarks>
        public void LinearInterpolateEmptyDataPoints()
        {
            for (int i = 0; i < StationData.Rows.Count; ++i)
            {
                DataRow row = StationData.Rows[i];

                // If this data point is empty
                if (IsDataPointEmpty(row))
                {
                    // Find next non-empty data point
                    int nextDataPointIndex = FindNextNonemptyDataPointIndex(i);

                    // Data begins with empty data point (case 2 and 4)
                    if (i == 0)
                    {
                        if (nextDataPointIndex < 0)
                        {
                            // CASE 4: All data points are empty, can't do anything
                            return;
                        }
                        else
                        {
                            // CASE 2: Copy first known data point backwards
                            DataRow firstKnownDataPoint = StationData.Rows[nextDataPointIndex];
                            for (; i < nextDataPointIndex; ++i)
                            {
                                DataRow currentRow = StationData.Rows[i];

                                // Copy the specified features
                                foreach (string featureName in Features)
                                {
                                    currentRow[featureName] = firstKnownDataPoint[featureName];
                                }
                                currentRow["Imputed"] = true;
                            }
                        }
                    }
                    else
                    {
                        DataRow lastKnownDataPoint = StationData.Rows[i - 1];

                        if (nextDataPointIndex < 0)
                        {
                            // CASE 3: Station data ends with missing data point, can't interpolate, must copy last known to end
                            for (; i < StationData.Rows.Count; ++i)
                            {
                                DataRow currentRow = StationData.Rows[i];

                                // Copy the specified features
                                foreach (string featureName in Features)
                                {
                                    currentRow[featureName] = lastKnownDataPoint[featureName];
                                }
                                currentRow["Imputed"] = true;
                            }
                        }
                        else
                        {
                            // CASE 1: Has a start and end index of missing data points, interpolate between these
                            DataRow nextKnownDataPoint = StationData.Rows[nextDataPointIndex];
                            int steps = nextDataPointIndex - i + 1;

                            // Linear interpolation of the specified features
                            foreach (string featureName in Features)
                            {
                                double lastKnownValue = (double)lastKnownDataPoint[featureName];
                                double nextKnownValue = (double)nextKnownDataPoint[featureName];
                                double slope = (nextKnownValue - lastKnownValue) / steps;

                                // Step though each missing data point of the current feature, calculate its linearly interpolated value and set it
                                for (int step = 1; step < steps; ++step)
                                {
                                    DataRow currentRow = StationData.Rows[i + step - 1];
                                    currentRow[featureName] = lastKnownValue + slope * step;
                                    currentRow["Imputed"] = true;
                                }
                            }

                            // Move row pointer to next known data point and continue from there
                            i = nextDataPointIndex;
                        }
                    }
                }
            }
        }
    }
}