﻿using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.IO;
using System.Threading;
using System.Threading.Tasks;

namespace E39.NetCDF
{
    /// <summary>
    /// Converter from NetCDF files to CSV files.
    /// </summary>
    /// <remarks>
    /// This class uses a Python script which does the conversion. This means
    /// that a Python interpreter needs to be installed on the user machine and available on
    /// the PATH. It also requires the Python package netCDF4 to be installed.
    /// </remarks>
    public class NetCDFConverter
    {
        private readonly string PythonInterpreterPath;
        private readonly string ConvertScriptPath;

        public NetCDFConverter(Config config)
        {
            PythonInterpreterPath = config.PythonExecutablePath;
            ConvertScriptPath = Resources.GetPathOfProgramFile("NetCDFConverter.py"); // Find the conversion script among the program files
            if (!File.Exists(ConvertScriptPath))
            {
                throw new FileNotFoundException("Python conversion script NetCDFConverter.py not found", ConvertScriptPath);
            }
        }

        /// <summary>
        /// Converts all variables in the root group of a NetCDF file to a csv file, using a Python conversion script.
        /// </summary>
        /// <remarks>
        /// This method calls a Python script which does the conversion. This means
        /// that a Python interpreter needs to be installed on the user machine and available on
        /// the PATH. It also requires the Python package netCDF4 to be installed.
        /// </remarks>
        /// <param name="netCDFFilePath">The path of the NetCDF input file.</param>
        /// <param name="csvFilePath">The path of the csv output file.</param>
        /// <exception cref="FileNotFoundException">Thrown if the input NetCDF file couldn't be found,
        /// or if the directory in which to put the csv file doesn't exist.</exception>
        /// <exception cref="Exception">Thrown if some error occured in the Python script during conversion.</exception>
        public void ConvertFile(string netCDFFilePath, string csvFilePath)
        {
            ProcessStartInfo procInfo = new ProcessStartInfo(PythonInterpreterPath)
            {
                UseShellExecute = false
            };

            // Supply the command line arguments to the python interpreter, which is the script to execute and the input and output files to convert
            procInfo.ArgumentList.Add(ConvertScriptPath);
            procInfo.ArgumentList.Add(netCDFFilePath);
            procInfo.ArgumentList.Add(csvFilePath);

            // Start the python process, which runs the script
            using (var proc = Process.Start(procInfo))
            {
                proc.WaitForExit();

                // Exit code 99 is returned from the script and indicates that the NetCDF file or the output directory doesn't exist
                if (proc.ExitCode == 99)
                {
                    throw new FileNotFoundException($"Input NetCDF file {netCDFFilePath} not found, or " +
                        $"the output directory {Path.GetDirectoryName(csvFilePath)} doesn't exist");
                }
                // Other nonzero exit codes indicates errors
                else if (proc.ExitCode != 0)
                {
                    throw new Exception($"An error occured in the Python script during conversion (exit code {proc.ExitCode})");
                }
            }
        }

        /// <summary>
        /// Converts all variables in the root groups of multiple NetCDF files to csv files in parallel.
        /// </summary>
        /// <remarks>
        /// This method tries to run multiple conversions in parallel by calling <see cref="ConvertFile(string, string)"/> in parallel.
        /// </remarks>
        /// <seealso cref="ConvertFile(string, string)"/>
        /// <param name="netCDFFilePaths">A <see cref="List{T}"/> of paths to the NetCDF files to convert.</param>
        /// <param name="csvOutputDirectoryPath">A path to the directory in which to put the output csv files.
        /// The files are named the same as the input files, plus a trailing ".csv".</param>
        /// <exception cref="AggregateException">Thrown if exceptions occured in <see cref="ConvertFile(string, string)"/>.</exception>
        /// <seealso cref="ConvertFilesParallel(List{string}, string, CancellationToken)"/>
        public void ConvertFilesParallel(List<string> netCDFFilePaths, string csvOutputDirectoryPath)
        {
            ConvertFilesParallel(netCDFFilePaths, csvOutputDirectoryPath, new CancellationToken());
        }

        /// <summary>
        /// Converts all variables in the root groups of multiple NetCDF files to csv files in parallel.
        /// </summary>
        /// <remarks>
        /// This method tries to run multiple conversions in parallel by calling <see cref="ConvertFile(string, string)"/> in parallel.
        /// </remarks>
        /// <seealso cref="ConvertFile(string, string)"/>
        /// <param name="netCDFFilePaths">A <see cref="List{T}"/> of paths to the NetCDF files to convert.</param>
        /// <param name="csvOutputDirectoryPath">A path to the directory in which to put the output csv files.
        /// The files are named the same as the input files, plus a trailing ".csv".</param>
        /// <param name="cancellationToken">A CancellationToken signalling to stop converting the files.</param>
        /// <exception cref="AggregateException">Thrown if exceptions occured in <see cref="ConvertFile(string, string)"/>.</exception>
        public void ConvertFilesParallel(List<string> netCDFFilePaths, string csvOutputDirectoryPath, CancellationToken cancellationToken)
        {
            Parallel.ForEach(netCDFFilePaths, new ParallelOptions { MaxDegreeOfParallelism = 5, CancellationToken = cancellationToken }, (filePath) =>
            {
                ConvertFile(filePath, Path.Combine(csvOutputDirectoryPath, Path.GetFileName(filePath) + ".csv"));
            });
        }
    }
}