﻿using E39.DataLoading;
using E39.Excel;
using E39.FileDownloading;
using E39.Imputation;
using E39.NetCDF;
using System;
using System.Collections.Generic;
using System.Data;
using System.Globalization;
using System.IO;
using System.Threading;

namespace E39
{
    /// <summary>
    /// Contains methods for processing individual part of the E39 wave data pipeline.
    /// </summary>
    public class Pipeline
    {
        private readonly Config Config;

        /// <summary>
        /// The UTC start date of the period to process data for, inclusive.
        /// </summary>
        public DateTimeOffset StartDate { get; }

        /// <summary>
        /// The UTC end date of the period to process data for, inclusive.
        /// </summary>
        public DateTimeOffset EndDate { get; }

        /// <summary>
        /// The station name letters to output Excel reports for.
        /// </summary>
        public List<string> OutputStationNames { get; }

        /// <summary>
        /// A CancellationToken which can be used for signalling to abort a running pipeline operation.
        /// </summary>
        public CancellationToken AbortToken { get; }

        /// <summary>
        /// Constructs a new Pipeline, which can process individual parts of the E39 wave data pipeline.
        /// </summary>
        /// <param name="config">The Config object with general settings for the processing.</param>
        /// <param name="startDate">The inclusive start date of the period to process data for. Will be converted to UTC if its not already.</param>
        /// <param name="endDate">The inclusive end date of the period to process data for. Will be converted to UTC if its not already.</param>
        /// <param name="outputStationNames">The station name letters to output Excel reports for.</param>
        public Pipeline(Config config, DateTimeOffset startDate, DateTimeOffset endDate, List<string> outputStationNames)
            : this(config, startDate, endDate, outputStationNames, new CancellationToken()) { }

        /// <summary>
        /// Constructs a new Pipeline, which can process individual parts of the E39 wave data pipeline.
        /// </summary>
        /// <param name="config">The Config object with general settings for the processing.</param>
        /// <param name="startDate">The inclusive start date of the period to process data for. Will be converted to UTC if its not already.</param>
        /// <param name="endDate">The inclusive end date of the period to process data for. Will be converted to UTC if its not already.</param>
        /// <param name="outputStationNames">The station name letters to output Excel reports for.</param>
        /// <param name="abortToken">A CancellationToken signalling to abort a running pipeline operation at appropriate points.</param>
        public Pipeline(Config config, DateTimeOffset startDate, DateTimeOffset endDate, List<string> outputStationNames, CancellationToken abortToken)
        {
            Config = config;
            StartDate = startDate.ToUniversalTime();
            EndDate = endDate.ToUniversalTime();
            OutputStationNames = outputStationNames;
            AbortToken = abortToken;
        }

        /// <summary>
        /// Downloads raw data files from met.no for all stations within the date range.
        /// </summary>
        public void DownloadFiles()
        {
            var fileDownloader = new FileGetter(Config);
            fileDownloader.DownloadFiles(StartDate.Year, StartDate.Month, EndDate.Year, EndDate.Month);
        }

        /// <summary>
        /// Converts all raw data NetCDF files to csv in the directory given in <see cref="Config.RawDataDirectory"/>.
        /// </summary>
        /// <remarks>The csv files will be put in the same directory as the NetCDF files, with a .csv suffix.</remarks>
        public void ConvertNetCDFToCsv()
        {
            var ncConverter = new NetCDFConverter(Config);

            string[] rawNcFilePaths = Directory.GetFiles(Config.RawDataDirectory, "*.nc", SearchOption.TopDirectoryOnly);

            ncConverter.ConvertFilesParallel(new List<string>(rawNcFilePaths), Config.RawDataDirectory, AbortToken);
        }

        /// <summary>
        /// Loads raw data from CSV files into a <see cref="DataTable"/> for each station and returns a <see cref="DataSet"/> with all station data.
        /// </summary>
        /// <remarks>
        /// This method searches for csv files in the directory given in <see cref="Config.RawDataDirectory"/>, and only loads those files which has
        /// data for a period within the start and end date range. The station name and data time period is determined
        /// from the filenames, which is expected to be on the format [yearMonth]_E39_[stationLetter]_[stationFjord]_[dataType].nc.csv
        /// </remarks>
        /// <returns>A <see cref="DataSet"/> with station data for all stations.</returns>
        /// <exception cref="OperationCanceledException">If the pipeline's cancellation token has signaled to abort operations.</exception>
        public DataSet LoadFromCsv()
        {
            string[] convertedCsvFilePaths = Directory.GetFiles(Config.RawDataDirectory, "*.nc.csv", SearchOption.TopDirectoryOnly);

            // Dictionary mapping a station name to its DataLoader object
            Dictionary<string, DataLoader> stationDataLoaderDict = new Dictionary<string, DataLoader>();

            // Find the start of the first month and the end of the last month we need the data files for
            var dataFilePeriodStart = new DateTimeOffset(StartDate.Year, StartDate.Month, 1, 0, 0, 0, TimeSpan.Zero);
            var dataFilePeriodEnd = new DateTimeOffset(EndDate.Year, EndDate.Month, 1, 0, 0, 0, TimeSpan.Zero).AddMonths(1).AddMinutes(-10);

            // Loop through all converted csv files and load its data into the corresponding DataLoader
            foreach (string csvFilePath in convertedCsvFilePaths)
            {
                // Check if we've been aborted and should stop working
                AbortToken.ThrowIfCancellationRequested();

                string csvFileName = Path.GetFileName(csvFilePath).Split('.')[0];  // Strip the trailing .nc.csv
                // The CSV files should be named as follows: <yearMonth>_E39_<stationLetter>_<stationFjord>_<dataType>
                string[] csvFileNameParts = csvFileName.Split('_');

                if (csvFileNameParts.Length == 5)
                {
                    // Find start and end of period this file contains data for
                    string yearMonth = csvFileNameParts[0];  // The format for yearMonth is YYYYMM
                    int year = int.TryParse(yearMonth.Substring(0, 4), out year) ? year : 0;
                    int month = int.TryParse(yearMonth.Substring(4), out month) ? month : 0;
                    var dataFilePeriod = new DateTimeOffset(year, month, 1, 0, 0, 0, TimeSpan.Zero);

                    // The stationLetter in the filename
                    string stationName = csvFileNameParts[2];

                    // The dataType in the filename, either WIND or WAVE
                    string dataType = csvFileNameParts[4].ToLower();

                    // Only process this file if it has data for a period between StartDate and EndDate and is either WIND or WAVE data
                    if (dataFilePeriod >= dataFilePeriodStart && dataFilePeriod <= dataFilePeriodEnd
                        && Config.FileDataTypes.Contains(dataType))
                    {
                        DataLoader stationDataLoader;

                        if (!stationDataLoaderDict.ContainsKey(stationName))
                        {
                            // Create a DataLoader object for this station if we haven't yet
                            stationDataLoader = new CsvDataLoader(Config, stationName);
                            stationDataLoader.FillInTimeColumn(StartDate, EndDate);
                            stationDataLoaderDict.Add(stationName, stationDataLoader);
                        }
                        else
                        {
                            stationDataLoader = stationDataLoaderDict[stationName];
                        }

                        stationDataLoader.LoadData(csvFilePath);
                    }
                }
            }

            // Create a DataSet containing the DataTables for all stations
            var dataSet = new DataSet("StationsData");
            foreach (var stationMapping in stationDataLoaderDict)
            {
                dataSet.Tables.Add(stationMapping.Value.Data);
            }

            return dataSet;
        }

        /// <summary>
        /// Imputes missing data points in-place using linear interpolation of adjacent data points.
        /// </summary>
        /// <param name="stationsData">The <see cref="DataSet"/> with station data for all stations.</param>
        /// <seealso cref="LinearInterpolation"/>
        public void ImputeMissingDataPointsLinearInterpolation(DataSet stationsData)
        {
            foreach (DataTable station in stationsData.Tables)
            {
                var linInt = new LinearInterpolation(station, Config.FeaturesToImpute, DataPointValidator.IsDataPointEmpty);
                linInt.LinearInterpolateEmptyDataPoints();
            }

            // Since time feature isn't imputed, copy it from timeInterval
            var postProcessor = new PostProcessor(stationsData);
            postProcessor.CopyTimeFeatureFromTimeInterval();
        }

        /// <summary>
        /// Imputes missing data points in-place using kNN regression.
        /// </summary>
        /// <param name="stationsData">The <see cref="DataSet"/> with station data for all stations.</param>
        /// <seealso cref="KNNRegression"/>
        public void ImputeMissingDataPointsKNNRegression(DataSet stationsData)
        {
            var features = new List<string>(Config.FeaturesToImpute);
            // Remove coordinate features, as they are linearly interpolated in preprocessing step
            features.Remove("latitude");
            features.Remove("longitude");

            var knn = new KNNRegression(stationsData, features, new KNNRegression.KNNParameters(Config.KNNParameter_K, Config.KNNParameter_N));

            // First, calculate missing coordinates for all stations
            knn.PrepareStationsData();

            // Find all empty data points and calculate neighbor weights
            var emptyDataPoints = knn.FindEmptyDataPointsAndNeighborWeights();

            // TODO: remove this after testing and cross-validation?
            var debugStream = StreamWriter.Null;
            if (Config.Debug)
            {
                debugStream = new StreamWriter("knn_weights.txt");
            }

            foreach (var dataPointNeighbors in emptyDataPoints)
            {
                // Calculate weighted average of neighbors and fill data point in-place
                knn.CalculateAndSetWeightedAverage(dataPointNeighbors);
                debugStream.WriteLine(dataPointNeighbors);
            }

            debugStream.Close();

            // Since time feature isn't imputed, copy it from timeInterval
            var postProcessor = new PostProcessor(stationsData);
            postProcessor.CopyTimeFeatureFromTimeInterval();
        }

        /// <summary>
        /// Fills the excel report templates with station data and writes the finished reports to the output directory.
        /// </summary>
        /// <remarks>This method loads the excel templates, fills them with station data and saves them to a subdirectory
        /// under the output directory, for each station.</remarks>
        /// <param name="stationsData">The <see cref="DataSet"/> with station data for all stations.</param>
        public void WriteExcelReports(DataSet stationsData)
        {
            string dateRangeString = $"{StartDate.Year}-{StartDate.Month}_{EndDate.Year}-{EndDate.Month}";
            string createdDate = DateTime.UtcNow.ToString("dd-MM-yyyy_HH.mm.ss", CultureInfo.InvariantCulture);

            using (var excelWriter = new ExcelWriter(Config, StartDate, EndDate, Config.ExcelTemplatesDirectory, stationsData.Tables[0].Rows.Count))
            {
                // Loop through each DataTable and write its data to Excel files, if it is in OutputStationNames
                foreach (DataTable station in stationsData.Tables)
                {
                    string stationLetter = (string)station.ExtendedProperties["StationName"];

                    if (OutputStationNames.Contains(stationLetter))
                    {
                        excelWriter.MakeReports(station);

                        // Path for the output excel file
                        string excelFinishedFilePath = Path.Combine(
                                Config.ExcelOutputDirectory,
                                $"{stationLetter}_{dateRangeString}__{createdDate}.xlsx"
                            );

                        // Path for the output Scat_total excel file
                        string scatTotalExcelFinishedFilePath = Path.Combine(
                                Config.ExcelOutputDirectory,
                                $"SCAT_total_{stationLetter}_{dateRangeString}__{createdDate}.xlsx"
                            );

                        // Path for the output Scat_swell excel file
                        string ScatSwellExcelFinishedFilePath = Path.Combine(
                                Config.ExcelOutputDirectory,
                                $"SCAT_swell_{stationLetter}_{dateRangeString}__{createdDate}.xlsx"
                            );

                        // Path for the output Scat_vind excel file
                        string ScatVindExcelFinishedFilePath = Path.Combine(
                                Config.ExcelOutputDirectory,
                                $"SCAT_vind_{stationLetter}_{dateRangeString}__{createdDate}.xlsx"
                            );

                        excelWriter.Save(excelFinishedFilePath, scatTotalExcelFinishedFilePath, ScatSwellExcelFinishedFilePath, ScatVindExcelFinishedFilePath);
                    }
                }
            }
        }
    }
}