﻿using CommandLine;
using CommandLine.Text;
using System;
using System.Collections.Generic;
using System.Data;
using System.Globalization;
using System.IO;
using System.Linq;

namespace E39
{
    public class Program
    {
        public class Options
        {
            [Option('r', "raw-data-dir", HelpText = "Path to a directory where raw NetCDF data files " +
                "either is to be downloaded or already exists, or converted CSV files reside. If not " +
                "specified, the value RawDataDirectory from the config file is used.")]
            public string RawDataDir { get; set; }

            [Option('o', "output-dir", HelpText = "Path to a directory in which the generated Excel " +
                "reports will be placed. If not specified, the value ExcelOutputDirectory from the " +
                "config file is used.")]
            public string OutputDir { get; set; }

            [Option('t', "templates-dir", HelpText = "Path to a directory in which the Excel template files " +
                "reside. If not specified, the value ExcelTemplatesDirectory from the config file is used.")]
            public string TemplatesDir { get; set; }

            [Option('d', "download-data", SetName = "DownloadFiles",
                HelpText = "Download raw data files from MET into raw data directory, convert them to CSV " +
                "and load them. If neither this nor --convert is specified, it's assumed converted CSV files " +
                "already exists in the raw data directory.")]
            public bool Download { get; set; }

            [Option('c', "convert", SetName = "ConvertFiles",
                HelpText = "Convert NetCDF files in the raw data directory to CSV and load them. If neither " +
                "this nor --download-data is specified, it's assumed converted CSV files already exists in the " +
                "raw data directory.")]
            public bool Convert { get; set; }

            [Option('i', "impute", HelpText = "Impute missing data points using kNN regression. " +
                "If not specified, empty data points will be left untouched.")]
            public bool Impute { get; set; }

            [Option('x', "dry-run", HelpText = "Dry run. Don't output any excel reports. " +
                "Useful when debugging with --debug.")]
            public bool DryRun { get; set; }

            [Option('v', "debug", HelpText = "Debug. Be verbose and output debug stat files.")]
            public bool Debug { get; set; }

            [Value(0, MetaName = "Start date", Required = true,
                HelpText = "Specifies UTC+00:00 start of date range to create reports for, on format YYYY-MM-DD. " +
                "The time of day is assumed to be 00:00. The range is inclusive.")]
            public string StartDate { get; set; }

            [Value(1, MetaName = "End date", Required = true,
                HelpText = "Specifies UTC+00:00 end of date range to create reports for, on format YYYY-MM-DD. " +
                "The time of day is assumed to be 23:50. The range is inclusive.")]
            public string EndDate { get; set; }

            [Value(2, MetaName = "Output stations", Required = true,
                HelpText = "The station letters to generate Excel reports for, separated by commas.")]
            public string OutputStationLetters { get; set; }

            [Usage]
            public static IEnumerable<Example> Examples { get; } = new List<Example>
            {
                new Example(
                    "General usage",
                    new Options {
                        RawDataDir = "<raw-data-dir>",
                        OutputDir = "<output-reports-dir>",
                        TemplatesDir = "<excel-templates-dir>",
                        StartDate = "<start-date>",
                        EndDate = "<end-date>",
                        OutputStationLetters = "<stations>"
                    }
                ),
                new Example(
                    "\nExample: Generate reports for january 2017 to april 2017 for stations A, B and C, with already converted raw data",
                    new Options {
                        RawDataDir = @"D:\E39 wave data\Raw\",
                        OutputDir = @"D:\E39 wave data\Reports\",
                        TemplatesDir = @"D:\E39 wave data\Templates\",
                        StartDate = "2017-01-01",
                        EndDate = "2017-04-30",
                        OutputStationLetters = "A,B,C"
                    }
                ),
                new Example(
                    "\nExample: Generate reports for january 2017 to april 2017 for station A, and impute missing data points",
                    new Options {
                        RawDataDir = @"D:\E39 wave data\Raw\",
                        OutputDir = @"D:\E39 wave data\Reports\",
                        TemplatesDir = @"D:\E39 wave data\Templates\",
                        Impute = true,
                        StartDate = "2017-01-01",
                        EndDate = "2017-04-30",
                        OutputStationLetters = "A,B,C"
                    }
                ),
            };
        }

        public static void RunPipeline(Options options)
        {
            Config config = new ConfigHandler() { Debug = options.Debug };

            // Set config from CLI options if specified
            if (!string.IsNullOrWhiteSpace(options.RawDataDir)) config.RawDataDirectory = options.RawDataDir;
            if (!string.IsNullOrWhiteSpace(options.OutputDir)) config.ExcelOutputDirectory = options.OutputDir;
            if (!string.IsNullOrWhiteSpace(options.TemplatesDir)) config.ExcelTemplatesDirectory = options.TemplatesDir;

            Directory.CreateDirectory(config.RawDataDirectory);
            Directory.CreateDirectory(config.ExcelOutputDirectory);

            DateTimeOffset startDate = DateTimeOffset.ParseExact(options.StartDate, "yyyy-MM-dd", CultureInfo.InvariantCulture, DateTimeStyles.AssumeUniversal);
            DateTimeOffset endDate = DateTimeOffset.ParseExact(options.EndDate, "yyyy-MM-dd", CultureInfo.InvariantCulture, DateTimeStyles.AssumeUniversal);
            endDate = endDate.AddHours(23).AddMinutes(50);  // Time was initially 00:00

            if (options.DryRun)
            {
                Console.WriteLine("Dry running");
            }

            var stationLetters = options.OutputStationLetters.Split(',', ';').ToList().ConvertAll(s => s.ToUpper());

            Console.WriteLine("Using config: " + config);
            Console.WriteLine("Stations to generate reports for: " + string.Join(", ", stationLetters));
            Console.WriteLine($"Creating reports from {startDate} to {endDate}");

            Pipeline pipeline = new Pipeline(config, startDate, endDate, stationLetters);
            DataSet stationsData = null;

            if (options.Download)
            {
                Console.WriteLine("Downloading files");
                pipeline.DownloadFiles();
                Console.WriteLine("Converting NetCDF files to CSV");
                pipeline.ConvertNetCDFToCsv();
            }
            else if (options.Convert)
            {
                Console.WriteLine("Converting NetCDF files to CSV");
                pipeline.ConvertNetCDFToCsv();
            }

            Console.WriteLine("Loading data from CSV files");
            stationsData = pipeline.LoadFromCsv();


            if (options.Impute)
            {
                Console.WriteLine("Imputing missing data points using kNN regression");
                pipeline.ImputeMissingDataPointsKNNRegression(stationsData);
            }

            if (!options.DryRun)
            {
                Console.WriteLine("Writing Excel reports");
                Directory.CreateDirectory(config.ExcelOutputDirectory);
                pipeline.WriteExcelReports(stationsData);
            }
        }

        public static void Main(string[] args)
        {
            Parser.Default.ParseArguments<Options>(args).WithParsed(RunPipeline);
            Console.WriteLine("Hit enter to exit");
            Console.ReadLine();
        }

        /// <summary>
        /// Helper method to print info about a <see cref="DataTable"/>.
        /// </summary>
        /// <param name="dt">The <see cref="DataTable"/> to print.</param>
        public static void PrintDataTable(DataTable dt)
        {
            Console.WriteLine($"Table {dt.TableName}");
            Console.WriteLine("Primary key: " + string.Join(", ", from col in dt.PrimaryKey select col.ColumnName));
            var numRows = dt.Rows.Count;
            Console.WriteLine("Num rows: " + numRows);
            Console.WriteLine("Columns: " + string.Join(", ", from DataColumn col in dt.Columns select col.ColumnName));

            for (int i = 0; i < Math.Min(3, numRows); ++i)
            {
                Console.WriteLine($"[{i}]: " + string.Join(", ", dt.Rows[i].ItemArray));
            }

            if (numRows > 6) Console.WriteLine("...");

            for (int i = Math.Max(numRows - 3, 3); i < numRows; ++i)
            {
                Console.WriteLine($"[{i}]: " + string.Join(", ", dt.Rows[i].ItemArray));
            }

            Console.WriteLine();
        }
    }
}