﻿using NUnit.Framework;
using E39;
using System;
using System.IO;
using E39.DataLoading;

namespace E39_Tests.E39_Lib.DataLoading
{
    class CsvDataLoaderTests
    {
        [SetUp]
        public void Setup()
        {
        }

        [Test]
        public void CreateDataTable_Returns_DataTable_With_24_Columns()
        {
            var ch = new ConfigHandler();
            var dl = new CsvDataLoader(ch, "Test");

            var dataTable = dl.CreateDataTable();
            var actualNumberOfColumns = dataTable.Columns.Count;

            Assert.AreEqual(24, actualNumberOfColumns);
        }
        
        [Test]
        public void FillInTimeColumn_Adds_The_Correct_Start_Date()
        {
            var ch = new ConfigHandler();
            var dl = new CsvDataLoader(ch, "Test");

            dl.FillInTimeColumn(
                new DateTimeOffset(2017, 12, 1, 0, 0, 0, TimeSpan.Zero), 
                new DateTimeOffset(2018, 1, 1, 0, 0, 0, TimeSpan.Zero));

            DateTime actualDateTime = (DateTime)dl.Data.Rows[0]["timeInterval"];
            DateTime expectedDateTime = new DateTime(2017, 12, 1, 0, 0, 0);

            Assert.AreEqual(expectedDateTime, actualDateTime);
        }

        [Test]
        public void FillInTimeColumn_Adds_The_Correct_End_Date()
        {
            var ch = new ConfigHandler();
            var dl = new CsvDataLoader(ch, "Test");

            dl.FillInTimeColumn(
                new DateTimeOffset(2017, 12, 1, 0, 0, 0, TimeSpan.Zero),
                new DateTimeOffset(2018, 1, 1, 0, 0, 0, TimeSpan.Zero));

            DateTime actualDateTime = (DateTime)dl.Data.Rows[dl.Data.Rows.Count - 1]["timeInterval"];
            DateTime expectedDateTime = new DateTime(2018, 1, 1, 0, 0, 0);

            Assert.AreEqual(expectedDateTime, actualDateTime);
        }

        [Test]
        public void FillInTimeColumn_Has_Interval_Of_10_Minutes()
        {
            var ch = new ConfigHandler();
            var dl = new CsvDataLoader(ch, "Test");

            dl.FillInTimeColumn(
                new DateTimeOffset(2018, 1, 1, 0, 0, 0, TimeSpan.Zero), 
                new DateTimeOffset(2018, 1, 2, 0, 0, 0, TimeSpan.Zero));

            DateTime dt1 = (DateTime)dl.Data.Rows[0]["timeInterval"];            
            DateTime dt2 = (DateTime)dl.Data.Rows[1]["timeInterval"];
            double actualNumberOfMinutes = dt2.Subtract(dt1).TotalMinutes;
            double numberOfMinutesExpected = 10;

            Assert.AreEqual(numberOfMinutesExpected, actualNumberOfMinutes);
        }

        [Test]
        public void FillInTimeColumn_Throws_AgrumentException_When_Dato_Arguments_Are_Invalid()
        {
            var ch = new ConfigHandler();
            var dl = new CsvDataLoader(ch, "Test");

            var e1 = Assert.Catch(() => dl.FillInTimeColumn(new DateTimeOffset(2019, 1, 1, 0, 0, 0, TimeSpan.Zero), new DateTimeOffset(2018, 1, 1, 0, 0, 0, TimeSpan.Zero)));
            var e2 = Assert.Catch(() => dl.FillInTimeColumn(new DateTimeOffset(2018, 2, 1, 0, 0, 0, TimeSpan.Zero), new DateTimeOffset(2018, 1, 1, 0, 0, 0, TimeSpan.Zero)));
            var e3 = Assert.Catch(() => dl.FillInTimeColumn(new DateTimeOffset(1969, 1, 1, 0, 0, 0, TimeSpan.Zero), new DateTimeOffset(2018, 1, 1, 0, 0, 0, TimeSpan.Zero)));

            Assert.IsInstanceOf<ArgumentException>(e1);
            Assert.IsInstanceOf<ArgumentException>(e2);
            Assert.IsInstanceOf<ArgumentException>(e3);
        }

        [Test]
        public void LoadData_FileNotFoundException_When_Unknow_CSV_URI_Is_Given()
        {
            var ch = new ConfigHandler();
            var dl = new CsvDataLoader(ch, "Test");

            var e = Assert.Catch(() => dl.LoadData("asd.csv"));

            Assert.IsInstanceOf<FileNotFoundException>(e);
        }

        [Test]
        public void LoadData_Does_Not_Crash_When_CSV_File_Runs_Out_Of_Rows_Before_The_DataTable_In_The_ExcelWriter_Class()
        {
            string csvFileWithMissingDataPath = Resources.GetPathOfProgramFile("data", "csvFileWithMissingDataOnEnd.csv");
            var ch = new ConfigHandler();
            var dl = new CsvDataLoader(ch, "Test");
            var startDate = new DateTimeOffset(2018, 4, 1, 0, 0, 0, TimeSpan.Zero);
            var endDate = startDate.AddMonths(1).AddMinutes(-10);
            dl.FillInTimeColumn(startDate, endDate);
            bool exceptionThrown = false;

            try
            {
                dl.LoadData(csvFileWithMissingDataPath);
            } catch (NullReferenceException)
            {
                exceptionThrown = true;
            }

            Assert.False(exceptionThrown);
        }
    }
}
