﻿using NUnit.Framework;
using System.Data;
using E39.Imputation;
using System.Collections.Generic;
using System;
using E39_Tests;

namespace E39_Tests.E39_Lib.Imputation
{
    class KNNRegressionTests
    {
        private DataTable StationData;
        private DataSet Stations;
        private KNNRegression Regression;

        public DataTable CreateDataTableStructure()
        {
            var emptyDt = new DataTable();
            emptyDt.Columns.Add("timeInterval", typeof(DateTime));
            emptyDt.Columns.Add("latitude", typeof(double));
            emptyDt.Columns.Add("longitude", typeof(double));
            emptyDt.Columns.Add("otherFeature1", typeof(double));
            emptyDt.Columns.Add(new DataColumn("Imputed", typeof(bool)) { DefaultValue = false });
            return emptyDt;
        }

        [SetUp]
        public void SetUp_CreateDataTableStructure()
        {
            StationData = CreateDataTableStructure();
            Stations = new DataSet("StationsData");
            Stations.Tables.Add(StationData);
            var features = new List<string> { "jeje" };
            Regression = new KNNRegression(Stations, features, new KNNRegression.KNNParameters(k: 3, n: 5));
        }

        [Test]
        public void HaversineDistance_DistanceBetweenTwoPoints_ReturnsCorrectDistance()
        {
            var table = CreateDataTableStructure();
            var dp1 = table.Rows.Add(new DateTime(2017, 1, 1, 12, 0, 0, DateTimeKind.Utc), 63.0, 10.0, 0);  // N63.0 E10.0
            var dp2 = table.Rows.Add(new DateTime(2017, 1, 1, 12, 0, 0, DateTimeKind.Utc), 63.1, 10.1, 0);  // N63.1 E10.1

            var expectedDistance = 12208.18;  // Expected 12208.18 meters
            var calculatedDistance = KNNRegression.HaversineDistance(dp1, dp2);

            Assert.AreEqual(expectedDistance, calculatedDistance, 0.01);  // Allow an inaccuracy of +- 0.01 meters
        }

        [Test]
        public void HaversineDistance_DistanceBetweenTwoPoints2_ReturnsCorrectDistance()
        {
            var table = CreateDataTableStructure();
            var dp1 = table.Rows.Add(null, 63.0, 10.0, 0);  // N63.0 E10.0
            var dp2 = table.Rows.Add(null, 63.1, 10.1, 0);  // N63.1 E10.1

            var expectedDistance = 12208.18;  // Expected 12208.18 meters
            var calculatedDistance = KNNRegression.HaversineDistance(dp1, dp2);

            Assert.AreEqual(expectedDistance, calculatedDistance, 0.01);  // Allow an inaccuracy of +- 0.01 meters
        }

        [Test]
        public void PrepareStationData_HasMissingDataPoints_OnlyCoordinateFeaturesInterpolated()
        {
            StationData.Rows.Add(new DateTime(10, DateTimeKind.Utc), 63, 10, 0.3);
            StationData.Rows.Add(new DateTime(0, DateTimeKind.Utc),  0,  0, 0);  // Only coordinates of these should be filled
            StationData.Rows.Add(new DateTime(0, DateTimeKind.Utc),  0,  0, 0);
            StationData.Rows.Add(new DateTime(40, DateTimeKind.Utc), 66, 13, 0.24);

            var expectedData = CreateDataTableStructure();
            expectedData.Rows.Add(new DateTime(10, DateTimeKind.Utc), 63, 10, 0.3,  false);
            expectedData.Rows.Add(new DateTime(0, DateTimeKind.Utc),  64, 11, 0,    true);
            expectedData.Rows.Add(new DateTime(0, DateTimeKind.Utc),  65, 12, 0,    true);
            expectedData.Rows.Add(new DateTime(40, DateTimeKind.Utc), 66, 13, 0.24, false);

            Regression.PrepareStationsData();

            AssertHelpers.AssertDataTablesAreEqual(expectedData, StationData);
        }
    }
}
