﻿using NUnit.Framework;
using System.Data;
using E39.Imputation;
using System.Collections.Generic;
using System;
using E39_Tests;

namespace E39_Tests.E39_Lib.Imputation
{
    class LinearInterpolationTests
    {
        private DataTable dt;
        private List<string> features;
        private Predicate<DataRow> coordinatesEmpty;
        private LinearInterpolation linearInterpolation;

        public DataTable CreateDataTableStructure()
        {
            var emptyDt = new DataTable();
            emptyDt.Columns.Add("latitude", typeof(double));
            emptyDt.Columns.Add("longitude", typeof(double));
            emptyDt.Columns.Add(new DataColumn("Imputed", typeof(bool)) { DefaultValue = false });
            return emptyDt;
        }

        [SetUp]
        public void SetUp_CreateDataTableStructure()
        {
            dt = CreateDataTableStructure();
            features = new List<string> { "latitude", "longitude" };
            coordinatesEmpty = dataPoint => (double)dataPoint["latitude"] == 0 && (double)dataPoint["latitude"] == 0;
            linearInterpolation = new LinearInterpolation(dt, features, coordinatesEmpty);
        }

        [Test]
        public void FindNextNonemptyCoordinatesIndex_NoNonemptyCoordinates_ReturnsNegativeOne()
        {
            dt.Rows.Add(0, 0);
            dt.Rows.Add(0, 0);
            dt.Rows.Add(0, 0);

            // Find next nonempty coord after index 0 (there are none)
            var index = linearInterpolation.FindNextNonemptyDataPointIndex(0);

            // Should be -1 as there are no nonempty coordinates
            Assert.AreEqual(-1, index);
        }

        [Test]
        public void FindNextNonemptyCoordinatesIndex_HasCoordinateWithOneEmptyElement_ReturnsCorrectIndex()
        {
            dt.Rows.Add(0, 0);
            dt.Rows.Add(0, 0);
            dt.Rows.Add(1, 0);

            // Find next nonempty coord after index 0
            var index = linearInterpolation.FindNextNonemptyDataPointIndex(0);

            // Index should be 2
            Assert.AreEqual(2, index);
        }

        [Test]
        public void FindNextNonemptyCoordinatesIndex_AfterIndexOutsideNumberOfRows_ReturnsNegativeOne()
        {
            dt.Rows.Add(0, 0);
            dt.Rows.Add(0, 0);
            dt.Rows.Add(1, 0);

            // Find next nonempty coord after index 2 (none, as it's out of range)
            var index = linearInterpolation.FindNextNonemptyDataPointIndex(2);

            // Should return -1 as the next index would be out of range
            Assert.AreEqual(-1, index);
        }

        [Test]
        public void FindNextNonemptyCoordinatesIndex_HasNonemptyCoordinates_ReturnsCorrectIndexAfterSpecifiedIndex()
        {
            dt.Rows.Add(63.1234, 10.777);
            dt.Rows.Add(0, 0);
            dt.Rows.Add(0, 0);
            dt.Rows.Add(63.1234, 10.777);

            // Find next nonempty coord after index 0
            var index = linearInterpolation.FindNextNonemptyDataPointIndex(0);

            // Index should be 3
            Assert.AreEqual(3, index);
        }

        [Test]
        public void FindNextNonemptyCoordinatesIndex_HasOnlyNonemptyCoordinates_ReturnsFirstIndexAfter()
        {
            dt.Rows.Add(63.1234, 10.777);
            dt.Rows.Add(63.1234, 10.777);
            dt.Rows.Add(63.1234, 10.777);
            dt.Rows.Add(63.1234, 10.777);
            dt.Rows.Add(63.1234, 10.777);

            // Find next nonempty coord after index 0
            var index = linearInterpolation.FindNextNonemptyDataPointIndex(0);

            // Index should be 1 as its the first nonempty coordinate *after* index 0
            Assert.AreEqual(1, index);
        }

        [Test]
        public void LinearInterpolateMissingStationCoordinates_OneMissingCoordinate_CorrectlyInterpolated()
        {
            dt.Rows.Add(63, 10);
            dt.Rows.Add(0, 0);  // Needs interpolation
            dt.Rows.Add(64, 11);
            dt.Rows.Add(63, 10);

            var expectedDt = CreateDataTableStructure();
            expectedDt.Rows.Add(63, 10);
            expectedDt.Rows.Add(63.5, 10.5, true);  // INTERPOLATED
            expectedDt.Rows.Add(64, 11);
            expectedDt.Rows.Add(63, 10);

            linearInterpolation.LinearInterpolateEmptyDataPoints();

            AssertHelpers.AssertDataTablesAreEqual(expectedDt, dt);
        }

        [Test]
        public void LinearInterpolateMissingStationCoordinates_AllCoordinatesMissing_NoInterpolationDone()
        {
            dt.Rows.Add(0, 0);  // All coordinates are missing
            dt.Rows.Add(0, 0);
            dt.Rows.Add(0, 0);
            dt.Rows.Add(0, 0);

            var expectedDt = CreateDataTableStructure();
            expectedDt.Rows.Add(0, 0);  // No interpolation can be done
            expectedDt.Rows.Add(0, 0);
            expectedDt.Rows.Add(0, 0);
            expectedDt.Rows.Add(0, 0);

            linearInterpolation.LinearInterpolateEmptyDataPoints();

            AssertHelpers.AssertDataTablesAreEqual(expectedDt, dt);
        }

        [Test]
        public void LinearInterpolateMissingStationCoordinates_FirstAndLastCoordinatesMissing_CopiesCoordinates()
        {
            dt.Rows.Add(0, 0);  // Missing first and last coordinate
            dt.Rows.Add(63, 10);
            dt.Rows.Add(64, 11);
            dt.Rows.Add(0, 0);

            var expectedDt = CreateDataTableStructure();
            expectedDt.Rows.Add(63, 10, true);  // First and last coordinate copied from first and last known
            expectedDt.Rows.Add(63, 10);
            expectedDt.Rows.Add(64, 11);
            expectedDt.Rows.Add(64, 11, true);

            linearInterpolation.LinearInterpolateEmptyDataPoints();

            AssertHelpers.AssertDataTablesAreEqual(expectedDt, dt);
        }

        [Test]
        public void LinearInterpolateMissingStationCoordinates_CoordinatesMissingAtBeginning_CopiesFirstKnownCoordinate()
        {
            dt.Rows.Add(0, 0);  // Coordinates are missing at the beginning
            dt.Rows.Add(0, 0);
            dt.Rows.Add(0, 0);
            dt.Rows.Add(63, 10);

            var expectedDt = CreateDataTableStructure();
            expectedDt.Rows.Add(63, 10, true);  // Copy the first known coordinate backwards
            expectedDt.Rows.Add(63, 10, true);
            expectedDt.Rows.Add(63, 10, true);
            expectedDt.Rows.Add(63, 10);

            linearInterpolation.LinearInterpolateEmptyDataPoints();

            AssertHelpers.AssertDataTablesAreEqual(expectedDt, dt);
        }

        [Test]
        public void LinearInterpolateMissingStationCoordinates_CoordinatesMissingAtEnd_CopiesLastKnownCoordinate()
        {
            dt.Rows.Add(63, 10);
            dt.Rows.Add(64, 11);
            dt.Rows.Add(0, 0);  // Coordinates are missing at the end
            dt.Rows.Add(0, 0);

            var expectedDt = CreateDataTableStructure();
            expectedDt.Rows.Add(63, 10);
            expectedDt.Rows.Add(64, 11);
            expectedDt.Rows.Add(64, 11, true);  // Copy the last known coordinate
            expectedDt.Rows.Add(64, 11, true);

            linearInterpolation.LinearInterpolateEmptyDataPoints();

            AssertHelpers.AssertDataTablesAreEqual(expectedDt, dt);
        }

        [Test]
        public void LinearInterpolateMissingStationCoordinates_NoMissingCoordinates_NoInterpolationDone()
        {
            dt.Rows.Add(63, 10);
            dt.Rows.Add(64, 11);
            dt.Rows.Add(65, 12);  // No missing coordinates
            dt.Rows.Add(66, 13);

            var expectedDt = CreateDataTableStructure();
            expectedDt.Rows.Add(63, 10);
            expectedDt.Rows.Add(64, 11);
            expectedDt.Rows.Add(65, 12);  // No interpolation done
            expectedDt.Rows.Add(66, 13);

            linearInterpolation.LinearInterpolateEmptyDataPoints();

            AssertHelpers.AssertDataTablesAreEqual(expectedDt, dt);
        }

        [Test]
        public void LinearInterpolateMissingStationCoordinates_TwoMissingCoordinates_CorrectlyInterpolated()
        {
            dt.Rows.Add(63, 10);
            dt.Rows.Add(0, 0);
            dt.Rows.Add(0, 0);  // Two missing coordinates
            dt.Rows.Add(66, 13);

            var expectedDt = CreateDataTableStructure();
            expectedDt.Rows.Add(63, 10);
            expectedDt.Rows.Add(64, 11, true);
            expectedDt.Rows.Add(65, 12, true);  // The two missing coordinated are interpolated
            expectedDt.Rows.Add(66, 13);

            linearInterpolation.LinearInterpolateEmptyDataPoints();

            AssertHelpers.AssertDataTablesAreEqual(expectedDt, dt);
        }

        [Test]
        public void LinearInterpolateMissingStationCoordinates_TwoPresentCoordinates_CorrectlyInterpolated()
        {
            dt.Rows.Add(0, 0);
            dt.Rows.Add(0, 0);
            dt.Rows.Add(0, 0);
            dt.Rows.Add(63, 10);
            dt.Rows.Add(0, 0);
            dt.Rows.Add(65, 12);

            var expectedDt = CreateDataTableStructure();
            expectedDt.Rows.Add(63, 10, true);  // Copied
            expectedDt.Rows.Add(63, 10, true);
            expectedDt.Rows.Add(63, 10, true);
            expectedDt.Rows.Add(63, 10);
            expectedDt.Rows.Add(64, 11, true);  // Interpolated
            expectedDt.Rows.Add(65, 12);

            linearInterpolation.LinearInterpolateEmptyDataPoints();

            AssertHelpers.AssertDataTablesAreEqual(expectedDt, dt);
        }
    }
}
