﻿using NUnit.Framework;
using E39.NetCDF;
using System.Collections.Generic;
using System.IO;
using E39;

namespace E39_Tests.E39_Lib.NetCDF
{
    class NetCDFConverterTests_NonexistentFiles
    {
        private Config Config;

        [SetUp]
        public void SetUp()
        {
            Config = new ConfigHandler();
        }

        [Test]
        public void ConvertFile_NonexistentPythonInterpreter_ThrowsException()
        {
            // Arrange
            Config.PythonExecutablePath = "montypython.exe";
            var converter = new NetCDFConverter(Config);
            var nonexistentFilePath = "nonexistent_file.nc";
            var csvFilePath = "";

            // Act and assert
            Assert.That(() => converter.ConvertFile(nonexistentFilePath, csvFilePath),
                Throws.Exception.TypeOf<FileNotFoundException>().Or.TypeOf<System.ComponentModel.Win32Exception>());
        }

        [Test]
        public void ConvertFile_NonexistentFile_ThrowsFileNotFoundException()
        {
            // Arrange
            var converter = new NetCDFConverter(Config);
            var nonexistentFilePath = "nonexistent_file.nc";
            var csvFilePath = "";

            // Act
            var exception = Assert.Catch(() => converter.ConvertFile(nonexistentFilePath, csvFilePath));

            // Assert
            Assert.IsInstanceOf<FileNotFoundException>(exception);
        }

        [Test]
        public void ConvertFilesParallel_NonexistentFiles_ThrowsFileNotFoundException()
        {
            // Arrange
            var converter = new NetCDFConverter(Config);
            var nonexistentFilePaths = new List<string> { "nonexistent_file.nc" };
            var csvDirectoryPath = Path.GetTempPath();

            // Act
            var exception = Assert.Catch(() => converter.ConvertFilesParallel(nonexistentFilePaths, csvDirectoryPath));

            // Assert
            Assert.IsInstanceOf<FileNotFoundException>(exception.InnerException);
        }
    }

    class NetCDFConverterTests_ExistingFiles
    {
        private Config Config;
        private string TestDataNetCDFFilePath;
        private string ExpectedCsvFilePath;
        private string ExpectedCsvContent;
        private string ConvertedCsvTempDirPath;
        private string ConvertedCsvFilePath;

        [SetUp]
        public void SetUpTestDataFiles()
        {
            Config = new ConfigHandler();

            TestDataNetCDFFilePath = Resources.GetPathOfProgramFile("data", "test_data.nc");
            ExpectedCsvFilePath = Resources.GetPathOfProgramFile("data", "test_data.csv");
            ExpectedCsvContent = File.ReadAllText(ExpectedCsvFilePath);

            ConvertedCsvTempDirPath = Path.Combine(Path.GetTempPath(), "E39_Tests");
            Directory.CreateDirectory(ConvertedCsvTempDirPath); // Create a temp directory in which to put the converted csv file
            ConvertedCsvFilePath = Path.Combine(ConvertedCsvTempDirPath, "test_data.nc.csv");
        }

        [TearDown]
        public void RemoveTestDataFiles()
        {
            Directory.Delete(ConvertedCsvTempDirPath, true);
        }

        [Test]
        public void ConvertFile_ExistingFile_CreatesCsvWithRootGroupValues()
        {
            // Arrange
            var converter = new NetCDFConverter(Config);

            // Act
            converter.ConvertFile(TestDataNetCDFFilePath, ConvertedCsvFilePath);
            var outputCsvContent = File.ReadAllText(ConvertedCsvFilePath);

            // Assert
            Assert.AreEqual(ExpectedCsvContent, outputCsvContent);
        }

        [Test]
        public void ConvertFilesParallel_ExistingFile_CreatesCsvWithRootGroupValues()
        {
            // Arrange
            var converter = new NetCDFConverter(Config);

            // Act
            converter.ConvertFilesParallel(new List<string> { TestDataNetCDFFilePath }, ConvertedCsvTempDirPath);
            var outputCsvContent = File.ReadAllText(ConvertedCsvFilePath);

            // Assert
            Assert.AreEqual(ExpectedCsvContent, outputCsvContent);
        }
    }
}
