﻿using E39.WebApplication.Jobs;
using E39.WebApplication.Models;
using Microsoft.AspNetCore.Authorization;
using Microsoft.AspNetCore.Mvc;
using Microsoft.Extensions.Options;
using System;
using System.Diagnostics;
using System.IO;
using System.Linq;
using System.Security.Claims;

namespace E39.WebApplication.Controllers
{
    [Route("")]
    [Route("CreateReport")]
    [Authorize]
    public class CreateReportController : Controller
    {
        private readonly IJobManager JobManager;
        private readonly WebAppConfig Config;

        public CreateReportController(IJobManager jobManager, IOptionsMonitor<WebAppConfig> config)
        {
            JobManager = jobManager;
            Config = config.CurrentValue;
        }

        [HttpGet]
        public IActionResult CreateReport()
        {
            return View();
        }

        [HttpGet("station-names")]
        public IActionResult GetStationNames()
        {
            return Json(Config.AllStationNames);
        }

        [HttpPost("jobs")]
        [ValidateAntiForgeryToken]
        public IActionResult CreateReportPost([FromBody] JobProperties jobProperties)
        {
            // TODO: validate dates and station names?

            var userIdentity = User.Identity as ClaimsIdentity; // Azure AD V2 endpoint specific
            jobProperties.StartedBy = userIdentity?.Claims.FirstOrDefault(c => c.Type == "preferred_username")?.Value;

            var job = JobManager.StartNewJob(jobProperties);

            return Json(new { success = true, guid = job.Id });
        }

        [HttpGet("jobs/{jobId}")]
        public IActionResult GetJob(Guid jobId)
        {
            var job = JobManager.GetJob(jobId);
            if (job != null)
            {
                return Json(new { exists = true, state = job.State });
            }
            else
            {
                return Json(new { exists = false });
            }
        }

        [HttpPost("jobs/{jobId}/abort")]
        [ValidateAntiForgeryToken]
        public IActionResult AbortJob(Guid jobId)
        {
            if (JobManager.GetJob(jobId) != null)
            {
                JobManager.AbortJob(jobId);
                return Ok();
            }
            else
            {
                return NotFound();
            }
        }

        // TODO remove this
        [HttpGet("config")]
        public IActionResult GetConfig()
        {
            return Content($"Server time now: {DateTimeOffset.Now}\nTime zone: {TimeZoneInfo.Local}\n\n{Config.ToString()}");
        }

        [HttpGet("jobs/{jobId}/reports")]
        public IActionResult DownloadReports(Guid jobId)
        {
            var job = JobManager.GetJob(jobId);
            if (job != null && job.State == JobState.COMPLETED)
            {
                var zipFileName = Path.GetFileName(job.ZipFilePath);
                var fileStream = System.IO.File.OpenRead(job.ZipFilePath);
                job.ZipDownloaded = true;

                return File(fileStream, "application/zip", zipFileName);
            }
            else
            {
                return NotFound();
            }
        }

        [ResponseCache(Duration = 0, Location = ResponseCacheLocation.None, NoStore = true)]
        public IActionResult Error()
        {
            return View(new ErrorViewModel { RequestId = Activity.Current?.Id ?? HttpContext.TraceIdentifier });
        }
    }
}