﻿using E39.WebApplication.Data;
using E39.WebApplication.Models;
using Microsoft.AspNetCore.Authorization;
using Microsoft.AspNetCore.Mvc;
using Microsoft.EntityFrameworkCore;
using Microsoft.WindowsAzure.Storage.Blob;
using System;
using System.Diagnostics;
using System.Linq;
using System.Threading.Tasks;

namespace E39.WebApplication.Controllers
{
    [Route("StoredReports")]
    [Authorize]
    public class StoredReportsController : Controller
    {
        private readonly FileDbContext dbContext;
        private readonly BlobManager blobManager;

        public StoredReportsController(FileDbContext dbContext, BlobManager blobManager)
        {
            this.dbContext = dbContext;
            this.blobManager = blobManager;
        }

        [HttpGet]
        public async Task<IActionResult> StoredReports()
        {
            return View(await dbContext.Reports.OrderByDescending(report => report.DateCreated).ToListAsync());
        }

        [HttpGet("{reportFileName}")]
        public async Task<IActionResult> DownloadReport(string reportFileName)
        {
            if (await dbContext.Reports.AnyAsync(report => report.Filename == reportFileName))
            {
                var blob = blobManager
                    .GetCloudBlobContainer("reports")
                    .GetBlockBlobReference(reportFileName);

                if (await blob.ExistsAsync())
                {
                    var sasToken = blob.GetSharedAccessSignature(
                        new SharedAccessBlobPolicy
                        {
                            Permissions = SharedAccessBlobPermissions.Read,
                            SharedAccessStartTime = DateTimeOffset.UtcNow.AddMinutes(-5), // Counter clock skewness
                            SharedAccessExpiryTime = DateTimeOffset.UtcNow.AddMinutes(10), // Let the blob download url be valid for 10 minutes
                        },
                        new SharedAccessBlobHeaders
                        {
                            ContentDisposition = $"attachment; filename={reportFileName}",
                            ContentType = "application/vnd.openxmlformats-officedocument.spreadsheetml.sheet"
                        }
                    );
                    return Redirect($"{blob.Uri}{sasToken}");
                }
                else
                {
                    return NotFound();
                }
            }
            else
            {
                return NotFound();
            }
        }

        [HttpDelete("{reportFilename}")]
        [ValidateAntiForgeryToken]
        public async Task<IActionResult> DeleteReport(string reportFilename)
        {
            var report = await dbContext.Reports.FindAsync(reportFilename);
            if (report != null)
            {
                dbContext.Reports.Remove(report);
                await dbContext.SaveChangesAsync();

                await blobManager
                    .GetCloudBlobContainer("reports")
                    .GetBlockBlobReference(reportFilename)
                    .DeleteIfExistsAsync();

                return Ok();
            }
            else
            {
                return NotFound();
            }
        }

        [ResponseCache(Duration = 0, Location = ResponseCacheLocation.None, NoStore = true)]
        public IActionResult Error()
        {
            return View(new ErrorViewModel { RequestId = Activity.Current?.Id ?? HttpContext.TraceIdentifier });
        }
    }
}