﻿using E39.WebApplication.Data;
using E39.WebApplication.Models;
using Microsoft.AspNetCore.Authorization;
using Microsoft.AspNetCore.Http;
using Microsoft.AspNetCore.Mvc;
using Microsoft.EntityFrameworkCore;
using Microsoft.WindowsAzure.Storage.Blob;
using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.Linq;
using System.Security.Claims;
using System.Threading.Tasks;

namespace E39.WebApplication.Controllers
{
    [Route("RawDataFiles")]
    [Authorize]
    public class UploadFilesController : Controller
    {
        private readonly FileDbContext dbContext;
        private readonly BlobManager blobManager;

        public UploadFilesController(FileDbContext dbContext, BlobManager blobManager)
        {
            this.dbContext = dbContext;
            this.blobManager = blobManager;
        }

        [HttpGet]
        public async Task<IActionResult> UploadFiles()
        {
            return View(await dbContext.RawDataFiles.OrderByDescending(r => r.Filename).ToListAsync());
        }

        [HttpGet("{filename}")]
        public async Task<IActionResult> DownloadRawDataFile(string filename)
        {
            if (await dbContext.RawDataFiles.AnyAsync(rawDataFile => rawDataFile.Filename == filename))
            {
                var blob = blobManager
                    .GetCloudBlobContainer("raw-data")
                    .GetBlockBlobReference(filename);

                if (await blob.ExistsAsync())
                {
                    var sasToken = blob.GetSharedAccessSignature(
                        new SharedAccessBlobPolicy
                        {
                            Permissions = SharedAccessBlobPermissions.Read,
                            SharedAccessStartTime = DateTimeOffset.UtcNow.AddMinutes(-5), // Counter clock skewness
                            SharedAccessExpiryTime = DateTimeOffset.UtcNow.AddMinutes(10), // Let the blob download url be valid for 10 minutes
                        },
                        new SharedAccessBlobHeaders
                        {
                            ContentDisposition = $"attachment; filename={filename}",
                            ContentType = "application/x-netcdf4"
                        }
                    );
                    return Redirect($"{blob.Uri}{sasToken}");
                }
                else
                {
                    return NotFound();
                }
            }
            else
            {
                return NotFound();
            }
        }

        [HttpDelete("{filename}")]
        [ValidateAntiForgeryToken]
        public async Task<IActionResult> DeleteRawDataFile(string filename)
        {
            // Load the raw data file from DB, and its related reports if any
            var rawDataFile = await dbContext.RawDataFiles
                .Include(r => r.ReportRelations)
                .ThenInclude(r => r.Report)
                .FirstOrDefaultAsync(r => r.Filename == filename);

            if (rawDataFile != null)
            {
                // Set Unknown status for the reports generated from this raw data file
                foreach (var reportRelation in rawDataFile.ReportRelations)
                {
                    reportRelation.Report.Status = "Unknown";
                }

                dbContext.RawDataFiles.Remove(rawDataFile);
                await dbContext.SaveChangesAsync();

                await blobManager
                    .GetCloudBlobContainer("raw-data")
                    .GetBlockBlobReference(filename)
                    .DeleteIfExistsAsync();

                await blobManager
                    .GetCloudBlobContainer("csv-data")
                    .GetBlockBlobReference($"{filename}.csv")
                    .DeleteIfExistsAsync();

                return Ok();
            }
            else
            {
                return NotFound();
            }
        }

        [HttpPost]
        [ValidateAntiForgeryToken]
        public async Task<IActionResult> Post(List<IFormFile> files)
        {
            var timeNow = DateTimeOffset.UtcNow;

            foreach (var formFile in files)
            {
                if (formFile.Length > 0)
                {
                    // Upload raw data file to blobstorage
                    using (var fileStream = formFile.OpenReadStream())
                    {
                        await blobManager.GetCloudBlobContainer("raw-data")
                            .GetBlockBlobReference(formFile.FileName)
                            .UploadFromStreamAsync(fileStream);
                    }

                    // Remove the corresponding CSV file if existing, which forces conversion of the new raw data file when needed
                    await blobManager.GetCloudBlobContainer("csv-data")
                        .GetBlockBlobReference($"{formFile.FileName}.csv")
                        .DeleteIfExistsAsync();

                    var username = (User.Identity as ClaimsIdentity)?
                        .Claims.FirstOrDefault(c => c.Type == "preferred_username")?.Value;

                    var rawDataFile = await dbContext.RawDataFiles
                        .Include(r => r.ReportRelations)
                        .ThenInclude(r => r.Report)
                        .FirstOrDefaultAsync(r => r.Filename == formFile.FileName);

                    // Raw data file does not exist in DB, add new entry
                    if (rawDataFile == null)
                    {
                        dbContext.RawDataFiles.Add(new RawDataFile(formFile.FileName, timeNow, timeNow, username, manuallyUploaded: true));
                    }
                    // Raw data file does exist in DB, update the entry and related reports
                    else
                    {
                        rawDataFile.DateChanged = timeNow;
                        rawDataFile.DateCached = timeNow;
                        rawDataFile.UploadedBy = username;
                        rawDataFile.ManuallyUploaded = true;

                        // When overwriting an existing raw data file, set status to outdated for related reports
                        foreach (var reportRelation in rawDataFile.ReportRelations)
                        {
                            reportRelation.Report.Status = "Outdated";
                        }
                    }

                    await dbContext.SaveChangesAsync();
                }
            }

            return Ok();
        }

        [ResponseCache(Duration = 0, Location = ResponseCacheLocation.None, NoStore = true)]
        public IActionResult Error()
        {
            return View(new ErrorViewModel { RequestId = Activity.Current?.Id ?? HttpContext.TraceIdentifier });
        }
    }
}