﻿using Microsoft.WindowsAzure.Storage;
using Microsoft.WindowsAzure.Storage.Blob;
using System;
using System.Collections.Generic;
using System.IO;
using System.Threading;
using System.Threading.Tasks;

namespace E39.WebApplication.Data
{
    public class BlobManager
    {
        private readonly CloudStorageAccount StorageAccount;

        public BlobManager(string connectionString)
        {
            StorageAccount = CloudStorageAccount.Parse(connectionString);
        }

        public CloudBlobContainer GetCloudBlobContainer(string containerName)
        {
            CloudBlobClient blobClient = StorageAccount.CreateCloudBlobClient();
            CloudBlobContainer container = blobClient.GetContainerReference(containerName);

            container.CreateIfNotExistsAsync().Wait();

            return container;
        }

        public void UploadBlob(string blobName, string filePath, CloudBlobContainer container)
        {
            CloudBlockBlob blob = container.GetBlockBlobReference(blobName);
            using (var fileStream = System.IO.File.OpenRead(filePath))
            {
                blob.UploadFromStreamAsync(fileStream).Wait();
            }
        }

        public async Task UploadBlobAsync(string blobName, string filePath, CloudBlobContainer container)
        {
            CloudBlockBlob blob = container.GetBlockBlobReference(blobName);
            using (var fileStream = File.OpenRead(filePath))
            {
                await blob.UploadFromStreamAsync(fileStream);
            }
        }

        public void DownloadBlob(string blobName, string savePath, CloudBlobContainer container)
        {
            CloudBlockBlob blob = container.GetBlockBlobReference(blobName);
            using (var fileStream = System.IO.File.OpenWrite(savePath))
            {
                blob.DownloadToStreamAsync(fileStream).Wait();
            }
        }

        public async Task<bool> BlobExistsAsync(string blobName, CloudBlobContainer container)
        {
            return await container.GetBlockBlobReference(blobName).ExistsAsync();
        }

        public async Task DownloadBlobsAsync(List<string> blobNames, string savePath, CloudBlobContainer container)
        {
            // Define the BlobRequestOptions on the download, including disabling MD5 hash validation for this example, this improves the download speed.
            BlobRequestOptions options = new BlobRequestOptions
            {
                DisableContentMD5Validation = true,
                StoreBlobContentMD5 = false
            };

            try
            {
                List<Task> tasks = new List<Task>();
                int max_outstanding = 100;
                int completed_count = 0;

                // Create a new instance of the SemaphoreSlim class to define the number of threads to use in the application.
                SemaphoreSlim sem = new SemaphoreSlim(max_outstanding, max_outstanding);

                foreach (string blobName in blobNames)
                {
                    // Get the blob and add a task to download the blob asynchronously from the storage account.
                    CloudBlockBlob blockBlob = container.GetBlockBlobReference(blobName);

                    await sem.WaitAsync();
                    tasks.Add(blockBlob.DownloadToFileAsync(Path.Combine(savePath, blobName), FileMode.Create, null, options, null).ContinueWith((t) =>
                    {
                        sem.Release();
                        Interlocked.Increment(ref completed_count);
                    }));
                }

                // Creates an asynchronous task that completes when all the downloads complete.
                await Task.WhenAll(tasks);
            }
            catch (Exception e)
            {
                Console.WriteLine("\nError encountered during transfer: {0}", e.Message);
            }
        }
    }
}