﻿using E39.WebApplication.Jobs;
using Microsoft.AspNetCore.SignalR;
using Microsoft.Extensions.Hosting;
using System.Threading;
using System.Threading.Tasks;

namespace E39.WebApplication.Hubs
{
    /// <summary>
    /// Service which propagates job events dispatched from <see cref="IJobManager"/> to clients via the SignalR hub <see cref="JobHub"/>.
    /// </summary>
    public class JobHubEventPropagator : IHostedService
    {
        private readonly IHubContext<JobHub> JobHubContext;
        private readonly IJobManager JobManager;

        public JobHubEventPropagator(IHubContext<JobHub> jobHubContext, IJobManager jobManager)
        {
            JobHubContext = jobHubContext;
            JobManager = jobManager;
        }

        public Task StartAsync(CancellationToken cancellationToken)
        {
            JobManager.OnJobStateChanged += JobManager_OnJobStateChanged;
            JobManager.OnJobLogMessageAdded += JobManager_OnJobLogMessageAdded;
            JobManager.OnJobTimeFinishedEstimated += JobManager_OnJobTimeFinishedEstimated;
            JobManager.OnJobProgressChanged += JobManager_OnJobProgressChanged;

            return Task.CompletedTask;
        }

        public Task StopAsync(CancellationToken cancellationToken)
        {
            JobManager.OnJobStateChanged -= JobManager_OnJobStateChanged;
            JobManager.OnJobLogMessageAdded -= JobManager_OnJobLogMessageAdded;
            JobManager.OnJobTimeFinishedEstimated -= JobManager_OnJobTimeFinishedEstimated;
            JobManager.OnJobProgressChanged -= JobManager_OnJobProgressChanged;

            return Task.CompletedTask;
        }

        private async void JobManager_OnJobStateChanged(Job senderJob, JobStateChangedEventArgs args)
        {
            await JobHubContext.Clients.All.SendAsync("JobStateChanged", senderJob.Id, args);
        }

        private async void JobManager_OnJobLogMessageAdded(Job senderJob, JobMessageAddedEventArgs args)
        {
            await JobHubContext.Clients.All.SendAsync("JobMessageAdded", senderJob.Id, args);
        }

        private async void JobManager_OnJobTimeFinishedEstimated(Job senderJob, JobTimeFinishedEstimatedEventArgs args)
        {
            await JobHubContext.Clients.All.SendAsync("JobTimeFinishedEstimated", senderJob.Id, args);
        }

        private async void JobManager_OnJobProgressChanged(Job senderJob, JobProgressChangedEventArgs args)
        {
            await JobHubContext.Clients.All.SendAsync("JobProgressChanged", senderJob.Id, args);
        }
    }
}