﻿using System;
using System.Collections.Generic;

namespace E39.WebApplication.Jobs
{
    /// <summary>
    /// A manager for <see cref="Job"/>s, which can start new jobs and notify changes to running jobs.
    /// </summary>
    public interface IJobManager
    {
        /// <summary>
        /// Event raised when a job appends a message to its log.
        /// </summary>
        event JobEventHandler<JobMessageAddedEventArgs> OnJobLogMessageAdded;

        /// <summary>
        /// Event raised when a job's state changes.
        /// </summary>
        event JobEventHandler<JobStateChangedEventArgs> OnJobStateChanged;

        /// <summary>
        /// Event raised when a job has estimated the time it should be finished.
        /// </summary>
        event JobEventHandler<JobTimeFinishedEstimatedEventArgs> OnJobTimeFinishedEstimated;

        /// <summary>
        /// Event raised when a job's progress percentage changes.
        /// </summary>
        event JobEventHandler<JobProgressChangedEventArgs> OnJobProgressChanged;

        /// <summary>
        /// Creates and starts a new <see cref="Job"/> for running the pipeline.
        /// </summary>
        /// <param name="properties">A <see cref="JobProperties"/> instance specifying how to run the job.</param>
        /// <returns>The newly created <see cref="Job"/>.</returns>
        Job StartNewJob(JobProperties properties);

        /// <summary>
        /// Returns a <see cref="Job"/> with the given GUID.
        /// </summary>
        /// <param name="jobId">The GUID of the job.</param>
        /// <returns>A <see cref="Job"/>, or null if the JobManager has no Job with that GUID.</returns>
        Job GetJob(Guid jobId);

        /// <summary>
        /// Returns all <see cref="Job"/>s that the job manager keeps track of.
        /// </summary>
        /// <returns>An <see cref="IEnumerable{T}"/> with all jobs.</returns>
        IEnumerable<Job> GetJobs();

        /// <summary>
        /// Attempts to abort a running <see cref="Job"/>.
        /// </summary>
        /// <param name="jobId">The GUID of the job to abort.</param>
        void AbortJob(Guid jobId);

        /// <summary>
        /// Removes a finished job from the JobManager. The job to remove must be in either COMPLETED, FAILED or ABORTED state.
        /// </summary>
        /// <param name="jobId">The GUID of the job to remove.</param>
        /// <exception cref="InvalidOperationException">If the job was in a state in which it can't be removed.</exception>
        void RemoveJob(Guid jobId);
    }
}