﻿using E39.DataLoading;
using E39.Excel;
using E39.NetCDF;
using E39.WebApplication.Data;
using E39.WebApplication.Models;
using E39.WebApplication.Scheduler;
using log4net;
using Microsoft.EntityFrameworkCore;
using Microsoft.WindowsAzure.Storage.Blob;
using System;
using System.Collections.Generic;
using System.Data;
using System.Diagnostics;
using System.Globalization;
using System.IO;
using System.IO.Compression;
using System.Linq;
using System.Text;
using System.Threading;
using System.Threading.Tasks;

namespace E39.WebApplication.Jobs
{
    public enum JobState
    {
        CREATED = 0,
        RUNNING = 1,
        COMPLETED = 2,
        FAILED = 3,
        ABORTED = 4
    }

    public class Job
    {
        public Guid Id { get; }
        public JobProperties Properties { get; }
        public WebAppConfig PipelineConfig { get; }
        public Task Task { get; set; }
        public CancellationTokenSource AbortJobTokenSource { get; }
        public List<JobMessage> Log { get; }
        public DateTimeOffset TimeStarted { get; private set; }
        public DateTimeOffset TimeFinished { get; private set; }
        public string ZipFilePath { get; private set; }
        public bool ZipDownloaded { get; set; }

        private readonly FileDbContext FileDbContext;
        private readonly BlobManager BlobManager;
        private readonly CloudBlobContainer RawDataContainer;
        private readonly CloudBlobContainer CsvDataContainer;
        private readonly CloudBlobContainer ReportsContainer;
        private Dictionary<string, List<string>> ReportCsvFilesMapping;
        private readonly List<string> RawDataToDownloadAndConvertFromBlob = new List<string>();
        private readonly List<string> RawDataToDownloadAndConvertFromMet = new List<string>();
        private readonly List<string> CsvFilesToDownloadFromBlob = new List<string>();
        private readonly List<string> ManuallyUploadedFiles = new List<string>();
        private readonly string TempPath;
        private readonly List<string> ReportsCreated;
        private readonly ILog Logger = LogManager.GetLogger(typeof(Job));
        private Stopwatch Stopwatch = new Stopwatch();
        private long EstimatedTime;
        private double TimeUsed;
        private readonly int RawDataFilesPerStation;
        private readonly Dictionary<string, long> LoadTimes;
        private Dictionary<string, long> UpdatedLoadTimes;

        private JobState _state = JobState.CREATED;

        public JobState State
        {
            get => _state;

            private set
            {
                if (value != _state)
                {
                    OnStateChanged?.Invoke(this, new JobStateChangedEventArgs(_state, value));
                    _state = value;
                }
            }
        }

        private float _progress;

        public float Progress
        {
            get => _progress;

            private set
            {
                value = value > 100 ? 100 : value;
                if (Math.Abs(value - _progress) > 0.01)
                {
                    OnProgressChanged?.Invoke(this, new JobProgressChangedEventArgs(_progress, value));
                    _progress = value;
                }
            }
        }

        private DateTimeOffset _estimatedTimeFinished;

        public DateTimeOffset EstimatedTimeFinished
        {
            get => _estimatedTimeFinished;

            private set
            {
                if (value != _estimatedTimeFinished)
                {
                    _estimatedTimeFinished = value;
                    OnTimeFinishedEstimated?.Invoke(this, new JobTimeFinishedEstimatedEventArgs(_estimatedTimeFinished));
                }
            }
        }

        public event JobEventHandler<JobMessageAddedEventArgs> OnLogMessageAdded;

        public event JobEventHandler<JobStateChangedEventArgs> OnStateChanged;

        public event JobEventHandler<JobTimeFinishedEstimatedEventArgs> OnTimeFinishedEstimated;

        public event JobEventHandler<JobProgressChangedEventArgs> OnProgressChanged;

        public Job(Guid id, JobProperties properties, WebAppConfig pipelineConfig, CancellationTokenSource abortJobTokenSource, FileDbContext dbContext, BlobManager blobManager)
        {
            Id = id;
            Properties = properties;
            PipelineConfig = pipelineConfig;
            AbortJobTokenSource = abortJobTokenSource;
            FileDbContext = dbContext;
            BlobManager = blobManager;
            Log = new List<JobMessage>();
            TimeStarted = DateTimeOffset.UtcNow;
            ReportCsvFilesMapping = new Dictionary<string, List<string>>();
            ReportsCreated = new List<string>();
            LoadTimes = new Dictionary<string, long>();
            UpdatedLoadTimes = new Dictionary<string, long>();
            RawDataContainer = BlobManager.GetCloudBlobContainer("raw-data");
            CsvDataContainer = BlobManager.GetCloudBlobContainer("csv-data");
            ReportsContainer = BlobManager.GetCloudBlobContainer("reports");
            TempPath = Path.Combine(pipelineConfig.BaseTempDirectory, $"{Id}_temp");
            RawDataFilesPerStation = ((Properties.EndDate.Year - Properties.StartDate.Year) * 12) + Properties.EndDate.Month - Properties.StartDate.Month + 1;
        }

        private void AddMessage(string text, JobMessageStyle style = JobMessageStyle.NORMAL)
        {
            var message = new JobMessage(text, style);
            Log.Add(message);
            Debug.WriteLine($"Message from job {Id}: {text}");
            Logger.Debug($"{Id}: {text}");
            OnLogMessageAdded?.Invoke(this, new JobMessageAddedEventArgs(message));
        }

        private void StartStopWatch()
        {
            Stopwatch.Start();
        }

        private void StopStopWatch(string operation, int numberOfOperations = 1)
        {
            Stopwatch.Stop();
            long millis = Stopwatch.ElapsedMilliseconds;
            Stopwatch.Reset();

            if (numberOfOperations != 0)
            {
                if (UpdatedLoadTimes[operation] == 0)
                {
                    UpdatedLoadTimes[operation] = (millis / numberOfOperations);
                }
                else
                {
                    long oldAverage = UpdatedLoadTimes[operation];
                    UpdatedLoadTimes[operation] = (long)((millis / numberOfOperations) * 0.1 + oldAverage * 0.9);
                }
            }
            TimeUsed += LoadTimes[operation] * numberOfOperations;
            Progress += (float)LoadTimes[operation] * numberOfOperations / EstimatedTime * 100.0f;
            EstimatedTimeFinished = DateTimeOffset.UtcNow.AddMilliseconds(EstimatedTime - TimeUsed);
        }

        private DataSet LoadFromCsv(List<string> stationNamesToLoad, CancellationToken cancellationToken)
        {
            cancellationToken.ThrowIfCancellationRequested();
            DataSet dataTables = new DataSet();

            // Loop through all the station names
            foreach (string station in stationNamesToLoad)
            {
                cancellationToken.ThrowIfCancellationRequested();
                // Only display messages for the stations which the user selected, even though we may load other stations as well
                var displayThisStation = Properties.Stations.Contains(station);

                if (displayThisStation)
                {
                    AddMessage($"Loading data files for station {station}:", JobMessageStyle.EMPHASIS);
                }

                // Check if we've been aborted and should stop working
                cancellationToken.ThrowIfCancellationRequested();

                // Initiate dataloader for the current station, and fill the TimeColumn
                DataLoader dataLoader = new CsvDataLoader(PipelineConfig, station);
                dataLoader.FillInTimeColumn(Properties.StartDate, Properties.EndDate);

                foreach (string csvFile in ReportCsvFilesMapping[station])
                {
                    cancellationToken.ThrowIfCancellationRequested();

                    if (displayThisStation)
                    {
                        AddMessage($"Loading csv file: {csvFile}");
                    }
                    dataLoader.LoadData(Path.Combine(TempPath, csvFile));
                }

                if (displayThisStation)
                {
                    AddMessage($"Done loading files for station {station}:", JobMessageStyle.EMPHASIS);
                    AddMessage($"Number of columns: {dataLoader.Data.Columns.Count}");
                    AddMessage($"Number of rows: {dataLoader.Data.Rows.Count}");
                    AddMessage($"Missing rows: {dataLoader.CountEmptyRows()}");
                }

                dataTables.Tables.Add(dataLoader.Data);
            }

            return dataTables;
        }

        /// <summary>
        /// Fills the excel report templates with station data and writes the finished reports to the output directory.
        /// </summary>
        /// <remarks>This method loads the excel templates, fills them with station data and saves them to a subdirectory
        /// under the output directory, for each station.</remarks>
        /// <param name="stationsData">The <see cref="DataSet"/> with station data for all stations.</param>
        private async Task WriteExcelReports(DataSet stationsData, CancellationToken cancellationToken)
        {
            StartStopWatch();
            AddMessage("Extending Excel formulas (this may take a while)", JobMessageStyle.EMPHASIS);

            string dateRangeString = $"{Properties.StartDate.Year}-{Properties.StartDate.Month}_{Properties.EndDate.Year}-{Properties.EndDate.Month}";
            string createdDate = TimeStarted.ToString("dd-MM-yyyy_HH.mm.ss", CultureInfo.InvariantCulture);

            using (var excelWriter = new ExcelWriter(PipelineConfig, Properties.StartDate, Properties.EndDate, TempPath, stationsData.Tables[0].Rows.Count))
            {
                StopStopWatch("load_excel_templates", stationsData.Tables[0].Rows.Count / 100);

                // Loop through each DataTable and write its data to Excel files, if it is in OutputStationNames
                foreach (DataTable station in stationsData.Tables)
                {
                    cancellationToken.ThrowIfCancellationRequested();
                    string stationName = (string)station.ExtendedProperties["StationName"];

                    if (Properties.Stations.Contains(stationName))
                    {
                        AddMessage($"Creating excel report for station {stationName}", JobMessageStyle.EMPHASIS);

                        // Check if the DataTable was filled using raw data files. Skip it if not.
                        if (ReportCsvFilesMapping[stationName].Count <= 0)
                        {
                            AddMessage($"Could not create report for {stationName}. No available raw data files.", JobMessageStyle.WARNING);
                        }
                        else
                        {
                            StartStopWatch();

                            // Convenience method which fills both templates with data and makes the graphs
                            AddMessage("Filling data and making graphs");
                            excelWriter.MakeReports(station);

                            StopStopWatch("write_to_excel");

                            StartStopWatch();
                            AddMessage("Saving excel reports");
                            // Path for the output excel file
                            string excelFilename = $"{stationName}_{dateRangeString}__{createdDate}.xlsx";
                            string excelFinishedFilePath = Path.Combine(
                                TempPath,
                                excelFilename
                            );
                            ReportsCreated.Add(excelFinishedFilePath);

                            // Path for the output Scat_total Excel file
                            string scatTotalExcelFilename = $"SCAT_total_{stationName}_{dateRangeString}__{createdDate}.xlsx";
                            string scatTotalExcelFinishedFilePath = Path.Combine(
                                TempPath,
                                scatTotalExcelFilename
                            );
                            ReportsCreated.Add(scatTotalExcelFinishedFilePath);

                            // Path for the output Scat_swell Excel file
                            string scatSwellExcelFilename = $"SCAT_swell_{stationName}_{dateRangeString}__{createdDate}.xlsx";
                            string scatSwellExcelFinishedFilePath = Path.Combine(
                                TempPath,
                                scatSwellExcelFilename
                            );
                            ReportsCreated.Add(scatSwellExcelFinishedFilePath);

                            // Path for the output Scat_vind Excel file
                            string scatVindExcelFilename = $"SCAT_vind_{stationName}_{dateRangeString}__{createdDate}.xlsx";
                            string scatVindExcelFinishedFilePath = Path.Combine(
                                TempPath,
                                scatVindExcelFilename
                            );
                            ReportsCreated.Add(scatVindExcelFinishedFilePath);

                            // Saving reports in TempPath
                            excelWriter.Save(excelFinishedFilePath, scatTotalExcelFinishedFilePath, scatSwellExcelFinishedFilePath, scatVindExcelFinishedFilePath);

                            // Uploading reports to Azure
                            BlobManager.UploadBlob(excelFilename, excelFinishedFilePath, ReportsContainer);
                            BlobManager.UploadBlob(scatTotalExcelFilename, scatTotalExcelFinishedFilePath, ReportsContainer);
                            BlobManager.UploadBlob(scatSwellExcelFilename, scatSwellExcelFinishedFilePath, ReportsContainer);
                            BlobManager.UploadBlob(scatVindExcelFilename, scatVindExcelFinishedFilePath, ReportsContainer);

                            // Divided by how long it takes to save a excel file per raw data file
                            StopStopWatch("save_excel_files", RawDataFilesPerStation);

                            // Save excel file names in database
                            AddMessage($"Saving reports in database:");
                            AddMessage($"{excelFilename}");
                            AddMessage($"{scatTotalExcelFilename}");
                            AddMessage($"{scatSwellExcelFilename}");
                            AddMessage($"{scatVindExcelFilename}");
                            StartStopWatch();

                            // First add Excel report filenames to DB
                            FileDbContext.Reports.Add(new Report(excelFilename, TimeStarted, Properties.StartedBy, "Up to date"));
                            FileDbContext.Reports.Add(new Report(scatTotalExcelFilename, TimeStarted, Properties.StartedBy, "Up to date"));
                            FileDbContext.Reports.Add(new Report(scatSwellExcelFilename, TimeStarted, Properties.StartedBy, "Up to date"));
                            FileDbContext.Reports.Add(new Report(scatVindExcelFilename, TimeStarted, Properties.StartedBy, "Up to date"));
                            await FileDbContext.SaveChangesAsync();

                            // Then add relations from reports to raw data files
                            foreach (string csvFile in ReportCsvFilesMapping[stationName])
                            {
                                string rawDataFileName = csvFile.Substring(0, csvFile.Length - 4);

                                FileDbContext.ReportRawDataFiles.Add(new ReportRawDataFile(excelFilename, rawDataFileName));
                                FileDbContext.ReportRawDataFiles.Add(new ReportRawDataFile(scatTotalExcelFilename, rawDataFileName));
                                FileDbContext.ReportRawDataFiles.Add(new ReportRawDataFile(scatSwellExcelFilename, rawDataFileName));
                                FileDbContext.ReportRawDataFiles.Add(new ReportRawDataFile(scatVindExcelFilename, rawDataFileName));
                            }
                            await FileDbContext.SaveChangesAsync();

                            StopStopWatch("store_reports_in_db", RawDataFilesPerStation);
                        }
                    }
                }
            }
        }

        private async Task LocateAvailableFilesAsync(List<string> stations, CancellationToken cancellationToken)
        {
            foreach (string station in stations)
            {
                cancellationToken.ThrowIfCancellationRequested();

                // Find the start of the first month and the end of the last month we need the data files for
                var dataFilePeriodStart = new DateTimeOffset(Properties.StartDate.Year, Properties.StartDate.Month, 1, 0, 0, 0, TimeSpan.Zero);
                var dataFilePeriodEnd = new DateTimeOffset(Properties.EndDate.Year, Properties.EndDate.Month, 1, 0, 0, 0, TimeSpan.Zero).AddMonths(1).AddMinutes(-10);

                // List for keeping track over which csv files were used.
                List<string> csvFilesUsed = new List<string>();

                foreach (string filetype in PipelineConfig.FileDataTypes)
                {
                    dataFilePeriodStart = new DateTimeOffset(Properties.StartDate.Year, Properties.StartDate.Month, 1, 0, 0, 0, TimeSpan.Zero);

                    while (dataFilePeriodStart < dataFilePeriodEnd)
                    {
                        cancellationToken.ThrowIfCancellationRequested();

                        bool csvFileLocated = false;

                        // Get filename for csv file.
                        string year = dataFilePeriodStart.ToString("yyyy");
                        string month = dataFilePeriodStart.ToString("MM");

                        string csvFilename = $"{year}{month}_E39_{station}_{filetype}.nc.csv";
                        string rawDataFilename = $"{year}{month}_E39_{station}_{filetype}.nc";

                        // First, check if the file exists in the database
                        if (await FileDbContext.RawDataFiles.AnyAsync(file => file.Filename == rawDataFilename))
                        {
                            var file = await FileDbContext.RawDataFiles.FindAsync(rawDataFilename);
                            if (file.ManuallyUploaded)
                            {
                                ManuallyUploadedFiles.Add($"{file.Filename} uploaded by {file.UploadedBy}");
                            }

                            // Then, check if the file *really* exists in the blobstorage (even though it should if it's in DB)
                            if (await BlobManager.BlobExistsAsync(csvFilename, CsvDataContainer))
                            {
                                csvFileLocated = true;
                                CsvFilesToDownloadFromBlob.Add(csvFilename);
                            }
                            else if (await BlobManager.BlobExistsAsync(rawDataFilename, RawDataContainer))
                            {
                                csvFileLocated = true;
                                RawDataToDownloadAndConvertFromBlob.Add(rawDataFilename);
                            }
                        }

                        // File not found in either DB or blobstorage
                        if (!csvFileLocated)
                        {
                            string fileInfoUrl = $"{PipelineConfig.CacherFileInfoUrl}/{year}/{month}/{rawDataFilename}.das";

                            var DateTimeOrNullObject = FileCacher.GetDateModified(fileInfoUrl);
                            if (DateTimeOrNullObject == null)
                            {
                                AddMessage($"The file; {rawDataFilename} does not exist", JobMessageStyle.WARNING);
                            }
                            else
                            {
                                csvFileLocated = true;
                                RawDataToDownloadAndConvertFromMet.Add($"{year}/{month}/{rawDataFilename}");
                            }
                        }

                        // If the file finally has been found in any way, remember it
                        if (csvFileLocated)
                        {
                            csvFilesUsed.Add(csvFilename);
                        }

                        dataFilePeriodStart = dataFilePeriodStart.AddMonths(1);
                    }
                }

                ReportCsvFilesMapping.Add(station, csvFilesUsed);
            }

            var diff = Properties.EndDate - Properties.StartDate;
            var rows = diff.Days * 24 * 6 + diff.Hours * 6 + diff.Minutes / 6;

            EstimatedTime =
               LoadTimes["download_csv_file_from_azure"] * CsvFilesToDownloadFromBlob.Count +
               LoadTimes["download_NetCDF_file_from_azure"] * RawDataToDownloadAndConvertFromBlob.Count +
               LoadTimes["convert_NetCDF_files_in_parallel"] * RawDataToDownloadAndConvertFromBlob.Count +
               LoadTimes["upload_csv_file_to_azure"] * RawDataToDownloadAndConvertFromBlob.Count +
               LoadTimes["download_from_met_convert_and_upload_to_azure"] * RawDataToDownloadAndConvertFromMet.Count +
               LoadTimes["download_templates"] +
               LoadTimes["load_csv_file"] * (CsvFilesToDownloadFromBlob.Count + RawDataToDownloadAndConvertFromBlob.Count + RawDataToDownloadAndConvertFromMet.Count) +
               LoadTimes["load_excel_templates"] * rows / 100 + // Time it takes to extend 100 rows of excel formulas
               LoadTimes["write_to_excel"] * Properties.Stations.Count +
               LoadTimes["save_excel_files"] * Properties.Stations.Count * RawDataFilesPerStation +
               LoadTimes["store_reports_in_db"] * Properties.Stations.Count * RawDataFilesPerStation +
               LoadTimes["create_zip_file"] * Properties.Stations.Count * RawDataFilesPerStation;
            //LoadTimes["average_time_diff"] * Properties.Stations.Count * RawDataFilesPerStation;

            if (Properties.Impute)
            {
                EstimatedTime += LoadTimes["impute_datatable"] * Properties.Stations.Count * RawDataFilesPerStation;
            }
        }

        private async Task DownloadFilesAsync(CancellationToken cancellationToken)
        {
            cancellationToken.ThrowIfCancellationRequested();
            // Download CSV files from blob storage
            if (CsvFilesToDownloadFromBlob.Count > 0) AddMessage("Downloading available csv files", JobMessageStyle.EMPHASIS);
            StartStopWatch();
            await BlobManager.DownloadBlobsAsync(CsvFilesToDownloadFromBlob, TempPath, CsvDataContainer);
            StopStopWatch("download_csv_file_from_azure", CsvFilesToDownloadFromBlob.Count);

            cancellationToken.ThrowIfCancellationRequested();
            // Download NetCDF raw data files from blob storage
            if (RawDataToDownloadAndConvertFromBlob.Count > 0) AddMessage("Downloading available raw data files (NetCDF)", JobMessageStyle.EMPHASIS);
            StartStopWatch();
            await BlobManager.DownloadBlobsAsync(RawDataToDownloadAndConvertFromBlob, TempPath, RawDataContainer);
            StopStopWatch("download_NetCDF_file_from_azure", RawDataToDownloadAndConvertFromBlob.Count);

            cancellationToken.ThrowIfCancellationRequested();
            // Create new filenames with path for the downloaded NetCDF files
            StartStopWatch();
            for (var i = 0; i < RawDataToDownloadAndConvertFromBlob.Count; i++)
            {
                RawDataToDownloadAndConvertFromBlob[i] = Path.Combine(TempPath, RawDataToDownloadAndConvertFromBlob[i]);
            }

            cancellationToken.ThrowIfCancellationRequested();
            // Convert the new NetCDF files in parallell to CSV
            if (RawDataToDownloadAndConvertFromBlob.Count > 0) AddMessage("Converting raw data files to csv", JobMessageStyle.EMPHASIS);
            NetCDFConverter netCDFConverter = new NetCDFConverter(PipelineConfig);
            netCDFConverter.ConvertFilesParallel(RawDataToDownloadAndConvertFromBlob, TempPath, cancellationToken);
            StopStopWatch("convert_NetCDF_files_in_parallel", RawDataToDownloadAndConvertFromBlob.Count);

            cancellationToken.ThrowIfCancellationRequested();
            // Upload newly converted files to blob blobstorage
            if (RawDataToDownloadAndConvertFromBlob.Count > 0) AddMessage("Uploading new csv files to Azure", JobMessageStyle.EMPHASIS);
            StartStopWatch();
            foreach (string RawDataFile in RawDataToDownloadAndConvertFromBlob)
            {
                cancellationToken.ThrowIfCancellationRequested();
                string csvFileName = $"{Path.GetFileName(RawDataFile)}.csv";
                await BlobManager.UploadBlobAsync(csvFileName, $"{RawDataFile}.csv", CsvDataContainer);
            }
            StopStopWatch("upload_csv_file_to_azure", RawDataToDownloadAndConvertFromBlob.Count);

            // Download NetCDF files from MET, convert them to CSV, update the database, and upload both to Azure blob storage
            StartStopWatch();
            FileCacher fileCacher = new FileCacher(PipelineConfig, FileDbContext, BlobManager);
            if (RawDataToDownloadAndConvertFromMet.Count > 0) AddMessage("Downloading, converting, and uploading files to azure:", JobMessageStyle.EMPHASIS);
            foreach (string rawDataDownloadUrl in RawDataToDownloadAndConvertFromMet)
            {
                cancellationToken.ThrowIfCancellationRequested();
                AddMessage($"{Path.GetFileName(rawDataDownloadUrl)}");
                string downloadUrl = $"{PipelineConfig.DownloadUrl}/{rawDataDownloadUrl}";
                string fileInfoUrl = $"{PipelineConfig.CacherFileInfoUrl}/{rawDataDownloadUrl}.das";
                await fileCacher.DownloadFileAndUpdateDBAsync(downloadUrl, TempPath, (DateTime)FileCacher.GetDateModified(fileInfoUrl));
            }
            StopStopWatch("download_from_met_convert_and_upload_to_azure", RawDataToDownloadAndConvertFromMet.Count);
        }

        private void LoadLoadTimes()
        {
            var loadTimes = FileDbContext.LoadTimes.First();
            LoadTimes["download_csv_file_from_azure"] = loadTimes.DownloadCsvFileFromAzure;
            LoadTimes["download_NetCDF_file_from_azure"] = loadTimes.DownloadNetCDFFileFromAzure;
            LoadTimes["convert_NetCDF_files_in_parallel"] = loadTimes.ConvertNetCDFFilesInParallel;
            LoadTimes["upload_csv_file_to_azure"] = loadTimes.UploadCsvFileToAzure;
            LoadTimes["download_from_met_convert_and_upload_to_azure"] = loadTimes.DownloadFromMetConvertAndUploadToAzure;
            LoadTimes["download_templates"] = loadTimes.DownloadTemplates;
            LoadTimes["load_csv_file"] = loadTimes.LoadCsvFile;
            LoadTimes["impute_datatable"] = loadTimes.ImputeDatatable;
            LoadTimes["load_excel_templates"] = loadTimes.LoadExcelTemplates;
            LoadTimes["write_to_excel"] = loadTimes.WriteToExcel;
            LoadTimes["save_excel_files"] = loadTimes.SaveExcelFiles;
            LoadTimes["store_reports_in_db"] = loadTimes.StoreReportsInDb;
            LoadTimes["create_zip_file"] = loadTimes.CreateZipFile;
            LoadTimes["average_time_diff"] = loadTimes.AverageTimeDiff;
            UpdatedLoadTimes = new Dictionary<string, long>(LoadTimes);
        }

        private async Task UpdateLoadTimes()
        {
            var loadTimes = FileDbContext.LoadTimes.First();
            loadTimes.DownloadCsvFileFromAzure = UpdatedLoadTimes["download_csv_file_from_azure"];
            loadTimes.DownloadNetCDFFileFromAzure = UpdatedLoadTimes["download_NetCDF_file_from_azure"];
            loadTimes.ConvertNetCDFFilesInParallel = UpdatedLoadTimes["convert_NetCDF_files_in_parallel"];
            loadTimes.UploadCsvFileToAzure = UpdatedLoadTimes["upload_csv_file_to_azure"];
            loadTimes.DownloadFromMetConvertAndUploadToAzure = UpdatedLoadTimes["download_from_met_convert_and_upload_to_azure"];
            loadTimes.DownloadTemplates = UpdatedLoadTimes["download_templates"];
            loadTimes.LoadCsvFile = UpdatedLoadTimes["load_csv_file"];
            loadTimes.ImputeDatatable = UpdatedLoadTimes["impute_datatable"];
            loadTimes.LoadExcelTemplates = UpdatedLoadTimes["load_excel_templates"];
            loadTimes.WriteToExcel = UpdatedLoadTimes["write_to_excel"];
            loadTimes.SaveExcelFiles = UpdatedLoadTimes["save_excel_files"];
            loadTimes.StoreReportsInDb = UpdatedLoadTimes["store_reports_in_db"];
            loadTimes.CreateZipFile = UpdatedLoadTimes["create_zip_file"];
            loadTimes.AverageTimeDiff = UpdatedLoadTimes["average_time_diff"];
            await FileDbContext.SaveChangesAsync();
        }

        public async Task RunAsync(CancellationToken cancellationToken)
        {
            State = JobState.RUNNING;
            AddMessage("Starting Pipeline", JobMessageStyle.EMPHASIS);

            try
            {
                Directory.CreateDirectory(TempPath);

                // STAGE 0:
                // Locate necessary raw data files and estimate time for job
                LoadLoadTimes();

                // If we should impute, load data for all available stations. If not, only the ones for the selected reports are necessary
                var stationNamesToLoad = Properties.Impute ? PipelineConfig.AllStationNames : Properties.Stations;

                AddMessage("Searching for available files: ", JobMessageStyle.EMPHASIS);
                await LocateAvailableFilesAsync(stationNamesToLoad, cancellationToken);
                TimeSpan ts = TimeSpan.FromMilliseconds(EstimatedTime);
                EstimatedTimeFinished = DateTimeOffset.UtcNow + ts;

                cancellationToken.ThrowIfCancellationRequested();

                Stopwatch sw = new Stopwatch();
                sw.Start();


                // Give the user information about which NetCDF file was manually uploaded by a user, if any.
                AddMessage("Using manually uploaded NetCDF files:", JobMessageStyle.EMPHASIS);
                if (ManuallyUploadedFiles.Count > 0)
                {
                    foreach (string file in ManuallyUploadedFiles)
                    {
                        AddMessage(file);
                    }
                } else
                {
                    AddMessage("None");
                }

                // STAGE 1:
                // Load data from CSV files into DataTables and DataTables into DataSets.
                // Convert NetCDF files if the CSV version doesn't exist.
                // Download NetCDF file it doesn't exist, and convert it.
                await DownloadFilesAsync(cancellationToken);

                AddMessage("Downloading templates", JobMessageStyle.EMPHASIS);
                StartStopWatch();
                // Copy Excel report templates to this job's temp directory, to avoid file opening conflicts
                await BlobManager.DownloadBlobsAsync(new List<string> { "excel_template.xlsx",
                                                                        "SCAT_total_excel_template.xlsx",
                                                                        "SCAT_swell_excel_template.xlsx",
                                                                        "SCAT_vind_excel_template.xlsx"
                                                                      }, TempPath, BlobManager.GetCloudBlobContainer("templates"));

                StopStopWatch("download_templates");

                AddMessage("Loading raw data", JobMessageStyle.EMPHASIS);
                StartStopWatch();
                DataSet dataTables = LoadFromCsv(stationNamesToLoad, cancellationToken);
                StopStopWatch("load_csv_file", CsvFilesToDownloadFromBlob.Count + RawDataToDownloadAndConvertFromBlob.Count + RawDataToDownloadAndConvertFromMet.Count);

                cancellationToken.ThrowIfCancellationRequested();

                // STAGE 2:
                // If Impute=true, impute
                if (Properties.Impute)
                {
                    StartStopWatch();
                    AddMessage("Imputing missing data points", JobMessageStyle.EMPHASIS);
                    var pipeline = new Pipeline(PipelineConfig, Properties.StartDate, Properties.EndDate, Properties.Stations);
                    pipeline.ImputeMissingDataPointsKNNRegression(dataTables);
                    StopStopWatch("impute_datatable", Properties.Stations.Count * RawDataFilesPerStation);
                }

                cancellationToken.ThrowIfCancellationRequested();

                // STAGE 3:
                // Write the CSV file data into excel template and excel SCAT template.
                // Update relation database
                AddMessage("Writing Excel reports", JobMessageStyle.EMPHASIS);
                await WriteExcelReports(dataTables, cancellationToken);

                cancellationToken.ThrowIfCancellationRequested();

                // STAGE 4:
                // Make reports available for download in a .zip folder
                StartStopWatch();

                AddMessage("Zipping generated Excel reports for download", JobMessageStyle.EMPHASIS);
                string startDate = Properties.StartDate.ToString("dd-MM-yyyy", CultureInfo.InvariantCulture);
                string endDate = Properties.EndDate.ToString("dd-MM-yyyy", CultureInfo.InvariantCulture);
                string createdDate = TimeStarted.ToString("dd-MM-yyyy_HH.mm.ss", CultureInfo.InvariantCulture);

                string stationsPart = "";
                foreach (string stationName in Properties.Stations)
                {
                    stationsPart += stationName.Split('_')[0];
                }

                string zipSrcDirPath = Path.Combine($"{TempPath}_zipFile", $"{startDate}_{endDate}_{stationsPart}_{createdDate}");
                Directory.CreateDirectory(zipSrcDirPath);

                // Copy each generated report file into the temp zip directory
                foreach (var reportPath in ReportsCreated)
                {
                    File.Copy(reportPath, Path.Combine(zipSrcDirPath, Path.GetFileName(reportPath)));
                }

                var jobLog = new StringBuilder();
                foreach (var logMessage in Log)
                {
                    jobLog.AppendLine(logMessage.Text);
                }
                File.WriteAllText(Path.Combine(zipSrcDirPath, "log.txt"), jobLog.ToString());

                ZipFilePath = $"{zipSrcDirPath}.zip";
                ZipFile.CreateFromDirectory(zipSrcDirPath, ZipFilePath);
                Directory.Delete(zipSrcDirPath, true);

                // Update average time diff
                StopStopWatch("create_zip_file", Properties.Stations.Count * RawDataFilesPerStation);
                sw.Stop();
                long millis = sw.ElapsedMilliseconds;

                if (UpdatedLoadTimes["average_time_diff"] == 0)
                {
                    UpdatedLoadTimes["average_time_diff"] = ((millis - EstimatedTime) / Properties.Stations.Count * RawDataFilesPerStation);
                }
                else
                {
                    long oldAverage = UpdatedLoadTimes["average_time_diff"];
                    UpdatedLoadTimes["average_time_diff"] = (long)(((millis - EstimatedTime) / (Properties.Stations.Count * RawDataFilesPerStation)) * 0.1 + oldAverage * 0.9);
                }

                // STAGE 4.1:
                // Update load_times database
                await UpdateLoadTimes();

                State = JobState.COMPLETED;
            }
            // A OperationCanceledException indicates task cancellation was signaled, set state to ABORTED and rethrow exception
            catch (OperationCanceledException)
            {
                State = JobState.ABORTED;
                AddMessage("Job was cancelled", JobMessageStyle.WARNING);
                // Rethrow the OperationCanceledException to the running Task
                throw;
            }
            catch (Exception ex)
            {
                State = JobState.FAILED;
                Logger.Fatal($"{Id}: Exception occured while running the pipeline: {ex.Message}", ex);
                AddMessage($"Exception occured while running the pipeline: {ex.Message}", JobMessageStyle.ERROR);
            }
            finally
            {
                Directory.Delete(TempPath, true);
                TimeFinished = DateTimeOffset.UtcNow;
            }

            Progress = 100;
            AddMessage($"Job finished with state {State}");
        }
    }
}