﻿using E39.WebApplication.Data;
using log4net;
using Microsoft.Extensions.DependencyInjection;
using Microsoft.Extensions.Options;
using System;
using System.Collections.Generic;
using System.Threading;
using System.Threading.Tasks;

namespace E39.WebApplication.Jobs
{
    public class JobManager : IJobManager
    {
        private readonly Dictionary<Guid, Job> Jobs;
        private readonly IServiceScopeFactory ServiceScopeFactory;
        private readonly IOptionsMonitor<WebAppConfig> ConfigMonitor;
        private readonly ILog Logger = LogManager.GetLogger(typeof(JobManager));

        public event JobEventHandler<JobMessageAddedEventArgs> OnJobLogMessageAdded;

        public event JobEventHandler<JobStateChangedEventArgs> OnJobStateChanged;

        public event JobEventHandler<JobTimeFinishedEstimatedEventArgs> OnJobTimeFinishedEstimated;

        public event JobEventHandler<JobProgressChangedEventArgs> OnJobProgressChanged;

        public JobManager(IServiceScopeFactory serviceScopeFactory, IOptionsMonitor<WebAppConfig> configMonitor)
        {
            Jobs = new Dictionary<Guid, Job>();
            ServiceScopeFactory = serviceScopeFactory;
            ConfigMonitor = configMonitor;
        }

        public Job StartNewJob(JobProperties properties)
        {
            // Create a DI scope for the new job, so it can have its own scoped services
            var jobScope = ServiceScopeFactory.CreateScope();
            var jobDbContext = jobScope.ServiceProvider.GetRequiredService<FileDbContext>();
            var jobBlobManager = jobScope.ServiceProvider.GetRequiredService<BlobManager>();

            var id = Guid.NewGuid();
            var abortJobTokenSource = new CancellationTokenSource();
            var job = new Job(id, properties, ConfigMonitor.CurrentValue, abortJobTokenSource, jobDbContext, jobBlobManager);

            job.OnLogMessageAdded += (sender, args) => OnJobLogMessageAdded?.Invoke(sender, args);
            job.OnStateChanged += (sender, args) => OnJobStateChanged?.Invoke(sender, args);
            job.OnTimeFinishedEstimated += (sender, args) => OnJobTimeFinishedEstimated?.Invoke(sender, args);
            job.OnProgressChanged += (sender, args) => OnJobProgressChanged?.Invoke(sender, args);

            Jobs.Add(id, job);  // TODO: handle exception if there's a guid collision?

            job.Task = Task.Run(() => job.RunAsync(abortJobTokenSource.Token), abortJobTokenSource.Token);

            Logger.Info($"Started new job with GUID {id}");
            Logger.Debug($"Job properties: " +
                $"StartDate: {properties.StartDate}, " +
                $"EndDate: {properties.EndDate}, " +
                $"Impute: {properties.Impute}, " +
                $"Stations: [{string.Join(", ", properties.Stations)}], " +
                $"StartedBy: {properties.StartedBy}");
            return job;
        }

        public Job GetJob(Guid jobId)
        {
            return Jobs.GetValueOrDefault(jobId, null);
        }

        public IEnumerable<Job> GetJobs()
        {
            return Jobs.Values;
        }

        public void AbortJob(Guid jobId)
        {
            if (Jobs.ContainsKey(jobId))
            {
                try
                {
                    Logger.Info($"Aborting job {jobId}");
                    Jobs[jobId].AbortJobTokenSource.Cancel();
                }
                catch (Exception ex)
                {
                    Logger.Warn($"Aborting of job {jobId} failed", ex);
                }
            }
        }

        public void RemoveJob(Guid jobId)
        {
            if (Jobs.ContainsKey(jobId))
            {
                var job = Jobs[jobId];
                if (job.State == JobState.COMPLETED || job.State == JobState.FAILED || job.State == JobState.ABORTED)
                {
                    Logger.Info($"Removing job {jobId} with state {job.State}");
                    Jobs.Remove(jobId);
                }
                else
                {
                    throw new InvalidOperationException("Job is not finished");
                }
            }
        }
    }
}