﻿using E39.FileDownloading;
using E39.NetCDF;
using E39.WebApplication.Data;
using E39.WebApplication.Models;
using log4net;
using Microsoft.WindowsAzure.Storage.Blob;
using System;
using System.IO;
using System.Net;
using System.Threading.Tasks;

namespace E39.WebApplication.Scheduler
{
    public class FileCacher
    {
        private readonly WebAppConfig config;
        private readonly ILog log = LogManager.GetLogger(typeof(FileCacher));
        private int filesDownloaded = 0;
        private int filesUpdated = 0;
        private int newFiles = 0;
        private string filesNotAvailableForDownload = "";
        private readonly FileDbContext FileDbContext;
        private readonly BlobManager BlobManager;
        private readonly CloudBlobContainer RawDataContainer;
        private readonly CloudBlobContainer CsvDataContainer;

        public FileCacher(WebAppConfig config, FileDbContext dbContext, BlobManager blobManager)
        {
            this.config = config;
            FileDbContext = dbContext;
            BlobManager = blobManager;
            RawDataContainer = BlobManager.GetCloudBlobContainer("raw-data");
            CsvDataContainer = BlobManager.GetCloudBlobContainer("csv-data");
        }

        /// <summary>
        /// Uses the the Response Header Last-Modified when asking for the thredds.met.no file with
        /// the .das ending. This gives the last modified date of the file.
        /// </summary>
        /// <param name="url">The url to the .das file.</param>
        /// <returns>Returns the date modified as a DateTime object if succesfull; null if not.</returns>
        public static DateTime? GetDateModified(string url)
        {
            try
            {
                DateTime lastModified;
                Uri myUri = new Uri(url);
                // Creates an HttpWebRequest for the specified URL.

                HttpWebRequest myHttpWebRequest = (HttpWebRequest)WebRequest.Create(myUri);
                HttpWebResponse myHttpWebResponse = (HttpWebResponse)myHttpWebRequest.GetResponse();

                if (myHttpWebResponse.StatusCode == HttpStatusCode.OK)
                {
                    lastModified = myHttpWebResponse.LastModified.ToUniversalTime();

                    // Releases the resources of the response.
                    myHttpWebResponse.Close();

                    return lastModified;
                }
            }
            catch (WebException)
            {
                return null;
            }

            return null;
        }

        /// <summary>
        /// Downloads the NetCDF file from the downloadURL, converts it to CSV and puts it in the path
        /// SaveDownloadPath (class variable).
        /// </summary>
        /// <param name="downloadURL">The URL where the .nc file can be downloaded from.</param>
        /// <param name="downloadDestinationUri">The Uri where the .nc file will be saved.</param>
        /// <param name="dateChanged">The date when the file from the downloadURL whas changed.</param>
        /// <returns>true if file was successfully downloaded and database updated; false if not.</returns>
        public async Task<bool> DownloadFileAndUpdateDBAsync(string downloadURL, string downloadDestinationUri, DateTime dateChanged)
        {
            FileGetter fileGetter = new FileGetter(config);
            bool successfullyDownloadedFile = fileGetter.DownloadFile(downloadURL, downloadDestinationUri);
            var dateCached = DateTime.UtcNow;

            if (successfullyDownloadedFile)
            {
                filesDownloaded++;

                var filename = FileGetter.GetFilename(downloadURL);
                if (filename == null)
                {
                    throw new ArgumentException($"Could not extract filename from {downloadURL}.");
                }

                await BlobManager.UploadBlobAsync(filename, Path.Combine(downloadDestinationUri, filename), RawDataContainer);

                NetCDFConverter netCDFConverter = new NetCDFConverter(config);
                netCDFConverter.ConvertFile(
                    Path.Combine(downloadDestinationUri, filename),
                    Path.Combine(downloadDestinationUri, $"{filename}.csv")
                );

                await BlobManager.UploadBlobAsync($"{filename}.csv", Path.Combine(downloadDestinationUri, $"{filename}.csv"), CsvDataContainer);

                var rawDataFile = await FileDbContext.RawDataFiles.FindAsync(filename);

                // If file already exists in database, update it (this previously checked the fileExistsInDatabase param)
                if (rawDataFile != null)
                {
                    filesUpdated++;

                    // Load related reports for this raw data file and set their status to outdated
                    await FileDbContext.LoadRelatedAsync(rawDataFile);

                    rawDataFile.DateChanged = dateChanged;
                    rawDataFile.DateCached = dateCached;
                    rawDataFile.UploadedBy = config.CacherUploadedBy;
                    rawDataFile.ManuallyUploaded = false;

                    foreach (var reportRelation in rawDataFile.ReportRelations)
                    {
                        reportRelation.Report.Status = "Outdated";
                    }
                }
                else
                {
                    newFiles++;
                    FileDbContext.RawDataFiles.Add(new RawDataFile(filename, dateChanged, dateCached, config.CacherUploadedBy, manuallyUploaded: false));
                }

                await FileDbContext.SaveChangesAsync();

                return true;
            }

            return false;
        }

        public void DeleteTemporaryFiles()
        {
        }

        /// <summary>
        /// Loops through all the months from startDate to endDate to check for modified files. For each month
        /// every station listed in the class variable StationNames is iterated through. For each station each file type
        /// from the class varible FileTypes is iterated through. The code checks if each file is downloaded/cached.
        /// If the file is not cached, the file gets downloaded/cached and a new entry in the database is created.
        /// The database entry contains filename (PK), date_changed, date_cached and uploaded_by.
        /// If the file is cached the date_created info is gathered from the file on the web server and compared with
        /// the cached files date_changed from the Database. If the file on the web server is newer than the file that is cached,
        /// the file on the web server is downloaded and replaces the cached file. The Database is then updated.
        /// </summary>
        /// <param name="startDate">Start date for checking for raw data files.</param>
        /// <param name="endDate">End date for checking for raw data files.</param>
        /// <param name="downloadDestinationUri">The Uri where the .nc file will be saved.</param>
        public async Task CheckForModifiedFilesAsync(DateTime startDate, DateTime endDate, string downloadDestinationUri)
        {
            if (startDate > endDate)
            {
                throw new ArgumentException($"Parameter startDate; {startDate.ToString("dd-MM-yyyy HH:mm:ss")} needs to be before endDate; {startDate.ToString("dd-MM-yyyy HH:mm:ss")}", nameof(startDate));
            }

            try
            {
                while (startDate < endDate)
                {
                    int year = startDate.Year;
                    string month = startDate.ToString("MM");

                    foreach (string station in config.AllStationNames)
                    {
                        foreach (string filetype in config.FileDataTypes)
                        {
                            string ncFilename = $"{year}{month}_E39_{station}_{filetype}.nc";
                            string downloadUrl = $"{config.DownloadUrl}/{year}/{month}/{ncFilename}";
                            string infoUrl = $"{config.CacherFileInfoUrl}/{year}/{month}/{ncFilename}.das";

                            DateTime? DateTimeOrNullObejct = GetDateModified(infoUrl);
                            /* If DateTimeOrNullObject is null, it was not possible to get date_created info from the NetCDF file. */
                            if (DateTimeOrNullObejct == null)
                            {
                                Console.WriteLine($"Could not get Last-Modified header from URL: {infoUrl}");
                                log.Debug($"Could not get Last-Modified header from URL: {infoUrl}");
                                continue;
                            }
                            DateTime dateChangedRemotely = (DateTime)DateTimeOrNullObejct;

                            var cachedRawDataFile = await FileDbContext.RawDataFiles.FindAsync(ncFilename);

                            /* If cachedRawDataFile is null, the NetCDF file does not exist in the database.
                               File will then get downloaded and the database will be updated. */
                            if (cachedRawDataFile == null)
                            {
                                Console.WriteLine($"The file: {ncFilename} does not exist in the database. Adding this file now.");
                                log.Debug($"The file: {ncFilename} does not exist in the database. Adding this file now.");

                                await DownloadFileAndUpdateDBAsync(downloadUrl, downloadDestinationUri, dateChangedRemotely);
                            }
                            else if (cachedRawDataFile.ManuallyUploaded)
                            {
                                // The file was manually uploaded by a user, so the cacher should not try to update it.
                                Console.WriteLine($"The file {ncFilename} was uploaded manually, leaving it as it is.");
                                log.Debug($"The file {ncFilename} was uploaded manually, leaving it as it is.");
                            }
                            else
                            {
                                /* This part of the code will run if the file exists in the database,
                                 * was not uploaded manually, and the date_created info has been
                                 * collected from the NetCDF file. */

                                if (!await BlobManager.BlobExistsAsync(ncFilename, RawDataContainer))
                                //if (!File.Exists(Path.Combine(config.RawDataDirectory, ncFilename)))
                                {
                                    // If the file exists in the database but not in blobstorage.
                                    Console.WriteLine($"The; {ncFilename}, exists in the database, but is not downloaded. Updating database and downloading file.");
                                    log.Debug($"The; {ncFilename}, exists in the database, but is not downloaded. Updating database and downloading file.");

                                    await DownloadFileAndUpdateDBAsync(downloadUrl, downloadDestinationUri, dateChangedRemotely);
                                }
                                else if (cachedRawDataFile.DateCached < dateChangedRemotely)
                                {
                                    // The cached file is outated.
                                    Console.WriteLine($"It exists a newer version of the cached file; {ncFilename}. Updating database and cached file.");
                                    log.Debug($"It exists a newer version of the cached file; {ncFilename}. Updating database and cached file.");

                                    await DownloadFileAndUpdateDBAsync(downloadUrl, downloadDestinationUri, dateChangedRemotely);
                                }
                                else
                                {
                                    // The cached file is the newest file.
                                    Console.WriteLine($"The cached file; {ncFilename} is up to date.");
                                    log.Debug($"The cached file; {ncFilename} is up to date.");
                                }
                            }
                        }
                    }

                    startDate = startDate.AddMonths(1);
                }
            }
            catch (Exception e)
            {
                log.Error($"There was an error when looking for modified files.", e);
            }

            log.Info($"Files downloaded: {filesDownloaded} -- Files updated: {filesUpdated} -- New files: {newFiles}");
            log.Info($"Files not available for download: \n{filesNotAvailableForDownload}");
        }
    }
}