﻿using E39.WebApplication.Data;
using E39.WebApplication.Hubs;
using E39.WebApplication.Jobs;
using E39.WebApplication.Scheduler;
using log4net;
using Microsoft.AspNetCore.Authentication;
using Microsoft.AspNetCore.Authentication.AzureAD.UI;
using Microsoft.AspNetCore.Authentication.OpenIdConnect;
using Microsoft.AspNetCore.Authorization;
using Microsoft.AspNetCore.Builder;
using Microsoft.AspNetCore.Hosting;
using Microsoft.AspNetCore.Http;
using Microsoft.AspNetCore.Mvc;
using Microsoft.AspNetCore.Mvc.Authorization;
using Microsoft.EntityFrameworkCore;
using Microsoft.Extensions.Configuration;
using Microsoft.Extensions.DependencyInjection;
using Microsoft.Extensions.Logging;
using Quartz;
using Quartz.Impl;
using Quartz.Spi;
using System;
using System.IO;

namespace E39.WebApplication
{
    public class Startup
    {
        private readonly ILog Log = LogManager.GetLogger(typeof(Startup));

        public IConfiguration Configuration { get; }

        public Startup(IConfiguration configuration)
        {
            Configuration = configuration;
        }

        // This method gets called by the runtime. Use this method to add services to the container.
        public void ConfigureServices(IServiceCollection services)
        {
            services.Configure<WebAppConfig>(Configuration.GetSection("PipelineConfig"));

            services.Configure<CookiePolicyOptions>(options =>
            {
                // This lambda determines whether user consent for non-essential cookies is needed for a given request.
                options.CheckConsentNeeded = context => true;
                options.MinimumSameSitePolicy = SameSiteMode.None;
            });

            services.AddAuthentication(AzureADDefaults.AuthenticationScheme)
                .AddAzureAD(options => Configuration.Bind("AzureAd", options));

            services.Configure<OpenIdConnectOptions>(AzureADDefaults.OpenIdScheme, options =>
            {
                options.Authority = options.Authority + "/v2.0/";
                options.TokenValidationParameters.ValidateIssuer = true;
            });

            services.AddAntiforgery(options =>
            {
                options.HeaderName = "X-CSRF-Token";
            });

            services.AddDbContext<FileDbContext>(options =>
            {
                options.UseSqlServer(Configuration.GetConnectionString("FileDb"));
            });

            services.AddScoped<BlobManager>(provider =>
            {
                return new BlobManager(Configuration.GetConnectionString("e39files_AzureStorageConnectionString"));
            });

            services.AddMvc(options =>
            {
                var policy = new AuthorizationPolicyBuilder()
                    .RequireAuthenticatedUser()
                    .Build();
                options.Filters.Add(new AuthorizeFilter(policy));
            })
            .SetCompatibilityVersion(CompatibilityVersion.Version_2_2);

            services.AddSingleton<IJobManager, JobManager>();
            services.AddHostedService<JobHubEventPropagator>();
            services.AddSignalR();

            /*
             * Add the scheduled jobs to the DI system, so they can have their required services injected.
             * Add the JobFactory to the DI system, so that it has access to the ServiceProvider when it creates new jobs.
             * Add Quartz scheduler as a singleton.
             * See Program.cs for the actual scheduling of the jobs.
             */
            services.AddTransient(typeof(DownloadFilesJob));
            services.AddTransient(typeof(CleanUpFilesJob));
            services.AddSingleton<IJobFactory, ScheduledJobFactory>();
            services.AddSingleton<IScheduler>(provider =>
            {
                var schedulerFactory = new StdSchedulerFactory();
                var scheduler = schedulerFactory.GetScheduler().Result;
                // Use our custom JobFactory, which enables injecting dependencies into the scheduled jobs
                scheduler.JobFactory = provider.GetService<IJobFactory>();
                scheduler.Start();
                return scheduler;
            });

            var webAppConfig = Configuration.GetSection("PipelineConfig").Get<WebAppConfig>();
            // Clean the temporary directory and create a new one
            try
            {
                Directory.Delete(webAppConfig.BaseTempDirectory, true);
            }
            catch (Exception) { }
            Directory.CreateDirectory(webAppConfig.BaseTempDirectory);
        }

        // This method gets called by the runtime. Use this method to configure the HTTP request pipeline.
        public void Configure(IApplicationBuilder app, IHostingEnvironment env, ILoggerFactory loggerFactory)
        {
            if (env.IsDevelopment())
            {
                app.UseDeveloperExceptionPage();
            }
            else
            {
                app.UseExceptionHandler("/Home/Error");
                // The default HSTS value is 30 days. You may want to change this for production scenarios, see https://aka.ms/aspnetcore-hsts.
                app.UseHsts();
            }

            app.UseHttpsRedirection();
            app.UseStaticFiles();
            app.UseCookiePolicy();

            app.UseAuthentication();

            app.UseSignalR(routes =>
            {
                routes.MapHub<JobHub>("/JobHub");
            });

            loggerFactory.AddLog4Net(); // << Add this line

            app.UseMvc(routes =>
            {
                routes.MapRoute(
                    name: "default",
                    template: "{controller=Home}/{action=Index}/{id?}");
            });

            Log.Info("E39 wave data pipeline web application started");
        }
    }
}