﻿/// <reference path="../lib/redux/index.d.ts" />

let currentJobGuid = window.localStorage.getItem('currentRunningJobGuid');
let jobHubConnection = null;
let jobEstimatedTimeFinished = null;

/**
 * Enum with the different states a Job may be in.
 * Equals the JobState enum in the backend.
 */
const JobState = {
    CREATED: 0,
    RUNNING: 1,
    COMPLETED: 2,
    FAILED: 3,
    ABORTED: 4
};

/**
 * Enum with the different styles a Job message may have.
 * Equals the JobMessageStyles enum in the backend.
 */
const JobMessageStyle = {
    NORMAL: 0,
    EMPHASIS: 1,
    WARNING: 2,
    ERROR: 3
};

/** Mapping from a job message style to a CSS class name */
const JobMessageCssClassMap = {
    [JobMessageStyle.NORMAL]: '',
    [JobMessageStyle.EMPHASIS]: 'job-message-emphasis',
    [JobMessageStyle.WARNING]: 'job-message-warning',
    [JobMessageStyle.ERROR]: 'job-message-error',
};

/** The initial view state of the GUI */
const init_state = {
    job_state: JobState.CREATED,
    job_log: "",
    job_progress: 0,
    time_left: null,

    valid_dates: true,
    stations_selected: false,

    alert_dismissed: true
};

/** A Redux store for managing the view state */
const store = Redux.createStore(reduce, init_state, window.__REDUX_DEVTOOLS_EXTENSION__ && window.__REDUX_DEVTOOLS_EXTENSION__({ trace: true }));
store.subscribe(() => render(store.getState()));

/**
 * Renders the whole view given the view state object.
 * @param {object} state the view state
 */
function render(state) {
    // The gui elements
    let e = {
        statusBox: $("#status-box"),
        etaHeader: $("#eta-header"),
        jobProgressBar: $("#job-progress-bar"),
        createReportButton: $("#create-report-button"),
        downloadReportButton: $("#download-report-button"),
        cancelJobButton: $("#cancel-job-button"),
        selectStationPicker: $("#select-station-picker"),
        fromDateOverlay: $("#from-date-overlay"),
        toDateOverlay: $("#to-date-overlay"),
        mlSwitch: $("#ml-switch"),
        alertTextWrapper: $('#alert-text-wrapper'),
        invalidInputAlert: $('#invalid-input-alert')
    };

    // Update estimated time left-header based on job state and time left
    if (state.job_state === JobState.RUNNING) {
        if (state.time_left) {
            if (state.time_left.asSeconds() > 0) {
                let min = Math.floor(state.time_left.asMinutes());
                let sec = state.time_left.seconds();
                if (min > 0) {
                    e.etaHeader.text('Estimated time left: ' + min + ' minutes and ' + sec + ' seconds');
                } else {
                    e.etaHeader.text('Estimated time left: ' + sec + ' seconds');
                }
            } else {
                e.etaHeader.text('Still working, please wait...');
            }
        } else {
            e.etaHeader.text('Estimating time...');
        }
    } else if (state.job_state === JobState.COMPLETED) {
        e.etaHeader.text('Completed');
    } else if (state.job_state === JobState.FAILED) {
        e.etaHeader.text('Failed');
    } else if (state.job_state === JobState.ABORTED) {
        e.etaHeader.text('Job cancelled');
    } else {
        e.etaHeader.text('Select dates and stations and create reports');
    }

    // Enable/disable input fields and animate progress bar based on job state
    if (state.job_state === JobState.RUNNING) {
        e.createReportButton.prop("disabled", true);
        e.cancelJobButton.prop("disabled", false);
        e.selectStationPicker.prop("disabled", true);
        e.fromDateOverlay.css("display", "block");
        e.toDateOverlay.css("display", "block");
        e.mlSwitch.prop("disabled", true);
        e.jobProgressBar.addClass('progress-bar-animated');
    } else {
        // Create report button may also be disabled because of invalid input
        if (!state.valid_dates || !state.stations_selected) {
            e.createReportButton.prop("disabled", true);
        } else {
            e.createReportButton.prop("disabled", false);
        }
        e.cancelJobButton.prop("disabled", true);
        e.selectStationPicker.prop("disabled", false);
        e.fromDateOverlay.css("display", "none");
        e.toDateOverlay.css("display", "none");
        e.mlSwitch.prop("disabled", false);
        e.jobProgressBar.removeClass('progress-bar-animated');
    }

    // Only enable download button when job is completed
    if (state.job_state === JobState.COMPLETED) {
        e.downloadReportButton.prop("disabled", false);
        e.downloadReportButton.addClass("ready-for-download");
    } else {
        e.downloadReportButton.prop("disabled", true);
        e.downloadReportButton.removeClass("ready-for-download");
    }

    // Set colors based on job state
    let statusBoxColor = '';
    let jobProgressBarClass = '';

    if (state.job_state === JobState.FAILED) {
        jobProgressBarClass = 'bg-danger';
    } else if (state.job_state === JobState.ABORTED) {
        statusBoxColor = '#B0B0B0';
        jobProgressBarClass = 'bg-secondary';
    } else if (state.job_state === JobState.COMPLETED) {
        jobProgressBarClass = 'bg-success';
    }

    e.statusBox
        .html(state.job_log.replace(/(?:\r\n|\r|\n)/g, '<br>'))
        .css('color', statusBoxColor);

    e.jobProgressBar
        .css('width', state.job_progress + '%')
        .prop('aria-valuenow', state.job_progress)
        .removeClass('bg-danger bg-secondary bg-success')
        .addClass(jobProgressBarClass);

    // Show alert if dates are invalid and the user hasn't dismissed the alert
    if (!state.valid_dates && !state.alert_dismissed) {
        e.invalidInputAlert.show();
        e.alertTextWrapper.html(
            '<strong>Invalid input!</strong> "From date" is after or at the same date ' +
            'and time as "To date". "From date" needs to be before "To date".'
        );
    } else {
        e.invalidInputAlert.hide();
        e.alertTextWrapper.html('');
    }
}

/**
 * The reducer function which Redux uses. Accepts the current state and an action, and returns a new state.
 * @param {object} state The current state object.
 * @param {object} action An action. Must have a type attribute, and may contain extra data related to the action.
 */
function reduce(state, action) {
    switch (action.type) {
        // SignalR actions

        case 'job_message_added':
            let msg = action.eventArgs.message.text;
            let cssClass = JobMessageCssClassMap[action.eventArgs.message.style];
            if (cssClass) {
                return { ...state, job_log: state.job_log + "<p class=\"" + cssClass + "\">" + msg + "</p>" };
            } else {
                return { ...state, job_log: state.job_log + "<p>" + msg + "</p>" };
            }

        case 'job_state_changed':
            return { ...state, job_state: action.eventArgs.newState };

        case 'job_progress_changed':
            return { ...state, job_progress: action.eventArgs.newProgress };

        case 'receive_all_job_info':
            let logHtml = '';
            for (let message of action.jobInfo.log) {
                let cssClass = JobMessageCssClassMap[message.style];
                if (cssClass) {
                    logHtml += "<p class=\"" + cssClass + "\">" + message.text + "</p>";
                } else {
                    logHtml += "<p>" + message.text + "</p>";
                }
            }

            return {
                ...state,
                job_log: logHtml,
                job_state: action.jobInfo.state,
                job_progress: action.jobInfo.progress
            };


        // Ajax callback actions

        case 'create_job_succeeded':
        case 'current_job_not_existing':
            return {
                ...state,
                job_state: JobState.CREATED,
                job_log: "",
                job_progress: 0
            };

        case 'create_job_failed':
            return {
                ...state,
                job_state: JobState.FAILED,
                job_log: "Couldn't start job",
                job_progress: 100
            };

        case 'download_reports_clicked':
            return {
                ...state,
                job_state: JobState.CREATED,
                job_progress: 0
            };


        // Input change/validation actions

        case 'input_date_changed':
            /* Check if fromDate is after or at the same time as toDate.
             If it is, tell the user that the chosen values are invalid. */
            if (action.fromDate >= action.toDate) {
                return { ...state, valid_dates: false, alert_dismissed: false };
            } else {
                return { ...state, valid_dates: true };
            }

        case 'input_stations_changed':
            return { ...state, stations_selected: action.stations.length > 0 };

        case 'alert_dismissed':
            return { ...state, alert_dismissed: true };

        case 'time_left_updated':
            return { ...state, time_left: action.timeLeft };


        case '@@INIT':
            return state;

        default:
            console.error("Unknown reducer action: " + action.type);
            return state;
    }
}

/**
 * Handler function for SignalR errors.
 * @param {any} err the error object
 */
function handleSignalRError(err) {
    return console.error(err.toString());
}

/**
 * Callback for succeeded ajax call which starts a new report generation job.
 * @param {Object} result the json result returned from the ajax call
 * @param {boolean} result.exists whether the job started successfully
 * @param {string} [result.guid] if exists is true, the GUID of the newly created job
 */
function startCreateReportJobSuccess(result) {
    if (result.success) {
        currentJobGuid = result.guid;
        window.localStorage.setItem('currentRunningJobGuid', result.guid);

        store.dispatch({ type: 'create_job_succeeded' });

        jobHubConnection.invoke("GetAllJobInfo", currentJobGuid).catch(handleSignalRError);
    }
    else {
        store.dispatch({ type: 'create_job_failed' });
    }
}

/**
 * Callback for succeeded ajax call which checks if a job exists.
 * @param {Object} result the json result returned from the ajax call
 * @param {boolean} result.exists whether the job exists
 * @param {number} [result.state] if exists is true, the state of the job
 */
function checkJobGetSuccess(result) {
    if (result.exists) {
        jobHubConnection.invoke("GetAllJobInfo", currentJobGuid).catch(handleSignalRError);
    }
    else {
        currentJobGuid = null;
        window.localStorage.removeItem('currentRunningJobGuid');
        store.dispatch({ type: 'current_job_not_existing' });
    }
}

/**
 * Sends a POST request for starting a new job for creating reports.
 * @param {string} fromDate The start date to create report for, on format DD-MM-YYYY HH:mm
 * @param {string} toDate The end date to create reports for, on format DD-MM-YYYY HH:mm
 * @param {string | string[]} stations An array of strings with station names to create reports for
 * @param {boolean} impute Whether to impute missing data points
 */
function startCreateReportJob(fromDate, toDate, stations, impute) {
    return $.ajax({
        url: 'jobs',
        type: 'POST',
        headers: {
            'X-CSRF-Token': antiXsrfRequestToken
        },
        dataType: 'json',
        contentType: 'application/json; charset=UTF-8',
        data: JSON.stringify({
            StartDate: fromDate,
            EndDate: toDate,
            Stations: stations,
            Impute: impute
        }),
        success: startCreateReportJobSuccess,
        error: function startCreateReportJobError(jqXHR, textStatus, errorThrown) {
            store.dispatch({ type: 'create_job_failed' });
        }
    });
}

/**
 * Populates the list of station names in the station picker.
 * @param {string[]} stationNames The list of station names
 */
function populateStationNamePicker(stationNames) {
    for (let station of stationNames) {
        $('.selectpicker').append('<option>' + station + '</option>');
        $('.selectpicker').selectpicker('refresh');
    }
}

/**
 * Requests cancellation of the currently running job.
 */
function tryAbortCurrentJob() {
    return $.ajax({
        url: 'jobs/' + currentJobGuid + '/abort',
        type: 'POST',
        headers: {
            'X-CSRF-Token': antiXsrfRequestToken
        },
        success: function tryAbortCurrentJobSuccess(result) {
            store.dispatch({
                type: 'job_message_added',
                eventArgs: { message: { text: "Cancelling...", style: JobMessageStyle.NORMAL } }
            });
        },
        error: function tryAbortCurrentJobError(jqXHR, textStatus, errorThrown) {
            console.log(jqXHR.responseJSON.value);
        }
    });
}

/**
 * Timer handler function for calculating estimating remaining time of the job.
 */
function estimatedTimeLeftCountdown() {
    if (jobEstimatedTimeFinished && store.getState().job_state === JobState.RUNNING) {
        let timeLeft = moment.duration(moment(jobEstimatedTimeFinished).diff(moment()));
        store.dispatch({ type: 'time_left_updated', timeLeft: timeLeft });
    }
    window.setTimeout(estimatedTimeLeftCountdown, 1000);
}

$(function () {
    /* Hide alert when x-button is clicked */
    $('#close-alert-button').click(function closeAlertButtonClicked() {
        store.dispatch({ type: 'alert_dismissed' });
    });

    /* Send AJAX to backend for creating a report on create report button click */
    $('#create-report-button').click(function createReportButtonClicked() {
        let fromDate = moment($('#datetime-picker-from-date').data('datetimepicker').date()).toISOString();
        let toDate = moment($('#datetime-picker-to-date').data('datetimepicker').date()).toISOString();
        let stations = $('#select-station-picker').val();
        let useML = $('#ml-switch').is(":checked");

        startCreateReportJob(fromDate, toDate, stations, useML);
    });

    /* Attempt to abort running job on cancel button click */
    $('#cancel-job-button').click(tryAbortCurrentJob);

    /* Used to check if the chosen dates are valid. */
    $('.datetime-picker').on('change.datetimepicker', function datetimePickerChanged() {
        let fromDate = $('#datetime-picker-from-date').data('datetimepicker'); // Get fromDate from dateTimePicker as a Moment object
        let toDate = $('#datetime-picker-to-date').data('datetimepicker'); // Get toDate from dateTimePicker as a Moment object

        if (typeof fromDate !== 'undefined' && typeof toDate !== 'undefined') {
            store.dispatch({
                type: 'input_date_changed',
                fromDate: fromDate.date(),
                toDate: toDate.date()
            });
        }
    });

    /* Used to check if at least 1 station is chosen. Enables the "Create report" button if it is. */
    $('#select-station-picker').on('change', function selectStationPickerChanged() {
        store.dispatch({
            type: 'input_stations_changed',
            stations: $('#select-station-picker').val()
        });
    });

    /* Download reports on button click, and remove the job guid from localstorage. */
    $('#download-report-button').click(function downloadReportButtonClicked() {
        window.open("jobs/" + currentJobGuid + "/reports", "_blank");
        currentJobGuid = null;
        window.localStorage.removeItem('currentRunningJobGuid');
        store.dispatch({ type: 'download_reports_clicked' });
    });

    /* Set the data for the fromDate dateTime picker */
    $('#datetime-picker-from-date').datetimepicker({
        inline: true,
        sideBySide: true,
        stepping: 10,
        timeZone: 'Etc/UTC',
        format: 'D-MMMM-Y HH:mm',
        defaultDate: moment().utc().startOf('month')
    });

    /* Set the data for the toDate dateTime picker */
    $('#datetime-picker-to-date').datetimepicker({
        inline: true,
        sideBySide: true,
        stepping: 10,
        timeZone: 'Etc/UTC',
        format: 'D-MMMM-Y HH:mm',
        defaultDate: moment().utc().endOf('month').set({ minute: 50, second: 0, millisecond: 0 })
    });

    /* Add stations names to the station names dropdown menu. */
    $.ajax({
        url: 'station-names',
        type: 'GET',
        success: populateStationNamePicker
    });

    // Connect to SignalR JobHub
    jobHubConnection = new signalR.HubConnectionBuilder().withUrl("/JobHub").build();

    jobHubConnection.on("JobMessageAdded", function signalRJobMessageAdded(jobId, eventArgs) {
        if (jobId === currentJobGuid) {
            store.dispatch({ type: 'job_message_added', eventArgs });
            $("#status-box").scrollTop($("#status-box").get(0).scrollHeight);
        }
    });

    jobHubConnection.on("JobStateChanged", function signalRJobStateChanged(jobId, eventArgs) {
        if (jobId === currentJobGuid) {
            store.dispatch({ type: 'job_state_changed', eventArgs });
        }
    });

    jobHubConnection.on("JobTimeFinishedEstimated", function signalRJobTimeFinishedEstimated(jobId, eventArgs) {
        if (jobId === currentJobGuid) {
            jobEstimatedTimeFinished = eventArgs.timeEstimatedFinished;
        }
    });

    jobHubConnection.on("JobProgressChanged", function signalRJobProgressChanged(jobId, eventArgs) {
        if (jobId === currentJobGuid) {
            store.dispatch({ type: 'job_progress_changed', eventArgs });
        }
    });

    jobHubConnection.on("ReceiveAllJobInfo", function signalRReceiveAllJobInfo(jobId, jobInfo) {
        if (jobId === currentJobGuid) {
            store.dispatch({ type: 'receive_all_job_info', jobInfo });
            // Check if estimated time finished actually has been calculated yet
            if (jobInfo.timeStarted < jobInfo.timeEstimatedFinished) {
                jobEstimatedTimeFinished = jobInfo.timeEstimatedFinished;
            }
            $("#status-box").scrollTop($("#status-box").get(0).scrollHeight);
        }
    });

    jobHubConnection.start().then(function signalRConnectedToJobHub() {
        console.log("SignalR connected");

        // On page load, if we have a job guid in localstorage, check if it's still running
        if (currentJobGuid !== null) {
            $.ajax({
                url: 'jobs/' + currentJobGuid,
                method: 'GET',
                success: checkJobGetSuccess
            });
        }
    }).catch(handleSignalRError);

    // Start the time left countdown timer
    estimatedTimeLeftCountdown();

    // Render the initial state
    render(store.getState());
});
